"""
This is the code for the Continuous Causal States method
described in the papers:

  Discovering Causal Structure with Reproducing-Kernel Hilbert Space ε-Machines
  by Nicolas Brodu and James P. Crutchfield

  Inferring Kernel ε-Machines: Discovering Structure in Complex Systems
  by Nicolas Brodu and Alexandra M. Jurgens1

  and possibly more to come
  
This code is free software, according to the MIT license
Author & Contact: nicolas.brodu@inria.fr
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
import scipy.linalg as lin
import time
from mpl_toolkits import mplot3d

last_time = time.time()

def print_elapsed(msg):
    now = time.time()
    global last_time
    print("%d: %s" % (now-last_time, msg))
    last_time = now

def plot3D(X, ax = None, color = None, split_indices = None):
    """Basic 3D plotter using matplotlib
    """
    
    try:
        import cupy as cp
        X = cp.asnumpy(X)
        if color is not None:
            color = cp.asnumpy(color)
    except:
        pass
    
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')
    
    if split_indices is None:
        split_indices = [ X.shape[0] ]

    start_idx = 0
    for si, split_idx in enumerate(split_indices):
        # Make the line multi-coloured by plotting it in segments which
        # change in colour across the whole time series.
        n = split_idx-start_idx
        s = (n+99)//100
        if type(color) is list:
            if type(color[si]) is np.ndarray:
                c = color[si]
            else:
                c = np.full(n, color[si])
        elif type(color) is np.ndarray:
            if color.ndim==1:
                c = np.full(n, color[si]) #color[start_idx:split_idx]
            else:
                c = color[:,si]
        else:
            c = np.linspace(0,1,n)
        for j in range(0,n-s,s):
            i = start_idx + j
            #ax.plot(X[i:i+s+1,0], X[i:i+s+1,1], X[i:i+s+1,2], color=(1,c[i],1-c[i]), alpha=0.4)
            ax.plot(X[i:i+s+1,0], X[i:i+s+1,1], X[i:i+s+1,2], color=cm.plasma(c[j]))
            #print(i,i+s+1)
        start_idx = split_idx

    # Remove all the axis clutter, leaving just the curve.
    #ax.set_axis_off()
    plt.show()
    return ax

def scatter3D(X, split_indices = None):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    if split_indices is None:
        split_indices = [ X.shape[0] ]
    start_idx = 0
    for split_idx in split_indices:
        n = split_idx-start_idx
        s = (n+99)//100
        c = np.linspace(0,1,n)
        for j in range(0,n-s,s):
            i = split_idx + j
            ax.scatter(X[i:i+s+1,0], X[i:i+s+1,1], X[i:i+s+1,2], color=cm.plasma(c[j]))
        start_idx = split_idx
    plt.show()


def plot2D(X, split_indices = None):
    fig = plt.figure()
    ax = fig.gca()
    if split_indices is None:
        split_indices = [ X.shape[0] ]
    start_idx = 0
    for split_idx in split_indices:
        n = split_idx-start_idx
        s = (n+99)//100
        c = np.linspace(0,1,n)
        for j in range(0,n-s,s):
            i = split_idx + j
            ax.plot(X[i:i+s+1,0], X[i:i+s+1,1], color=cm.plasma(c[j]))
        start_idx = split_idx
    plt.show()

def scatter2D(X, split_indices = None):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    if split_indices is None:
        split_indices = [ X.shape[0] ]
    start_idx = 0
    for split_idx in split_indices:
        n = split_idx-start_idx
        s = (n+99)//100
        c = np.linspace(0,1,n)
        for j in range(0,n-s,s):
            i = split_idx + j
            ax.scatter(X[i:i+s+1,0], X[i:i+s+1,1], color=cm.plasma(c[j]))
        start_idx = split_idx
    plt.show()



def plot_series(series, predictions = None, extrapolations = None, timestamps = None, pred_timestamps = None, labels = None, title = None, qcflags = None, selected_series = -1, filename = None, display = True, renderer='browser'):
    
    import plotly.graph_objects as go
    import plotly.offline
    import plotly.subplots

    series_list = series if isinstance(series,list) or isinstance(series,tuple) else [ series ]

    ts_list = timestamps if isinstance(timestamps,list) or isinstance(timestamps,tuple) else [ timestamps ] * len(series_list)
    
    qc_list = qcflags if isinstance(qcflags,list) or isinstance(qcflags,tuple) else [ qcflags ] * len(series_list)
    
    # the default argument results in a list of None
    # Predictions may be provided partially, for some components only...
    pred_list = predictions if isinstance(predictions,list) or isinstance(predictions,tuple) else [ predictions ] * len(series_list)

    # ...and with possibly different temporal schemes    
    pred_ts_list = pred_timestamps if isinstance(pred_timestamps,list) or isinstance(pred_timestamps,tuple) else [ pred_timestamps ] * len(series_list)
    
    ex_list = extrapolations if isinstance(extrapolations,list) or isinstance(extrapolations,tuple) else [ extrapolations ] * len(series_list)
    
    labels_list = labels if isinstance(labels,list) or isinstance(labels,tuple) else [ f"Series {i+1}" for i in range(len(series_list))]
    
    ########### Plots of each target series + gap filling #############
    
    fig = go.Figure()
    traces_xs = []
    traces_ys = []
    traces_onoff = []
    
    for ser, ts, qc, pred, pred_ts, ex in zip(series_list, ts_list, qc_list, pred_list, pred_ts_list, ex_list):
        ser_xs = []
        ser_ys = []
        
        n = ser.shape[0] if hasattr(ser, "shape") else len(ser)
        
        if ts is None:
           ts = np.arange(n)
        
        # Need to make the same number of traces, even if some are missing
        # (no predictions, no extrapolations...) for some series
        # => use nan values and TODO: show/hide the trace in the legend
        
        # Note: There is some kind of bug in pyplot, so the x/y of the
        # first trace are used even if NaN => need to start with the data
        
        # (valid) data
        ser_xs += [ ts ]
        if qc is None:
            valid_data = ser[:,0]
        else:
            valid_data = ser[:,0].copy()
            valid_data[qc<=0] = np.nan
        ser_ys += [ valid_data ]
        nvalid = np.sum(~np.isnan(valid_data))
        
        # predictions?
        if pred is None:
            ser_xs += [ [np.nan] ]
            ser_ys += [ [np.nan] ]
            npred = 0
        else:
            npred = pred.shape[0] if hasattr(pred, "shape") else len(pred)
            # no timestamps => auto
            if pred_ts is None:
                ser_xs += [ ts[-1] + np.arange(npred+1) ]
            else:
                ser_xs += [ np.concatenate([[ts[-1]], pred_ts]) ]
            ser_ys += [ np.squeeze(np.concatenate((ser[-1:,:], pred), axis=0)) ]

        
        # invalid data, with junction to nearby valid data
        if qc is None:
            ser_xs += [ [np.nan] ]
            ser_ys += [ [np.nan] ]
            ninvalid = 0
        else:
            tvidx = np.arange(n)[qc>0]
            gaplocs = np.arange(len(tvidx)-1)[np.diff(tvidx)!=1]
            gaplens = tvidx[gaplocs+1] - tvidx[gaplocs] + 1
            gaplocs = tvidx[gaplocs]
            if qc[0]<=0:
                gaplocs = np.concatenate([[0], gaplocs])
                gaplens = np.concatenate([[tvidx[0]+1], gaplens])
            if qc[-1]<=0:
                gaplocs = np.concatenate([[tvidx[-1]], gaplocs])
                gaplens = np.concatenate([[n-tvidx[-1]], gaplens])
            invalid_data = np.full(n, np.nan)
            for gloc,glen in zip(gaplocs,gaplens):
                invalid_data[gloc:gloc+glen] = ser[gloc:gloc+glen,0]
            ser_xs += [ ts ]
            ser_ys += [ invalid_data ]
            ninvalid = np.sum(~np.isnan(invalid_data))
            
        # extrapolated data
        if ex is None:
            ser_xs += [ [np.nan] ]
            ser_ys += [ [np.nan] ]
            ngap = 0
            ser_xs += [ [np.nan] ]
            ser_ys += [ [np.nan] ]
            nfit = 0
        else:
            # gap filling
            if qc is None:
                ser_xs += [ [np.nan] ]
                ser_ys += [ [np.nan] ]
                ngap = 0
            else:
                ser_xs += [ ts ]
                gapfilled = np.full(ser.shape[0], np.nan)
                for gloc,glen in zip(gaplocs,gaplens):
                    gapfilled[gloc:gloc+glen] = ex[gloc:gloc+glen,0]
                ser_ys += [ gapfilled ]
                ngap = np.sum(~np.isnan(gapfilled))
            
            # data fit
            if qc is None:
                fit_data = ex[:,0]
            else:
                fit_data = ex[:,0].copy()
                fit_data[qc<=0] = np.nan
            ser_xs += [ ts ]
            ser_ys += [ fit_data ]
            nfit = np.sum(~np.isnan(fit_data))
            
        traces_xs += [ser_xs]
        traces_ys += [ser_ys]
        traces_onoff += [ [bool(nvalid>0), bool(npred>0), bool(ninvalid>0), bool(ngap>0), bool(nfit>0)] ]
    
    sel = len(series_list) + selected_series if selected_series < 0 else selected_series
    
    fig.add_scatter(
        x = traces_xs[sel][0],
        y = traces_ys[sel][0],
        mode = 'lines',
        line = dict(color='blue'),
        name = 'Data',
        showlegend = traces_onoff[sel][0]
    )
    fig.add_scatter(
        x = traces_xs[sel][1],
        y = traces_ys[sel][1],
        mode = 'lines',
        line = dict(color='black'),
        name = 'Predictions',
        legendrank = 1001,  # plotly >= v5.0, put after the others
        showlegend = traces_onoff[sel][1]
    )
    fig.add_scatter(
        x = traces_xs[sel][2],
        y = traces_ys[sel][2],
        mode = 'lines',
        line = dict(color='red'),
        name = 'Invalid data',
        showlegend = traces_onoff[sel][2]
    )
    fig.add_scatter(
        x = traces_xs[sel][3],
        y = traces_ys[sel][3],
        mode = 'lines',
        line = dict(color='green'),
        name = 'Gap filling',
        showlegend = traces_onoff[sel][3]
    )
    fig.add_scatter(
        x = traces_xs[sel][4],
        y = traces_ys[sel][4],
        mode = 'lines',
        line = dict(color='#008080'),
        name = 'Data fit',
        visible='legendonly',
        showlegend = traces_onoff[sel][4]
    )
    
    fig.update_layout(updatemenus=[dict(
        active=sel,
        buttons=[dict(label=f"{labels_list[i]}",
                      method="update",
                      args=[{"x":traces_xs[i],
                             "y":traces_ys[i],
                             "showlegend":traces_onoff[i]
                      }],
                 ) for i in range(len(series))
        ],
        showactive=True,
        xanchor='left',
        yanchor='bottom',
        x=0,
        y=1,
        pad_b=5,
    )])
    fig.update_layout(legend=dict(yanchor='top',xanchor='right',y=0.99,x=0.99,title_text='Click to show/hide:'))
    fig.update_layout(showlegend=True)
    fig.update_layout(hovermode=False)
    fig.update_layout(clickmode='select')
    fig.update_layout(margin = dict(l=0,r=0,t=0,b=30))
    if title is not None:
        fig.update_layout(annotations=[dict(xanchor='center',yanchor='bottom',x=0.5,y=1,showarrow=False,text=title,font_size=25,xref='paper',yref='paper')])
    config = {'displayModeBar': True, 'displaylogo': False, 'scrollZoom': True, 'responsive':True, 'toImageButtonOptions': {'format':'svg'}, 'modeBarButtonsToRemove': ['hoverCompareCartesian','toggleSpikelines']}
    
    if display and filename is None:
        fig.show(renderer=renderer, config=config)
    else:
        # full_html=True generates only a very minimal html - do not use
        div = plotly.io.to_html(fig, config=config, full_html=False)
        html = f'''<html>
<head><meta charset="utf-8" />
<title>Data plots – {title}</title>
</head>
<body>
{div}
</body>
</html>
'''
        if filename is None:
            return html
        
        html_file = open(filename,"w")
        html_file.write(html)
        html_file.close()
        
        if display:
            # Source code inspection shows this is used by plotly
            # so we can safely import it here
            import webbrowser
            webbrowser.open(filename)



def plot_trajectories(coords, index_map, eigenvalues, predictions = None, extrapolations = None, annotations = None, annoted_coords = None, annotations_legend = '', annotations_alignment=None, annotation_colors = None, mark_annotations = False, marker_size = 3, marker_colors=None, marker_colorscale = None, marker_symbols = None, title = None, color_by_values = None, color_by_scales = None, color_by_names = None, add_default_color_schemes = True, line_width = None, qc_extra = None, num_vars = 5, num_spectrum_vars = None, split_segments = False, axes_mode = 1, axes_colors = None, plot2D = False, N = None, discard_first_components=1, filename = None, display = True, renderer='browser', cam_eye = None, cam_up = None, main_plot_only = False):
    #series, predictions = None, extrapolations = None, timestamps = None, pred_timestamps = None, labels = None, title = None, qcflags = None, selected_series = -1, filename = None, display = True, renderer='browser'):
            
    # num_vars=5: Consider the first 5 inferred components (= coords dimensions)
    # => 10 combinations of 3 elements to plot as x,y,z components
    
    # GPU objects don't do well with plotly
    try:
        import cupy as cp
        # copy on CPU if not there already. no-op otherwise
        coords = cp.asnumpy(coords)
        index_map = cp.asnumpy(index_map)
        eigenvalues = cp.asnumpy(eigenvalues)

    except:
        # cannot import cupy - assume CPU
        pass
    
    # HACK : num_vars=-1 only displays the main plot, suitable for inclusion
    # in a presentation as iframe, with only 3 components
    #main_plot_only = False
    #if num_vars==-1:
    #    main_plot_only = True
    #    num_vars = 3
    if main_plot_only:
        num_vars = 3
        
    if num_spectrum_vars is None:
        num_spectrum_vars = num_vars
    
    num_vars = np.min([num_vars, coords.shape[1]-discard_first_components])
    num_spectrum_vars = np.min([num_spectrum_vars, coords.shape[1]-discard_first_components])
    
    if line_width is None:
        if plot2D:
            line_width = 2
        else:
            line_width = 4
    
    if N is None:
        # auto-infer from colorby or extrapolated components
        if extrapolations is not None:
            N = extrapolations.shape[0]
        elif color_by_values is not None:
            if isinstance(color_by_values[0],list) or isinstance(color_by_values[0],tuple):
                N = len(color_by_values[0][0])
            else:
                N = len(color_by_values[0])
        else:
            N = int(np.max(index_map) + 1)
    
    import plotly.graph_objects as go
    import plotly.offline
    import plotly.io
    import plotly.subplots
        
    fig = go.Figure()

    if not main_plot_only:
        fig.set_subplots(rows=3, cols=4, horizontal_spacing=0, column_widths=[0.10,0.05,0.75,0.1],
                     specs=[ [{'type':'xy','colspan':2},{},{'rowspan':3, 'type':('xy' if plot2D else 'scene')},{'rowspan':3}],
                             [{}, {}, None, None],
                             [{}, {'type':'heatmap'}, None, None]
                            ]
        )
    
    
    traces_xs = []
    traces_ys = []
    traces_zs = []
    traces_onoff = []
    uvw = []
    
    nseg = 1
    
    # indices of the state components
    for u in range(discard_first_components,num_vars+discard_first_components):
        for v in range(u+1,num_vars+discard_first_components):
            for w in range(v+1,(v+2 if plot2D else num_vars+discard_first_components)):
                # Label the component names from the first
                uvw += [ [u-discard_first_components+1,v-discard_first_components+1,w-discard_first_components+1] ]
                xs = []
                ys = []
                zs = []
                
                # states for the valid ranges (possibly refined)
                sc = np.full((N,3), np.nan)
                sc[index_map,:] = coords[:,[u,v,w]]
                
                if split_segments:
                    dimap = np.diff(index_map)
                    # number of separate segments
                    nseg = np.sum(dimap>1) + 1
                    if nseg>1:
                        split_pts = np.concatenate([np.where(dimap>1)[0], [-1]])
                        beg = 0
                        for si,s in enumerate(split_pts):
                            end = index_map[s]+1 if s>0 else N
                            xs += [ sc[beg:end,0] ]
                            ys += [ sc[beg:end,1] ]
                            zs += [ sc[beg:end,2] ]
                            beg = index_map[s+1]
                    else:
                        split_segments = False # only one contiguous segment
                
                if not split_segments:
                    xs += [sc[:,0]] ; ys += [sc[:,1]] ; zs += [sc[:,2]]
                
                nvalid = np.sum(~np.isnan(sc).any(axis=1))
                
                # states for the begin/end intermediate data
                if extrapolations is None:
                    xs += [ [np.nan] ] ; ys += [ [np.nan] ] ; zs += [ [np.nan] ]
                    ninter = 0
                    xs += [ [np.nan] ] ; ys += [ [np.nan] ] ; zs += [ [np.nan] ]
                    ngap = 0
                else:
                    if qc_extra is None:
                        qc_extra = np.full(N, 0)
                        qc_extra[index_map] = 2 # original data
                        qc_extra[:index_map[0]] = 1 # begin/end
                        qc_extra[index_map[-1]:] = 1
                    oridx = np.arange(N)[index_map]
                    # original valid segment, ends and starts
                    be = oridx[np.concatenate([np.diff(oridx)>1,[True]])]
                    bs = np.flip(oridx)[np.concatenate([np.diff(np.flip(oridx))<-1,[True]])]
                    # valid states according to the qc flags
                    validx = np.arange(N)[qc_extra>=1]
                    # add the very last state before predictions as an exception
                    validx = np.concatenate([validx, [N-1]])
                    # begin/end extrapolations
                    inter_states = np.setdiff1d(validx, oridx)
                    # join with the original segments
                    inter_states = np.union1d(inter_states, bs)
                    inter_states = np.union1d(inter_states, be)
                    sc = np.full((N,3), np.nan)
                    sc[inter_states,:] = extrapolations[inter_states,:][:,[u,v,w]]
                    xs += [sc[:,0]] ; ys += [sc[:,1]] ; zs += [sc[:,2]]
                    ninter = np.sum(~np.isnan(sc).any(axis=1))
                
                    # states for the gap filled regions
                    # begin end end of each valid segment
                    be = validx[np.concatenate([np.diff(validx)>1,[True]])]
                    bs = np.flip(validx)[np.concatenate([np.diff(np.flip(validx))<-1,[True]])]
                    invalidx = np.setdiff1d(np.arange(N), validx)
                    # join curves with the valid segments
                    invalidx = np.union1d(invalidx, bs)
                    invalidx = np.union1d(invalidx, be)
                    sc = np.full((N,3), np.nan)
                    sc[invalidx,:] = extrapolations[invalidx,:][:,[u,v,w]]
                    xs += [sc[:,0]] ; ys += [sc[:,1]] ; zs += [sc[:,2]]
                    ngap = np.sum(~np.isnan(sc).any(axis=1))
                    
                    
                # predictions
                if predictions is None:
                    xs += [ [np.nan] ] ; ys += [ [np.nan] ] ; zs += [ [np.nan] ]
                    npred = 0
                else:
                    xs += [ predictions[:,u] ]
                    ys += [ predictions[:,v] ]
                    zs += [ predictions[:,w] ]
                    npred = np.sum(~np.isnan(predictions).any(axis=1))

                # dates
                if annoted_coords is None or annotations is None:
                    xs += [ [np.nan] ] ; ys += [ [np.nan] ] ; zs += [ [np.nan] ]
                    nanno = 0
                    # markers
                    xs += [ [np.nan] ] ; ys += [ [np.nan] ] ; zs += [ [np.nan] ]
                else:
                    xs += [annoted_coords[:,u]]
                    ys += [annoted_coords[:,v]]
                    zs += [annoted_coords[:,w]]
                    nanno = np.sum(~np.isnan(annoted_coords).any(axis=1))
                
                    # date markers
                    if mark_annotations:
                        xs += [annoted_coords[:,u]]
                        ys += [annoted_coords[:,v]]
                        zs += [annoted_coords[:,w]]
                    else:
                        xs += [ [np.nan] ] ; ys += [ [np.nan] ] ; zs += [ [np.nan] ]
                
                # eigenspectrum
                xs += [np.arange(1,num_spectrum_vars+1)]
                ys += [eigenvalues[discard_first_components:num_spectrum_vars+discard_first_components]/eigenvalues[discard_first_components]*100]
                zs += [None]
                
                # heat map
                xs += [['']]
                ys += [None]
                zs += [np.arange(0,1,0.01)[:,None]]
                
                traces_xs += [xs]
                traces_ys += [ys]
                traces_zs += [zs]
                
                traces_onoff += [ [bool(nvalid>0)] * nseg + [ bool(ninter>0), bool(ngap>0), bool(npred>0), bool(nanno>0), False] ]
    
    sel = 0 # state components 1,2,3

    if plot2D is True:
        # UNIMPLEMENTED FEATURE in plotly : color by value is not available
        color_schemes = [ ['blue'] * (2+nseg) + ['black'] * 5 ]
        colormap_scale = [ [ 'Greys', 0, 0.01 ] ]
        
    else:
        color_schemes = []
        colormap_scale = []
        if color_by_names is None:
            color_by_names = []
        if color_by_values is not None:
            for i, v in enumerate(color_by_values):
                if isinstance(v, np.ndarray) and v.dtype=='<U7':
                    colval = v
                    colormap_scale += [ [ 'Greys', 0, 0.01 ] ]
                else:
                    if isinstance(v,list) or isinstance(v,tuple):
                        v, vmin, vmax = v[0], v[1], v[2]
                    else:
                        vmin = np.nanmin(v)
                        vmax = np.nanmax(v)
                    #color_by_names
                    scale = 'Rainbow' if color_by_scales is None else color_by_scales[i]
                    v_01 = (v-vmin)/(vmax-vmin)
                    v_01[np.isnan(v_01)] = 0.5 # should be out of index_map anyway
                    colval = plotly.colors.sample_colorscale(scale, v_01, colortype='tuple')
                    colormap_scale += [ [ scale, vmin, (vmax-vmin)/100 ] ]
                    
                color_schemes += [ 
                    [ colval ] * (2+nseg) + ['black'] * 5
                ]
    
            if not color_by_names:
                color_by_names = [ f'Color scheme {i+1}' for i in range(len(color_schemes)) ]

        # Default color schemes can be computed even 
        # in the absence of data values
        if add_default_color_schemes:
            # At least one discontinuity
            dimap = np.diff(index_map)
            if np.sum(dimap>1)>0:
                color_by_names += ["Continuous segments"]
                split_pts = np.concatenate([np.where(dimap>1)[0], [-1]])
                beg = 0
                ccol = ['#000000'] * N
                discrete_colors = plotly.colors.qualitative.G10 + plotly.colors.qualitative.Dark24
                cscheme_segs = []
                cdataidx_segs = []
                for si,s in enumerate(split_pts):
                    end = index_map[s]+1 if s>0 else N
                    thecolor = discrete_colors[si % len(discrete_colors)]
                    for i in range(beg,end):
                        ccol[i] = thecolor
                    cscheme_segs += [ thecolor ]
                    cdataidx_segs += [ plotly.colors.sample_colorscale('Rainbow', np.arange(N)/(N-1))[beg:end] ]
                    beg = index_map[s+1]
                if nseg>1:
                    color_schemes += [ 
                        cscheme_segs + [ ccol ] * 2 + ['black'] * 5
                    ]
                else:
                    color_schemes += [ 
                        [ ccol ] * (2+nseg) + ['black'] * 5
                    ]
                colormap_scale += [ [ discrete_colors, 0, len(discrete_colors)/100 ] ]
                    
            color_by_names += ["Data index"]
            if nseg>1:
                color_schemes += [ 
                    cdataidx_segs + [ plotly.colors.sample_colorscale('Rainbow', np.arange(N)/(N-1)) ] * 2 + ['black'] * 5
                ]
            else:
                color_schemes += [ 
                    [ plotly.colors.sample_colorscale('Rainbow', np.arange(N)/(N-1)) ] * (2+nseg) + ['black'] * 5
                ]
            colormap_scale += [ [ 'Rainbow', 0, N/100 ] ]
            
            color_by_names += ["No coloring"]
            color_schemes += [ ['black'] * 7 ]
            colormap_scale += [ [ 'Greys', 0, 0.01 ] ]

        if not color_schemes:
            color_by_names += ["No coloring"]
            color_schemes += [ ['black'] * 7 ]
            colormap_scale += [ [ 'Greys', 0, 0.01 ] ]

    # First provided color scheme, or the unique default
    colsel = 0
    
    if main_plot_only:
        row_col_kwargs = {}
    else:
        row_col_kwargs = {'row':1, 'col':3}
    
    for segi in range(nseg):
        fig.add_trace(go.Scattergl(
            x = traces_xs[sel][segi],
            y = traces_ys[sel][segi],
            mode = 'lines',
            line = dict(color=color_schemes[colsel][segi], width=line_width),
            hoverinfo='none',
            name = f'Segment {segi+1}' if nseg>1 else 'Data',
            showlegend = traces_onoff[sel][segi],
        ) if plot2D else go.Scatter3d(
            x = traces_xs[sel][segi],
            y = traces_ys[sel][segi],
            z = traces_zs[sel][segi],
            mode = 'lines+markers',
            marker = dict(size=2.75),
            line = dict(color=color_schemes[colsel][segi], width=line_width),
            hoverinfo='none',
            name = f'Segment {segi+1}' if nseg>1 else 'Data',
            showlegend = traces_onoff[sel][segi],
        ), **row_col_kwargs)
        
    fig.add_trace(go.Scattergl(
        x = traces_xs[sel][nseg],
        y = traces_ys[sel][nseg],
        mode = 'lines',
        line = dict(color=color_schemes[colsel][nseg], width=line_width),
        hoverinfo='none',
        name = 'Partial past/future',
        showlegend = traces_onoff[sel][nseg],
    ) if plot2D else go.Scatter3d(
        x = traces_xs[sel][nseg],
        y = traces_ys[sel][nseg],
        z = traces_zs[sel][nseg],
        mode = 'lines',
        line = dict(color=color_schemes[colsel][nseg], width=line_width),
        hoverinfo='none',
        name = 'Partial past/future',
        showlegend = traces_onoff[sel][nseg],
    ), **row_col_kwargs)
    fig.add_trace(go.Scattergl(
        x = traces_xs[sel][nseg+1],
        y = traces_ys[sel][nseg+1],
        mode = 'lines',
        line = dict(color=color_schemes[colsel][nseg+1], width=line_width),
        hoverinfo='none',
        name = 'Gap filling',
        showlegend = traces_onoff[sel][nseg+1],
    ) if plot2D else go.Scatter3d(
        x = traces_xs[sel][nseg+1],
        y = traces_ys[sel][nseg+1],
        z = traces_zs[sel][nseg+1],
        mode = 'lines',
        line = dict(color=color_schemes[colsel][nseg+1], width=line_width),
        hoverinfo='none',
        name = 'Gap filling',
        showlegend = traces_onoff[sel][nseg+1],
    ), **row_col_kwargs)

    fig.add_trace(go.Scattergl(
        x = traces_xs[sel][nseg+2],
        y = traces_ys[sel][nseg+2],
        mode = 'lines',
        line = dict(color='black', width=line_width+2),
        hoverinfo='none',
        name = 'Predictions',
        visible= True if traces_onoff[sel][nseg+2] else 'legendonly',
        showlegend = traces_onoff[sel][nseg+2],
    ) if plot2D else go.Scatter3d(
        x = traces_xs[sel][nseg+2],
        y = traces_ys[sel][nseg+2],
        z = traces_zs[sel][nseg+2],
        mode = 'lines',
        line = dict(color='black', width=line_width+2),
        hoverinfo='none',
        name = 'Predictions',
        visible= True if traces_onoff[sel][nseg+2] else 'legendonly',
        showlegend = traces_onoff[sel][nseg+2],
    ), **row_col_kwargs)
    
    fig.add_trace(go.Scattergl(
        x = traces_xs[sel][nseg+3],
        y = traces_ys[sel][nseg+3],
        mode = 'text',
        hoverinfo='none',
        text = annotations,
        textfont = dict(color = annotation_colors if annotation_colors is not None else 'black'),
        name = annotations_legend,
        showlegend = traces_onoff[sel][nseg+3],
    ) if plot2D else go.Scatter3d(
        x = traces_xs[sel][nseg+3],
        y = traces_ys[sel][nseg+3],
        z = traces_zs[sel][nseg+3],
        mode = 'text',
        hoverinfo='none',
        text = annotations,
        textfont = dict(color = annotation_colors if annotation_colors is not None else 'black'),
        textposition = annotations_alignment,
        name = annotations_legend,
        showlegend = traces_onoff[sel][nseg+3],
    ), **row_col_kwargs)

    if marker_colors is None:
        marker_colors = annotation_colors
        
    if marker_colors is None:
        marker_colors = "black"
        
    if marker_symbols is None:
        marker_symbols = 'circle'
        
    fig.add_trace(go.Scattergl(
        x = traces_xs[sel][nseg+4],
        y = traces_ys[sel][nseg+4],
        mode = 'markers',
        hoverinfo='none',
        marker = dict(color='black', size=3),
        showlegend = False,
    ) if plot2D else go.Scatter3d(
        x = traces_xs[sel][nseg+4],
        y = traces_ys[sel][nseg+4],
        z = traces_zs[sel][nseg+4],
        mode = 'markers',
        hoverinfo='none',
        #marker = dict(color='black', size=3),
        #marker = dict(color=np.linspace(0,1,len(traces_xs[sel][nseg+4])), colorscale="Rainbow", size=4),
        marker = dict(color=marker_colors, colorscale=marker_colorscale, size=marker_size, symbol=marker_symbols),
        showlegend = False,
    ), **row_col_kwargs)


    # Experimental and ugly, does not work in 2D
    # bit flags:
    # axes_mode & 1 : background grid on/off
    # axes_mode & 2 : lines along each axis on/off
    # Would need to redo all ticks, etc. Not really worth it.
    if axes_mode & 2:
        
        #axes_colors = ["magenta", "cyan", "grey"]
        if axes_colors is None:
            axes_colors = "black"
        
        if not (isinstance(axes_colors,list) or isinstance(axes_colors,tuple)):
            axes_colors = [ axes_colors ] * 3
        
        axc = []
        for c in axes_colors:
            axc += [ c if isinstance(c,list) or isinstance(c,tuple) else [ c, "lightgrey" ] ]
   
            
        fig.add_trace(go.Scatter3d(
            x = np.array([0,1]),
            y = np.array([0,0]),
            z = np.array([0,0]),
            mode = 'lines',
            line = dict(color=axc[0][0], width=line_width+2),
            hoverinfo='none',
            showlegend = False,
        ))
            
        fig.add_trace(go.Cone(
            x = np.array([1]),
            y = np.array([0]),
            z = np.array([0]),
            u = np.array([0.2]),
            v = np.array([0]),
            w = np.array([0]),
            colorscale = [axc[0][0], axc[0][0]],
            hoverinfo='none',
            showlegend = False,
            showscale = False,
        ))
            
        fig.add_trace(go.Scatter3d(
            x = np.array([0,0]),
            y = np.array([0,1]),
            z = np.array([0,0]),
            mode = 'lines',
            line = dict(color=axc[1][0], width=line_width+2),
            hoverinfo='none',
            showlegend = False,
        ))
            
        fig.add_trace(go.Cone(
            x = np.array([0]),
            y = np.array([eigenvalues[2]/eigenvalues[1]]),
            z = np.array([0]),
            u = np.array([0]),
            v = np.array([0.2]),
            w = np.array([0]),
            colorscale = [axc[1][1], axc[1][1]],
            hoverinfo='none',
            showlegend = False,
            showscale = False,
        ))
            
        fig.add_trace(go.Cone(
            x = np.array([0]),
            y = np.array([1]),
            z = np.array([0]),
            u = np.array([0]),
            v = np.array([0.2]),
            w = np.array([0]),
            colorscale = [axc[1][0], axc[1][0]],
            hoverinfo='none',
            showlegend = False,
            showscale = False,
        ))
            
        fig.add_trace(go.Scatter3d(
            x = np.array([0,0]),
            y = np.array([0,0]),
            z = np.array([0,1]),
            mode = 'lines',
            line = dict(color=axc[2][0], width=line_width+2),
            hoverinfo='none',
            showlegend = False,
        ))
            
        fig.add_trace(go.Cone(
            x = np.array([0]),
            y = np.array([0]),
            z = np.array([eigenvalues[3]/eigenvalues[1]]),
            u = np.array([0]),
            v = np.array([0]),
            w = np.array([0.2]),
            colorscale = [axc[2][1], axc[2][1]],
            #colorscale = ['lightgrey', 'lightgrey'],
            #colorscale = [axes_colors[2], axes_colors[2]],
            hoverinfo='none',
            showlegend = False,
            showscale = False,
        ))
            
        fig.add_trace(go.Cone(
            x = np.array([0]),
            y = np.array([0]),
            z = np.array([1]),
            u = np.array([0]),
            v = np.array([0]),
            w = np.array([0.2]),
            colorscale = [axc[2][0], axc[2][0]],
            hoverinfo='none',
            name = 'X',
            showlegend = False,
            showscale = False,
        ))


        
    # spectrum & heat map
    if not main_plot_only:
        """
        fig.add_trace(go.Scattergl(
            x = traces_xs[sel][nseg+5],
            y = traces_ys[sel][nseg+5],
            mode = 'lines',
            line = dict(color='black'),
            showlegend=False
        ), row=1, col=1)
        """
        fig.add_trace(go.Bar(
            x = traces_xs[sel][nseg+5],
            y = traces_ys[sel][nseg+5],
            marker_color = '#4488aa',
            showlegend=False
        ), row=1, col=1)
        
        fig.add_trace(go.Heatmap(
            z=np.arange(0,1,0.01)[:,None],
            x=[''],
            y0=colormap_scale[0][1],
            dy=colormap_scale[0][2],
            hoverongaps = False,
            showscale = False,
            colorscale = colormap_scale[0][0]
        ), row=3, col=2)


    updatemenus=[dict(
        active=sel,
        buttons=[dict(label=f"Components {uvw[i][0]},{uvw[i][1]}",
                      method="update",
                      args=[{"x":traces_xs[i],
                             "y":traces_ys[i],
                             "showlegend":traces_onoff[i]
                      }],
                 ) for i in range(len(traces_xs))
        ] if plot2D else [dict(label=f"Components {uvw[i][0]},{uvw[i][1]},{uvw[i][2]}",
                      method="update",
                      args=[{"x":traces_xs[i],
                             "y":traces_ys[i],
                             "z":traces_zs[i],
                             "showlegend":traces_onoff[i]
                      }],
                 ) for i in range(len(traces_xs))
        ],
        showactive=True,
        xanchor='left',
        yanchor='bottom',
        x=0,
        y=1,
        pad_b=5,
    )]
    
    annotations = []
    
    if len(color_schemes)>0 and not plot2D:
        updatemenus += [dict(
                active=colsel,
                buttons=[dict(label=f"{color_by_names[i]}",
                              method="update",
                              args=[{"line.color":color_schemes[i],"colorscale":colormap_scale[i][0],"y0":colormap_scale[i][1],"dy":colormap_scale[i][2]
                              }],
                         ) for i in range(len(color_schemes))
                ],
                showactive=True,
                xanchor='left', yanchor='top',x=0,y=0.5,
            )   
        ]
                       
        if not main_plot_only:
            annotations += [dict( xanchor='left', yanchor='bottom', x=0, y=0.5, showarrow=False, text="Color by", font_size=17, xref='paper',yref='paper')]
            

    if not main_plot_only:
        fig.update_layout(updatemenus=updatemenus)
        fig.update_xaxes(title_text="Spectrum (% of first)", row=1, col=1)
    
    if not plot2D:
        nospikes = dict(showspikes=False)
        if not axes_mode & 1:
            nospikes["visible"] = False
        fig.update_layout(scene_xaxis = nospikes)
        fig.update_layout(scene_yaxis = nospikes)
        fig.update_layout(scene_zaxis = nospikes)
        fig.update_layout(scene_aspectratio=dict(x=1, y=1, z=1))
        camera = dict(
            up=dict(x=0, y=0, z=1) if cam_up is None else dict(x=cam_up[0], y=cam_up[1], z=cam_up[2]),
            eye=dict(x=0.9, y=0.9, z=0.9) if cam_eye is None else dict(x=cam_eye[0], y=cam_eye[1], z=cam_eye[2]),
            center=dict(x=0, y=0, z=0),
        )
        fig.update_layout(scene_camera=camera)
        fig.update_layout(scene_dragmode='orbit')

    if main_plot_only:
        fig.update_layout(showlegend=False)
    else:
        fig.update_layout(legend=dict(yanchor='top',xanchor='right',y=0.99,x=0.99,title_text='Click to show/hide:'))
        fig.update_layout(showlegend=True)
    fig.update_layout(hovermode=False)
    fig.update_layout(margin = dict(l=0,r=0,t=0,b=0))
    fig.update_layout(clickmode='select')

    if title is not None:
        annotations += [dict(xanchor='center', yanchor='bottom', x=0.5, y=1, showarrow=False, text=title, font_size=25, xref='paper', yref='paper')]
    
    if len(annotations)>0:
        fig.update_layout(annotations=annotations)
    
    config = {'displayModeBar': True, 'displaylogo': False, 'scrollZoom': True, 'responsive':True, 'modeBarButtonsToRemove': ['hoverCompareCartesian','toggleSpikelines','toImage'], 'modeBarButtonsToAdd':['toImage']}
    
    
    if display and filename is None:
        fig.show(renderer=renderer, config=config)
    else:
        # full_html=True generates only a very minimal html - do not use
        div = plotly.io.to_html(fig, config=config, full_html=False)
        html = f'''<html>
<head><meta charset="utf-8" />
<title>{title} - Causal diffusion components explorer</title>
</head>
<body>
{div}
</body>
</html>
'''
        if filename is None:
            return html
        
        html_file = open(filename,"w", encoding="utf-8")
        html_file.write(html)
        html_file.close()
        
        if display:
            import webbrowser
            webbrowser.open(filename, new=1)









