#!/usr/bin/env python3

"""
Source code from the Continuous Causal States method,
described in the paper :
  Discovering Causal Structure with Reproducing-Kernel Hilbert Space ε-Machines
  by Nicolas Brodu and James P. Crutchfield

This code is free software, according to the MIT license
Author & Contact: nicolas.brodu@inria.fr
"""

from . import npda_helper

import numpy as np

def spectral_basis(Gs, num_basis=None, scaled = False, alpha=1, use_dask = None, overwrite=False):
    """
    Spectral decomposition of the similarity matrix, *after normalization*.
    This uses a variant of the diffusion map algorithm, applied directly to the provided similarity matrix, which replaces the "diffusion" step of the diffusion maps.

    Parameters
    ----------
    Gs : Array of size (N, N)
        A precomputed similarity matrix.
        
    num_basis : number, optional
        Specifies the number of components to retain:
            
        - If num_basis is None (default), then N//2 eigenvalues are retained.
        - If num_basis <= 0, all components are retained
        - If num_basis >=1, this specifies directly the number of components
        - If 0 < num_basis < 1, this specifies a threshold below which eigenvalues are discarded.
        
        The default of N//2 is somewhat arbitrary and you should check what value is adequate for your problem. However, the hope is that the exact value does not matter much: preliminary tests on diverse data sets suggest there is a large plateau with acceptable performances (and then, a decrease as num_basis increases). Note that using all eigenvalues is generally a bad idea: the (necessary) eps parameter used in embed_states introduces a minimal and spurious eigenvalue. Due to the MMD test, the distance between states is also inacurate to some extent (depending on N). And the shift operator definition implicitly uses a matrix inverse, which is badly conditionned when spurious and meaningless eigenvalues are retained. Also, the kernel scale used for building Gx and Gy matters, since it sets the similarity between points, states, hence the sensitivity to small details or to large fluctuations, hence ultimately the eigenspectrum of Gs. One possiblity is to use cross-validation, for example using the accuracy of reconstructing the last nfuture values of the series. 
        
    alpha : float, optional
        Normalization exponent for the diffusion-map like algorithm. 
        
        The default of 1 allows the inference of the geometry, with minimal interference from the density.
        
    use_dask : bool, optional
        Whether to distribute the computations with Dask:
        
        - None (default): Use the package flag (itself defaulting to False)
        - False or 0: Do not use Dask. Numpy and scipy have better linear algebra support than Dask, so it may better to use higher-level parallelism and perform DMap_K computations as a single Dask task, even if Dask is used for other functions
        - True or 1: Use Dask for this function. If the data would not fit in memory on one machine, or if the numpy/scipy are not parallelized and multiple cores are available, then it may be advantageous to activate Dask internally in this function to parallelize computations.
        
    overwrite : Bool, optional
        Whether to use the memory in Gs for the computations. This has no effect when using Dask. False by default.

    Returns
    -------
    eigenvalues : array of size M
        The eigenvalues of the spectral decomposition. M is the number of components retained, see the num_basis parameter. The first eigenvalue is always 1 due to the normalization. All other eigenvalues should be greater than 0, unless Gs is badly built (in that case, check the data scale parameter in series_Gxy).
    
    basis : array of size (N, M)
        The M basis vectors, each of size N, represented as columns. These are also the left eigenvectors of the spectral decomposition. The first vector corresponds to a density and is normalized to sum to 1. The others should all sum to 0. The basis vectors all have the same norm.
        
    coords : array of size (N, M)
        The N coordinates, one for each entry in Gs. Each coordinate is a vector or of dimension M, matching the M basis vectors. The first coordinate is always the constant 1 and can be discarded for purposes such as distance comparisons (knn queries), manifold reconstruction, visualization, etc. The coordinates are the right eigenvectors of the spectral decomposition, scaled to be bi-orthogonal with the basis.
    
    Notes
    -----
    
    The pre-normalization step makes this *not* equivalent to a classical eigen or singular decomposition. Distances using the resulting coordinates would correspond to "diffusion distances", assuming Gs is a "diffusion matrix" built like the "diffusion map" algorithm. Although, since we use a completely different building  scheme, this is not a diffusion at all.
    
    The choice is to provide coordinates scaled with a constant of 1 in the dimension of the eigenvalue 1. This provides consistent values for coordinates along each dimension, with a range close to 1. This also greatly simplifies the computations of the shift operator and avoids division by small values, while being mathematically equivalent in the prediction results. One should however remember that the coordinates should be scaled by the eigenvalues for their relative influence in distance-based algorithms, such as finding nearest neighbors. The scaled distance has some interpretation as explained in the "diffusion map litterature".
    
    The use of a precision for selecting the number of eigenbasis components is not implemented in Dask mode, since all Dask does is create an operation graph and this would trigger a computation.
    
    """
    
    if use_dask is None:
        use_dask = npda_helper.use_dask()

    npda = npda_helper.import_dask_or_numpy(use_dask)
    
    N = Gs.shape[0]
    
    eigen_cutoff = -1
    
    if num_basis is None:
        num_basis = N // 2
    elif num_basis<=0:
        num_basis = N
    elif num_basis>=1:
        num_basis = int(num_basis)
    else:
        eigen_cutoff = num_basis
        num_basis = N
    
    alpha = np.clip(alpha, 0, 1)

    
    # The Dask part below follows the route of the diffusion map
    # papers which seek the eigenvalues/vectors of the diffusion matrix
    # D**-1 @ M, with M = mat and D is a diagonal matrix with entries q
    # But this matrix is not symmetric, so these papers (and code below)
    # build a symmetric matrix with the same eigenvalues and related
    # eigenvectors, and then recover eigenvectors later. 
    # Then, I normalize so the left eigenvector is a density and right
    # eigenvector has entries all ones.
    # This is fine and allows to use a classic eigenvalue problem  
    # A x = l x  with A the transformed matrix below
    # Now, it is _also_ possible to use the generalized
    # eigenvalue problem M x = l D x directly, when supported
    # Since the matrix D is diagonal, this is equivalent to M x = D (l x) 
    # and also to D**-1 @ M x = l x, which is the initial problem we seek.
    
    # Numpy / scipy implementation using a generalized eigenvalue solver
    if use_dask == False:
        try:
            # Use cupy if Gs is on the GPU
            import cupy as cp
            xp = cp.get_array_module(Gs)
        except:
            xp = np
    
        if overwrite:
            mat = Gs
        else:
            mat = Gs.copy() # no-op for dask

        if alpha>0:
            q = xp.sum(mat, axis=1)
            if alpha!=1:
                q = xp.power(q, alpha)
            q = 1. / q
            # This is equivalent to np.diag(qinv) @ mat @ np.diag(qinv),
            # but using broadcasting and reusing memory
            mat *= q[:,None]
            mat *= q[None,:]
        
        q = xp.sum(mat, axis=1)

        if xp is np:
            import scipy.linalg
            try:
                # subset_by_index is not supported by all scipy implementations
                # driver 'gv' is more robust and uses less memory. May be slower
                eigval, eigvec = scipy.linalg.eigh(mat, np.diag(q), overwrite_a=True, subset_by_index = [N-num_basis, N-1]) #, driver='gv')
            except:
                eigval, eigvec = scipy.linalg.eigh(mat, np.diag(q), overwrite_a=True)
                if num_basis<N:
                    # copies avoid to keep the full matrix referenced
                    eigval = eigval[N-num_basis:N].copy()
                    eigvec = eigvec[:, N-num_basis:N].copy()

            # sort from largest to lowest
            eigval = eigval[::-1]
            eigvec = eigvec[:, ::-1]
            
            if eigen_cutoff >= 0:
                num_basis = np.sum(eigval>=eigen_cutoff)
                # This was for preserving the trace proportion
                #cs = np.cumsum(eigval)
                #num_basis = np.sum( cs / cs[-1] < 1 - eigen_cutoff) + 1
            
            if num_basis<eigval.shape[0]:
                # copies avoid to keep the full matrix referenced
                eigval = eigval[:num_basis].copy()
                eigvec = eigvec[:, :num_basis].copy()

            # Normalization so that eigvec_r[:,0] entries are all 1
            # and that eigvec_l[:,0] matches a density
            eigvec_r = eigvec / eigvec[0,0]
            eigvec_l = eigvec * (q * eigvec[0,0])[:, None]

        # cuda implementation does not have generalized eigenvalue decomposition
        else:
            Dinvsqrt = 1. / xp.sqrt(q)
            mat *= Dinvsqrt[:,None]
            mat *= Dinvsqrt[None,:]
            eigval, eigvec_sym = xp.linalg.eigh(mat)
            # sort from largest to lowest
            eigval = eigval[::-1]
            eigvec_sym = eigvec_sym[:, ::-1]
            if eigen_cutoff >= 0:
                num_basis = xp.sum(eigval>=eigen_cutoff)
            if num_basis<eigval.shape[0]:
                eigval = eigval[:num_basis].copy()
            
            eigvec_r = eigvec_sym[:, :num_basis] * Dinvsqrt[:,None]
            eigvec_l = eigvec_sym[:, :num_basis] * (1/Dinvsqrt)[:,None]
            
            eigvec_l /= npda.sum(eigvec_l[:,0])
            normdiag = xp.diagonal(eigvec_l.T @ eigvec_r)
            eigvec_r /= normdiag[None,:]
        
        
        # If we want symmetric eigenvectors, but still wish to normalize,
        # for example for use with SEC code, then another possibility
        # is to note that, with u = eigvec:
        # (u.T / u[0,0]) @ (np.diag(D) * u[0,0]**2) @ (u / u[0,0])
        # Hence, we can use eigvec_r in place of u
        # and replace D by D*u[0,0]**2 and 
        
        if scaled: return eigval, eigvec_l, eigvec_r * eigval
        return eigval, eigvec_l, eigvec_r

    # Dask implemention. Not even a simple eigenvalue solver is available,
    # so a workaround is found with Cholesky and SVD
    else:
        
      import warnings
      with warnings.catch_warnings():
        # Avoid the Dask warning that many chunks are produced
        # rechunk below before using this variable
        warnings.simplefilter("ignore")
        
        if eigen_cutoff != -1:
            raise NotImplementedError("Specifying an eigenvalue cutoff would force Dask computations")
        
        if alpha>0:
            q = npda.sum(mat, axis=1)
            if alpha!=1:
                q = npda.power(q, alpha)
            q = 1. / q
            #print("q: ", q.chunks)
            # This is equivalent to np.diag(qinv) @ mat @ np.diag(qinv),
            # but using broadcasting and reusing memory
            #print("mat0: ", mat.chunks)
            mat *= q[:,None]
            #print("mat1: ", mat.chunks)
            mat *= q.rechunk((-1,))[None,:]
            #print("mat2: ", mat.chunks)
        
        q = npda.sum(mat, axis=1)
        #print("q: ", q.chunks)
        
        Dinvsqrt = 1. / npda.sqrt(q) # need to keep q as return value
        
        # See above
        mat *= Dinvsqrt[:,None]
        #print("mat3: ", mat.chunks)
        mat *= Dinvsqrt.rechunk((-1,))[None,:]
        #print("mat4: ", mat.chunks)
        
        # subset_by_index not defined on old scipy implementations (or Intel's)
        #eigval, eigvec_sym = scipy.linalg.eigh(mat, overwrite_a=True)#, subset_by_index = [N-num_basis, N-1])
    
        # dask does not have an eigh implementation :-(
        # eigval, eigvec_sym = scipy.linalg.eigh(mat)
        # the following returns the same, and in decreasing magnitude order
        #_, r = npda.linalg.qr(mat)
        r = npda.linalg.qr(mat)[-1]
        #print("r: ", mat.chunks)
        # This works, but returns a single-chunk revt
        #_, eigval, revt = npda.linalg.svd(r)
        # Re-chunk before so the result is also chunked, better fits in memory
        eigval, revt = npda.linalg.svd(r.rechunk((-1,npda_helper.chunk_size())))[1:3]
        eigvec_sym = revt.T
        #print("eigval: ", eigval.chunks)
        #print("eigvec_sym: ", eigvec_sym.chunks)
        
        eigvec_r = eigvec_sym * Dinvsqrt[:,None]
        eigvec_l = eigvec_sym * (1/Dinvsqrt)[:,None]
        #print("eigvec_r: ", eigvec_r.chunks)
        #print("eigvec_l: ", eigvec_l.chunks)
        
        # renormalizing the vectors, so that eigvec_l[:,0] is the density
        # and so the first right eigenvector components is all 1
        #eigvec_r = eigvec_r / np.sqrt(np.sum(eigvec_r * eigvec_r, 0) / N)
        #eigvec_l_d = np.diagonal(eigvec_l.T @ eigvec_r)
        # the first left eigenvector is now exactly the density, up to a sign
        #eigvec_l = eigvec_l / eigvec_l_d[None,:]
        
        # It is more stable and more direct to do it the other way around
        # This rescales all left eigenvectors by the same value, so that the first
        # is now also the density
        eigvec_l /= npda.sum(eigvec_l[:,0])
        # get the diagonal for the bi-orthogonality
        normdiag = npda.diagonal(eigvec_l.T @ eigvec_r)
        # restore the bi-orthogonality
        eigvec_r /= normdiag[None,:]
        # At this stage, eigvec_r[:,0] should be all ones
        
        # This is the density, but should also be the left eigenvector
        # associated to eigenvalue 1
        #q /= npda.sum(q)
        
        # ensure the sign is correct
        #if npda.dot(eigvec_l[:,0], q) < 0:
        #    eigvec_l = -eigvec_l
        #    eigvec_r = -eigvec_r # ensures bi-orthogonality is respected
        # Problem: data is not computed computed at that point, the values
        #          are not yet available and only the task graph is built
        # a compute() could be forced... but may be useless
        # as the sign was just a convenience to match q and eigvec_l[:,0]
        # another possibility is to force-copy q into eigvec_l[:,0]
        # or do nothing, they should match anyway
        
        # Now we can get the embedding coordinates
        # but this can also be done by the caller
        # -- better to keep all the information
        # psi = eigvec_r * eigval[None,:]
        
        #print("eigval: ", eigval.chunks)
        #print("eigvec_l: ", eigvec_l.chunks)
        #print("eigvec_r: ", eigvec_r.chunks)
        
        # Now normalize the coords with the eigenvalues
        if scaled: return eigval, eigvec_l, eigvec_r * eigval
        return eigval, eigvec_l, eigvec_r
        




