#!/usr/bin/env python3

import numpy as np
import scipy.linalg
from . import kernels

def new_coords(Ks, Gs, coords, method='nnls'):
    """
    Compute a state distribution

    Parameters
    ----------
    Ks : array of size (N, L)
        These are the kernel similarity vectors between a new state estimate and reference states. Such vectors can be computed using the series_newKx for time series data and embedding that result with embed_Kx
        L is the number of such vectors, one per column
        
    Gs : array of size (N, N)
        The Gram matrix between known states
        
    coords : array of size (N, M)
        diffusion map coordinates for each sample. These are returned by the spectral_basis function.
        
    method : string, optional
        What method to use for computing the state distribution. This parameter is subject to change. Values are
            + 'nnls': (default) uses non-negative least squares to ensure that the new state estimate remains within the boundaries of existing states. This is similar to kernel moment matching, see the paper
            + 'unbounded': unbounded estimate, equivalent to a Nyström extension, which is then normalized into a pseudo-distribution.

    Returns
    -------
    state_dist : array of size (M, L)
        The state distribution, represented as coefficients in the eigenbasis. The first entry of the state distribution is always be 1. In order to retreive a proper probability distribution, you can do q = basis @ state_dist. Note that, after evolution by the shift operator, the expression of the causal states as linear combinations of RKHS basis elements may induce that some entries of q are negative. q is both the distribution, and its expression as a linear combination on the RKHS samples.
    """
    
    if method=='unbounded':
        omega = np.linalg.solve(Gs, Ks)
    else: 
        # if method=='convex':
        # Note that the Non-Negative constraint is enough in this case,
        # Since the solution sums to 1 with the diagonal entries in Gs
        from . import nnls_impl
        omega = nnls_impl.nnls(Gs, Ks)
        
    q0 = omega
    q0 /= np.sum(q0,axis=0)
    # q0 is specified in the original RKHS representation, for each sample. 
    # Turn it into an eigen space representation (scaled if coords are scaled)
    nc = q0.T @ coords
    # Since q0 is normalized, and since all the first components of the coordinates are 1, the first entry of the state distribution should also be 1
    if not np.all(np.isclose(nc[:,0], 1)):
        import warnings
        warnings.warn("The resulting state distribution must be normalized to have 1 as its first entry, but this is not the case. Did you pass the correct coordinates?")
    # ensure there are no numerical roundoff
    nc /= nc[:,:1]
    return nc


def shift_operator(scaled_coords, index_map):
    """
    Compute a shift operator for the dynamical system, assuming coords are the time series of causal states
    Handles NaN values and split / multi series with index maps
    
    Arguments:
        scaled_coords: the coordinates of the causal states, scaled by the eigenvalues
        index_map: see series_Gxy
        
    Returns:
        The forward operator for state coordinates, represented in the eigenbasis.
    """
    
    try:
        # Use cupy if coords are on GPU
        # Use cupy if Gx or Gy is on the GPU, use GPU
        import cupy as cp
        if cp.get_array_module(scaled_coords) is cp or cp.get_array_module(index_map) is cp:
            use_GPU=1
        else:
            use_GPU=0
            
        if use_GPU==0:
            xp = np
            # copy on CPU if not there already
            scaled_coords = cp.asnumpy(scaled_coords)
            index_map = cp.asnumpy(index_map)
        else:
            xp = cp
            # copy on GPU if not there already
            scaled_coords = cp.asarray(scaled_coords)
            index_map = cp.asarray(index_map)
    except:
        xp = np
        if use_GPU==1:
            print("Warning: GPU is not available. Using CPU.")
    
    # TODO: generator, meaning the log in eigen decomposition form
    # This would allow to compute powers more efficiently

    # Shift operator, expressing the new coordinates as a combination of the old ones. This definition is similar to that of the kernel Koopman operator in RKHS, but done here in diffusion map coordinates space
    
    if index_map is None:
        valid_prev = xp.arange(scaled_coords.shape[0]-1)
        indices_no_next = xp.arange(scaled_coords.shape[0]-1,scaled_coords.shape[0])
    else:
        # abs is a hack to allow the use of the same function 
        # with np.flip'd arrays, in order to build reverse operators
        valid_prev = xp.arange(scaled_coords.shape[0]-1)[xp.abs(xp.diff(index_map))==1]
        # also known as gaplocs, plus the latest state
        indices_no_next = xp.setdiff1d(xp.arange(scaled_coords.shape[0]), valid_prev)

    # To replace in formulas:
    # [:-1] => valid_prev
    # [1:] => valid_next
    # [-1:] => indices_no_next
    valid_next = valid_prev + 1
    
    shift_op = xp.zeros((scaled_coords.shape[1], scaled_coords.shape[1]))
    shift_op[0,0] = 1

    # we want shift_op @ coords[valid_prev,:].T = coords[valid_next,:].T
    # and some generalization outside valid_prev
    # below are raw/naive options
    # This is the fastest
    #shift_op[1:,:] = coords[valid_next,1:].T @ scipy.linalg.pinv2(coords[valid_prev,:].T)
    # These two are mathematically equivalent, but slower.
    # The second of these two may require less memory, so be kept as an option
    # coords[1:,1:].T @ np.linalg.pinv(coords[:-1,:].T)
    # scipy.linalg.lstsq(coords[:-1,:], coords[1:,1:])[0].T


    # The idea now is to express the coordinates with no valid next values
    # as a combination of known coordinates, so that the shift operator
    # expresses coordinates only using the known / good examples

    # Using inverses yields the worst results
    #X = scipy.linalg.pinv2(coords[valid_prev,:].T) @ coords[indices_no_next,:].T
    
    # Using nnls instead works better, see the first paper
    from . import nnls_impl
    # weights for each entry with no successor
    X = nnls_impl.nnls(scaled_coords[valid_prev,:].T, scaled_coords[indices_no_next,:].T)
    X /= xp.sum(X, axis=0)[None,:] # should already be close to 1
    
    # Using distances and softmax is too sensitive to spurious neighbors
    #from . import kernels
    #import scipy.spatial.distance
    #dists = scipy.spatial.distance.cdist(coords[valid_prev,:]*eigenvalues, coords[indices_no_next,:]*eigenvalues)
    #X = kernels.dist_to_weights(dists, knn=5)
    
    # NOTE: off-diagonal => T.shape = (coords.shape[0], coords.shape[0]) 
    T = xp.diag(np.ones(scaled_coords.shape[0]-1), -1)
    
    # put that back into T. Process column by column, due to indexing scheme
    # that would make the matrix read-only
    for ic, c in enumerate(indices_no_next):
        T[:, c] = 0
        T[valid_next, c] = X[:,ic]

    shift_op[1:,1:] = xp.linalg.lstsq(scaled_coords[:,1:], T.T @ scaled_coords[:,1:], rcond=None)[0].T    
    return shift_op



def immediate_future(data, indices):
    """
    Equivalent to lambda d,i: d[i+1,:] . This function is used as a default argument in expectation_operator. See the documentation there.
    """
    return data[indices+1, :]

def expectation_operator(scaled_coords, index_map, targets, function = immediate_future):
    """
    Builds the expectation operator, mapping a state distribution expressed in the eigenbasis, into numerical values, expressed in the original series units.

    Parameters
    ----------
        
    scaled_coords : array
        the coordinates of the causal states, scaled by the eigenvalues
        
    index_map : int array
        This is the index_map returned by the series_Gxy function. It indicates the index in the series for the (past, future) pair matching each coordinate entry. That index is that of the present, the last value in the past series. If the series consist of several discontiguous blocks, the index refers to the valid entries in the series that would be made by concatenating these blocks (that series would have invalid entries at each time discontinuity, at nan values, etc). The targets parameter may have its own validity pattern, and one that differs for each heterogenous data source - it may be that a computed state, has no matching target for one data source, but a valid target for another. Expectation operators are computed using all valid data, for each heterogenous data source.
        
    targets : array or list of array
        Target values that we wish to predict from causal states, to build an operator from. Each array must have the same temporal structure (number of samples, discontinuous blocks, etc) as the data series that were used to build the causal states - one target is needed for each original series measurement. Some targets may be NaN, in which case they will be ignored for building the operator. These NaN patterns may differ from these implied by the causal state index_map. If a list is provided as the targets argument, these are observations from multiple (possibly heterogenous) data source, and possibly different sources from those used for building the causal states. Then, a list of expectation operators is returned, one of each data source.
        
    function : callable, or list of callable, optional
        This is the function of the targets, which expectation is computed by the operator. Different functions can be provided for each heterogenous data source by providing a list of callables. Each function takes as parameter a data array (which is in fact the stacked targets) and an index array (the index_map), and should return the function computed on the data at each specified index. The default is to use the immediate_future function (defined in the concaust scope), which is equivalent to lambda d,i: d[i+1,:]. Due to the way the index_map is built, the function is applied only on target data values that match valid causal states. Thus, it is guaranteed that data in the range i-npast+1:i+nfuture is valid: i is the "present", the last entry of the past sequence in the consecutive (past, future) pair. Assuming nfuture>=1, then the next value at i+1 always exists and it is the first entry in the future sequence.


    Returns
    -------
    expect_ops : array or list of arrays
        This is the expectation operator for the function (or functions) that compute a value from the current state. The expectation is implicitly performed by the reproducing property and using mean maps for the probability distributions. If a list of data sources was provided as the series argument, return one search operator for each data source.
        
    Notes
    ----
    
    - As suggested by the 'targets' parameter name, these need not be the same as the 'series' argument of the series_Gxy that were used to build the causal states. You may very well construct the causal states from some data sources, and try to predict only a subset of these sources, or maybe other observables entirely that you assume depend on the causal states. The only restriction is that the series and the targets variable share the same temporal pattern: the same number of samples in the same number of continuous data blocks. Use NaN if some target values are not available for all sample times.
    
    - Theoretically, any function of the causal state is useable, so long as it can be estimated for each observed (past, future) data pair in the target measurements. The function could thus very well take non-numeric data as argument, but, currently, it should only return numeric values (scalar or vectorial).
    
    - TODO: Given the above, it would be possible to extend other machine learning algorithms taking as input the (past,future) pairs and producing some value of interest. The machine learning trained instance could be fed as the function parameter, then it would benefit from the causal state machinery.
    
    """
    
    if index_map is None:
        raise ValueError("You must provide a valid index map")
    
    targets_list = targets if isinstance(targets,list) or isinstance(targets,tuple) else [ targets ]
    f_list = function if isinstance(function,list) or isinstance(function,tuple) else [ function ] * len(targets_list)
    
    eoplist = []
    
    sci = scipy.linalg.pinv(scaled_coords.T)
    
    for tar, fun in zip(targets_list,f_list):
        
        if isinstance(tar,list) or isinstance(tar,tuple):
            tar = np.concatenate(tar, axis=0)
        
        if tar.ndim==1:
            tar = tar[:,None]
            
        fvalues = fun(tar, index_map)
        
        valid_f = ~np.isnan(fvalues).any(axis=1)

        eoplist += [ fvalues[valid_f,:].T @ sci[valid_f,:] ]
        
    return eoplist if isinstance(targets,list) or isinstance(targets,tuple) else eoplist[0]



def extrapolate_contiguous(scaled_coords, index_map, shift_op, expect_ops, targets = None, qc_targets = None, reverse_op = None, verbose = 0, extent = 1e-6, extent_nan = 0.1, adjust_ori = True, N = None):
    """
    Extrapolate the states given by scaled_coords, so that each series entry has 
    a matching state. 
    
    Three kind of extrapolations are performed:
    - At the start of the series, before npast values are observed. These are 
      extrapolated backward from the first state
    - At the end of the series, to fill the missing nfuture-1 values before
      reaching the "present" state. These are extrapolated forward from the 
      last known state
    - In the middle of the series, when there are nan gaps. These are
      extrapolated in both directions and a linear combination is performed
      for these two extrapolations with weights matching the distance to
      each end

    Parameters
    ----------
    scaled_coords : TYPE
        DESCRIPTION.
    index_map : TYPE
        DESCRIPTION.
    shift_op : TYPE
        DESCRIPTION.
    series : TYPE
        DESCRIPTION.

    verbose: <=0: off, 1: detailed progress, 2: very detailed (solver info)

    Returns
    -------
    ecoords, qcmap if targets is None
    extrapolated_targets, ecoords, qcmap if targets is not None
    
    Quality control for the current state based on the NEXT target value:
      -3: no target provided
      -2: issue with numerical solver, pure interpolation/prediction
      -1: no valid next data, but this is expected (last point in the series)
      ratio 0..1 : num_non-nans / num_targets
      => 0 = all nans, interpolation + fit towards smoothed trajectory
      => 1 = in the begin/end areas, with valid data to fit.
      2: original state. Valid data, not gap filled, no begin/end
      3: adjusted state using valid data, not gap filled, no begin/end
    So:
      qc <= 0 : no data at all, or no next data (very last point)
      qc < 1 : some nans
      qc >= 1 : correct data or extrapolation
      
    Note: the very last entry of the series has qc=-1 because no future data 
    are available to fit it, but this is still the best we can do so it is
    suitable for predictions. It is set to a value <0 in order to rebuild
    an index map suitable for feeding to the expectation operator function:
    
    ex, exstates, qcmap = extrapolate_contiguous(...)
    index_map2 = np.arange(len(qcmap))[qcmap>=1]
    coords2 = exstates[qcmap>=1, :]
    shift_op2 = shift_operator(coords2, index_map2)
    expect_op2 = expectation_operator(coords2, index_map2, target_series)

    """
    
    targets_list = targets if isinstance(targets,list) or isinstance(targets,tuple) else [ targets ]
    eop_list = expect_ops if isinstance(expect_ops,list) or isinstance(expect_ops,tuple) else [ expect_ops ]
    # Assume each target series is a contiguous block here -- no second-level lists
    if N is None:
        N = targets_list[0].shape[0]

    if qc_targets is None:
        qcflags_list = [ None for t in targets_list ]
    else: 
        qcflags_list = qc_targets if isinstance(qc_targets,list) or isinstance(qc_targets,tuple) else [ qc_targets ]
    
    # forward_op = shift_operator(scaled_coords, index_map)
    # the shift_operator code is tweeked to allow reverse mapping as well
    if reverse_op is None:
        reverse_op = shift_operator(np.flip(scaled_coords, axis=0), np.flip(index_map, axis=0))
    
    ecoords = np.zeros((N, scaled_coords.shape[1]))
    ecoords[index_map,:] = scaled_coords
    qcmap = np.zeros(N)
    qcmap[index_map] = 2 # to start with, filled below

    
    if targets is not None:
        import warnings
        import cvxpy as cp
        constraints_list = []
        x = cp.Variable((scaled_coords.shape[1],1))
        constraints_list += [ x >= np.min(scaled_coords.T * (1-extent), axis=1)[:,None], x <= np.max(scaled_coords.T * (1+extent), axis=1)[:,None]]
        
        target_x = cp.Parameter((scaled_coords.shape[1],1))
        min_bounds = [ cp.Parameter((1,t.shape[1])) for t in targets_list ]
        max_bounds = [ cp.Parameter((1,t.shape[1])) for t in targets_list ]
        min_bound_values = [ np.nanmin(t, axis=0)[None,:]*(1-extent) for t in targets_list ]
        max_bound_values = [ np.nanmax(t, axis=0)[None,:]*(1+extent) for t in targets_list ]
        
        constraints_list += [ x[0] == 1 ]
        for i, eop in enumerate(eop_list):
            valid_t = ~np.isnan(targets_list[i]).any(axis=1)
            if qcflags_list[i] is not None:
                valid_t[qcflags_list[i]<=0] = False
            scale = np.std(targets_list[i][valid_t,:], axis=0)
            shift = np.mean(targets_list[i][valid_t,:], axis=0)
            eop_value = (eop @ x - shift) / scale
            constraints_list += [ eop_value >= (min_bounds[i] - shift) / scale]
            constraints_list += [ eop_value <= (max_bounds[i] - shift) / scale]
                
        problem = cp.Problem(cp.Minimize(cp.sum_squares(x - target_x)), constraints_list)
        
        solver_args = dict()
        solver = cp.OSQP
        solver_args['verbose'] = verbose>=2
        solver_args['eps_abs'] = 1e-6
        solver_args['eps_rel'] = 1e-6
        solver_args['eps_prim_inf'] = 1e-6
        solver_args['eps_dual_inf'] = 1e-6
        solver_args['solver'] = solver
        solver_args['max_iter'] = 1000
    
    # end of the series - extrapolate forwards from the last known state
    # Make it so the extrapolated states best match the observed data
    # => they should stay on the "attractor" instead of starting to converge
    # towards the limit distribution
    fs = scaled_coords[-1:,:].T
    fsm = fs.copy()
    for i in range(index_map[-1]+1,N):
        fs = shift_op @ fs
        if verbose>0: print(f"filling last value {i-index_map[-1]}/{N-index_map[-1]-2}")
        qc = 0
        if i < N-1 and targets is not None:
            fsm = shift_op @ fsm
            # want to be as close as possible to the trajectory
            x.value = fs
            target_x.value = fsm
            try:
                # while respecting the target value constraints
                for bi, tar in enumerate(targets_list):
                    ext = extent
                    # i+1: expect_op with immediate future: TODO function arg
                    tval = tar[i+1:i+2,:]
                    is_valid = np.isfinite(tval).all(axis=1)[0]
                    if qcflags_list[bi] is not None and qcflags_list[bi][i+1]<=0:
                        is_valid = False
                    if is_valid:
                        qc += 1
                    else:
                        tval = eop_list[bi] @ fs
                        ext = extent_nan
                    min_bounds[bi].value = np.max([min_bound_values[bi], tval - np.abs(tval)*ext], axis=0)
                    max_bounds[bi].value = np.min([max_bound_values[bi], tval + np.abs(tval)*ext], axis=0)
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    problem.solve(**solver_args)
            except:
                x.value = None
                qc = 0
            # None for no solution
            if x.value is not None:
                fs = x.value[:,:1]
                qc /= len(targets_list)
        else: 
            if targets is None:
                qc = -3
            else:
                qc = -1

        ecoords[i,:] = fs[:,0]
        qcmap[i] = qc
    
    # beginning of the series - extrapolate backwards from the first known state
    bs = scaled_coords[0:1,:].T
    bsm = bs.copy()
    for i in range(index_map[0]-1,-1,-1):
        bs = reverse_op @ bs
        
        if targets is not None:
            bsm = reverse_op @ bsm
            if verbose>0: print(f"filling first value {index_map[0]-1-i}/{index_map[0]-1}")
    
            # want to be as close as possible to the trajectory
            x.value = bs
            target_x.value = bsm
            # while respecting the target value constraints
            try:
                qc = 0
                for bi, tar in enumerate(targets_list):
                    # i+1: expect_op with immediate future: TODO function arg
                    ext = extent
                    tval = tar[i+1:i+2,:]
                    is_valid = np.isfinite(tval).all(axis=1)[0]
                    if qcflags_list[bi] is not None and qcflags_list[bi][i+1]<=0:
                        is_valid = False
                    if is_valid:
                        qc += 1
                    else:
                        tval = eop_list[bi] @ bs
                        ext = extent_nan
                    min_bounds[bi].value = np.max([min_bound_values[bi], tval - np.abs(tval)*ext], axis=0)
                    max_bounds[bi].value = np.min([max_bound_values[bi], tval + np.abs(tval)*ext], axis=0)
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    problem.solve(**solver_args)
            except:
                x.value = None
                qc = 0
           # None for no solution
            if x.value is not None:
                bs = x.value[:,:1]
                qc /= len(targets_list)
        else:
            qc = -3
            
        ecoords[i,:] = bs[:,0]
        qcmap[i] = qc
    
    anterior = reverse_op @ bs
    
    # NaN gaps only - no begin or end of the series
    # it is assumed the series are contiguous in this function
    gaplocs = np.arange(scaled_coords.shape[0]-1)[np.diff(index_map)!=1]
    gaplens = index_map[gaplocs+1] - index_map[gaplocs]
    for gapidx, (gloc, glen) in enumerate(zip(gaplocs, gaplens)):
        # forward and backwards states
        fs = scaled_coords[gloc:gloc+1,:].T
        bs = scaled_coords[gloc+1:gloc+2,:].T
        # stay close to the original unperturbed trajectory
        # otherwise, errors can accumulate and the trajectory diverges
        fsm = fs.copy()
        bsm = bs.copy()
        
        # interpolate all states in between
        ei = index_map[gloc]
        for gi in range(1,glen):
            i = ei+gi # index in series, ecoords
            w = (glen-gi) / glen # w decreases as it moves away
            
            fs = shift_op @ fs
            fsm = shift_op @ fsm
            
            if targets is not None:
                x.value = fs
                target_x.value = fsm.copy()
                try:
                    qcfs = 0
                    for bi, tar in enumerate(targets_list):
                        ext = extent
                        tval = tar[i+1:i+2,:]
                        is_valid = np.isfinite(tval).all(axis=1)[0]
                        if qcflags_list[bi] is not None and qcflags_list[bi][i+1]<=0:
                            is_valid = False
                        if is_valid:
                            qcfs += 1
                        else:
                            # errors do not accumulate here, fs is close to fsm
                            # so take advantage of starting from the closer point
                            # from previous targets
                            tval = eop_list[bi] @ fs
                            ext = extent_nan
                        min_bounds[bi].value = np.max([min_bound_values[bi], tval - np.abs(tval)*ext], axis=0)
                        max_bounds[bi].value = np.min([max_bound_values[bi], tval + np.abs(tval)*ext], axis=0)
                        
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        problem.solve(**solver_args)
                except:
                    x.value = None
                    qcfs = 0
                if x.value is not None:
                    fs = x.value
                    qcfs /= len(targets_list)
            else:
                qcfs = -3

            qcmap[i] += qcfs
            
            ecoords[i,:] += w * fs[:,0]
            
            
            i = ei+glen-gi
            bs = reverse_op @ bs
            bsm = reverse_op @ bsm

            if targets is not None:
                x.value = bs
                target_x.value = bsm.copy()
                try:
                    qcbs = 0
                    for bi, tar in enumerate(targets_list):
                        ext = extent
                        tval = tar[i+1:i+2,:]
                        is_valid = np.isfinite(tval).all(axis=1)[0]
                        if qcflags_list[bi] is not None and qcflags_list[bi][i+1]<=0:
                            is_valid = False
                        if is_valid:
                            qcbs += 1
                        else:
                            # errors do not accumulate here, bs is close to bsm
                            # so take advantage of starting from the closer point
                            # from previous targets
                            tval = eop_list[bi] @ bs
                            ext = extent_nan
                        min_bounds[bi].value = np.max([min_bound_values[bi], tval - np.abs(tval)*ext], axis=0)
                        max_bounds[bi].value = np.min([max_bound_values[bi], tval + np.abs(tval)*ext], axis=0)
                        
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        problem.solve(**solver_args)
                except:
                    x.value = None
                    qcbs = 0
                if x.value is not None:
                    bs = x.value
                    qcbs /= len(targets_list)
            else:
                qcbs = -3
                
            qcmap[i] += qcbs
            
            # same w, i has changed, run in reverse direction
            ecoords[i,:] += w * bs[:,0]

        ei = index_map[gloc]
        for gi in range(1,glen):
            i = ei+gi
            qcmap[i] /= 2
            
        # eops are linear, so the interpolated ec = fs * w + bs * (1-w)
        # also produce the eop targets


    # adjust the original valid states so the expectation operators
    # best match the observed data -- optional flag, this may overfit
    if adjust_ori:
        for imi, i in enumerate(index_map):
            fs = scaled_coords[imi:imi+1,:].T
            if verbose>0: print(f"adjusting the original states {imi+1}/{index_map.shape[0]}")
            if targets is not None:
                x.value = fs
                target_x.value = fs
                qc = 2 # original state / data
                try:
                    for bi, tar in enumerate(targets_list):
                        tval = tar[i+1:i+2,:] # valid by construction of index_map
                        min_bounds[bi].value = np.max([min_bound_values[bi], tval - np.abs(tval)*ext], axis=0)
                        max_bounds[bi].value = np.min([max_bound_values[bi], tval + np.abs(tval)*ext], axis=0)
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        problem.solve(**solver_args)
                except:
                    x.value = None
                if x.value is not None:
                    fs = x.value[:,:1]
                    qc = 3 # adjusted state
            else:
                qc = 2 # original data
            ecoords[i,:] = fs[:,0]
            qcmap[i] = qc
    
    if targets is None:
        # no extrapolated targets
        return ecoords, qcmap
    
    # update for computing the extrapolated targets below
    prev_states = np.concatenate([anterior.T, ecoords[:-1,:]], axis=0)
    
    extrapolated_targets = []
    for eop, tar in zip(eop_list, targets_list):
        extrapolated_targets += [ (eop @ prev_states.T).T ]
    
    if len(extrapolated_targets)==1:
        return extrapolated_targets[0], ecoords, qcmap

    return extrapolated_targets, ecoords, qcmap


# TODO: Check and/or complete this function
def extrapolate_states(scaled_coords, index_map, shift_op, series):
    
    series_list = series if isinstance(series,list) or isinstance(series,tuple) else [ series ]
    
    splitmaps = kernels.split_concatenated_index_map(series, index_map)
    
    cstart = 0
    ret = []
    # NOOOO: second-level = discontiguous, not first
    for ser, smap in zip(series_list, splitmaps):
        c = scaled_coords[cstart:cstart+smap.shape[0],:]
        ret += [ extrapolate_contiguous(c, smap, shift_op, len(ser)) ]
        cstart += smap.shape[0]
     
    return ret if isinstance(series,list) or isinstance(series,tuple) else ret[0]
    
    

def predict(npred, state_dist, expect_op, shift_op = None, knn_analog = None, return_dist = 0, verbose = False, bounds = None, expect_targets = None, expect_target_weights = None, knn_convexity = None, solver = None, **kwargs):
    """
    Predict values from the current causal states distribution

    Parameters
    ----------
    npred : int
        Number of predictions to generate
        
    state_dist : array of size (num_basis, 1)
        Current state distribution, expressed in the eigen basis
        
    expect_op : array of size (data_dim, num_basis) or a list of such arrays
        Operator, or list of operators, that take a distribution of causal states and generate a data value compatible with the original series
        
    shift_op : array of size (num_basis, num_basis), optional
        Operator to evolve the distributions in time
        If this argument is None, then knn_analog must be specified

    knn_analog : int, optional
        If not specified, then shift_op must be. Both cannot be specified together.
        If specified, predictions in state space are made using analog forecast with this number of neighbors. That is, the neighbors of the current state are looked up. Then, the prediction for the current state is the average of all their successors. Additional arguments are:
          
          + coords: The coordinates of each observed state in the eigen basis. This argument is required.
          
          + eigenvalues: eigenvalues for weighting distance computations
          
          + index_map: required, used for filtering states without successors
          
          + analog_weight_factor: for weighting the forecast. The weight for each neighbor is w = exp(- analog_weight_factor * dist / mean(dists)), then normalized by sum(w). The default is 3.
        
          + knn_dim (optional): Restrict the number of data dimensions to make the nearest neighbors search on. This can be useful when a low-dimensional attractor is embedded in a large-dimension space, where only the first few dimensions contribute to the distances, for accelerating the computations. knn_dim can be a number, in which case the first dimensions below that number are selected, or an array-like object of indices. The default is to use all available dimensions in the coords array.
        
    return_dist : {0, 1, 2}, optional
        Whether to return the state distribution, or update it:
        - 0 (default): do not return a state vector
        - 1: return the updated state vector
        - 2: return an array of state_dist vectors, one row for each prediction
        
    verbose : int, optional
        Whether to print progress. The default is False (0). True (1) displays basic progress for each prediction step. A value >1 also sets the verbosity flag in the solver extra arguments. See the doc for 'solver' below.

    bounds : list, optional
        Bounds for each expect_op operator prediction. 
        This list should have the same length as expect_op.
        Each list entry is itself an array-like of two elements, one for the lower and one for the upper bound. Use None or +/- inf values to leave a bound unspecified. For vector data, use a tuple (or list) of two vectors, for each bound.
        As a convenience, when there is a single expect_op, the bounds can be specified directly, without being wrapped in a single-element list.

    expect_targets : list or array, optional
        Some known values could be used as guides to optimize the predictions
        of other unknown values. Must be either None, or a list where each
        unknown parameter is None and each target is an array with as many 
        values as the number of predictions to make. 
        
    expect_target_weights : list, optional
        weights given to respecting these targets, with respect to other
        constraints such as knn
    
    knn_convexity : int or None, optional
        If specified, restrict each state estimate to the closest point in the convex enveloppe of its nearest neighbors. That envelop can be slightly enlarged with the extent parameter, because some valid states may exist out of the currently observed data points. knn_convexity specifies the number of neighbors to use. This number should be large enough so as to reach multiple nearby trajectories. Otherwise, neighbors consist only of the points just before/after along trajectories and the state synchronizes to the training data. Additional arguments coords and knn_dim as for knn_analog, plus:
       
          + extent (optional): Change the bounds of the convexity constraints from [0,1], to [0-extent,1+extent]. The default is 0, but you should really try to increase this to something like 0.05, possibly adding data bounds if needed.

        
    solver : str, optional
        Which CVXPY solver to use in order to enforce the bounds and knn_convexity constraints. Valid values are:

            + 'OSQP_AUTO' (default): use either OSQP_MKL if MKL is available, or OSQP_QDLDL otherwise
            
            + 'OSQP_QDLDL' : use the QDLDL implementation of CVXPY to solve the problem. This is single-threaded
            
            + 'OSQP_MKL' : use the MKL implementation of CVXPY to solve the problem. This is multi-threaded, but requires MKL to be installed. IMPORTANT NOTE : There is a library loading bug with some versions of CVXPY / MKL that may cause the program to crash. See also : https://github.com/oxfordcontrol/osqp-python/issues/52

            + other_string : one of the solvers from cvxpy.installed_solvers(). For example, the CVXOPT solver is multi-threaded, but may be slower but more precise than OSQP_QDLDL

       Additional solver arguments may be passed as a dict(), itself passed as an additional kwargs extra argument called 'solver_args'. One of these extra arguments is 'verbose', which you can set to True like this: sargs=dict(); sargs['verbose']=True; then pass solver_args=sargs as an argument to this function. As a convenience, a value of verbose > 1 for this function, also sets the solver verbose flag to True. See the doc for 'verbose' above.
        
    **kwargs : dict, optional
        Optional arguments, see the notes

    Returns
    -------
    predictions: array, or list of arrays, matching the expect_op type
        As many series of npred values as there are expectation operators. Either a list, or a single array, matching the expect_op type
    
    updated_dist: optional, array
        If return_dist>0, the updated state distribution or all such distributions are returned as a second argument.
    
    Notes
    -----
    
    The causal state space (conditional distributions) is not convex: linear combinations of causal states do not necessarily correspond to a valid value in the conditionned domain. We are dealing here with distributions of causal states, but these distributions are ultimately represented as linear combinations of the data span in a Reproducing Kernel Hilbert Space, or, in this case, as combinations of eigenbasis vectors (themselves represented in RKHS). Therefore, the state distributions are also linear combinations of other causal states. Ultimately, the continuous-time model converges to a limit distribution, which is an average distribution that need not correspond to any single state, so the non-convexity is not an issue for that limit distribution. Making predictions with the expectation operator is still feasible, and we get the expected value from the limit distribution as a result. 
    
    If, instead, one wants a trajectory, and not the limit average, then a method is required to ensure that each predicted state remain valid as a result of applying a linear shift operator. The knn_convexity argument is an attempt to solve this issue. The API is not definitive and may change in the future. The preimage issue is a recurrent problem in machine learning and no single answer can currently solve all cases.
       
    WORK IN PROGRESS - MAY NOT WORK YET - API BELOW MAY HAVE CHANGED - USE WITH CAUTION. The 'project_to_manifold' extra argument builds an estimate of a manifold best fitting the data, using tangent subspaces found with Spectral Exterior Calculus (SEC). See the source code for possible arguments and concaust.SEC for a self-contained demo. Any additional extra parameter is passed on to the tangent_frames, tangent_bases and project_to_surface methods, some of which can be quite influencial on the quality of the result. Additional kwargs are passed to the numerical solver (API still in progress)          
          + coords: The coordinates of each observed state in the eigen basis. This argument is required.
        
          + 'tangent_dim' (semi-optional): The dimension of the tangent subspaces. This parameter has no effect if the 'tangent_bases' parameter is provided, otherwise it is strongly recommended. See the documentation in tangent_bases.
          
          + 'tangent_bases' (optional): You can precompute tangent bases with the function of the same name and pass the result as argument. This can be useful for testing the influence of other parameters, or for calling this "predict" function multiple times, without having to recompute the tangent frames and bases at each call.
        
          + 'SEC_scale' (semi-optional): Used for the SEC code, homogeneous to a distance between scaled coordinates. Should cover at least a few neighbors. Too small or too large values will screw the directions of the tangent bases. This parameter has no effect if the 'tangent_bases' parameter is provided, otherwise it must be provided. If you really want to try auto-tuning, set it to None, but you will probably have to adjust the printed scale.
       
          + knn (optional): If an integer value is passed, it is interpreted as a number of neighors to use for computing the tangent subspaces. If a float value is passed, it is interpreted as a radius in scaled coordinate space. The default is integer 5.
    
    
    """
    
    eoplist = expect_op if isinstance(expect_op,list) or isinstance(expect_op,tuple) else [ expect_op ]
    
    state_dist = state_dist.copy()
    
    all_state_dists = []
    
    constraints_list = []
    x = None
    target_x = None
    problem = None
    nbrs = None
    
    if knn_analog is not None:
        if shift_op is not None:
            raise ValueError("knn_analog and shift_op cannot be specified together")
        if knn_convexity is not None:
            raise ValueError("knn_convexity cannot be specified if knn_analog is")
        analog_weight_factor = kwargs['analog_weight_factor'] if 'analog_weight_factor' in kwargs else 3
    
    if knn_analog is None and shift_op is None:
        raise ValueError("knn_analog or shift_op must be specified")
    
    # Hidden feature for debug purposes, useless in practice
    enforce_positive_weights = kwargs['enforce_positive_weights'] if 'enforce_positive_weights' in kwargs else False
    
    if enforce_positive_weights:
        #from . import nnls_impl
        coords = kwargs['coords']
        cinv = scipy.linalg.pinv(coords.T)
        # CVXPY constraint works, but yields worse results
        #import cvxpy as cp
        #w = cp.Variable((coords.shape[0], 1))
        #x = coords.T @ w
        #constraints_list += [w >= 0]
    
    elif knn_convexity or knn_analog:
        import cvxpy as cp
        if 'coords' not in kwargs:
            raise ValueError("coords are needed for knn methods")
        coords = kwargs['coords']
        if 'eigenvalues' not in kwargs:
            raise ValueError("eigenvalues are needed for knn methods")
        eigenvalues = kwargs['eigenvalues']
        knn_dim = kwargs['knn_dim'] if 'knn_dim' in kwargs else coords.shape[1]
        import numbers
        if isinstance(knn_dim, numbers.Integral):
            knn_dim = np.arange(knn_dim, dtype=int)
        from sklearn.neighbors import NearestNeighbors
        if verbose: print("knn: building the nearest neighbors search structure")
        if knn_analog:
            if 'index_map' not in kwargs:
                raise ValueError("index_map is needed for knn_analog")
            index_map = kwargs['index_map']
            # Consider only as potential neighbors, these for which
            # there is a successor. See the code in shift_op
            valid_prev = np.arange(coords.shape[0]-1)[np.abs(np.diff(index_map))==1]
            nbrs = NearestNeighbors(n_neighbors=knn_analog).fit((coords*eigenvalues)[valid_prev,:][:,knn_dim])
            
        # convexity has lesser priority if analog is also specified
        else:
            extent = kwargs['extent'] if 'extent' in kwargs else 0
            nbrs = NearestNeighbors(n_neighbors=knn_convexity).fit((coords*eigenvalues)[:,knn_dim])
            # Neighbors, to find below
            M = cp.Parameter((coords.shape[1], knn_convexity))
            w = cp.Variable((knn_convexity,1))
            # for all purposes, x is defined as a combination of neighors
            # the problem formulation is otherwise the same
            x = M @ w
            constraints_list += [w >= -extent, w <= 1+extent, cp.sum(w) == 1]
            
    
    if bounds is not None:
        if not hasattr(bounds, "__len__"):
            raise ValueError("bounds must be list or array")
            
        if len(expect_op)==1:
            if len(bounds) == 2:
                bounds = [ bounds ]

        if len(bounds) != len(expect_op):
            raise ValueError("bounds should be specified for each expect_op. To disable bounds, use None or inf values")
    
        bound_values = []
        for b in bounds:
            u = np.empty((npred,2))
            if b is not None and hasattr(b,'shape'):
                if b.ndim == 1 and len(b)==2:
                    u[:,0] = b[0]
                    u[:,1] = b[1]
                elif b.ndim==2 and b.shape[0]>=npred and b.shape[1]==2:
                    u = b
                else:
                    raise ValueError("invalid bound array specification, size does not match")
            else:
                if b is not None and (not hasattr(b, "__len__") or len(b) != 2):
                    raise ValueError("invalid bound list specification, size does not match")
                u[:,0] = b[0]
                u[:,1] = b[1]
            idx = np.where(np.isinf(u))
            u[idx] = np.finfo(u.dtype).max * np.sign(u[idx])
            bound_values += [ u ]
            
        import cvxpy as cp
        
        bound_params = []
        
        # x could be defined by knn_convexity already
        if x is None:
            x = cp.Variable(state_dist.shape)
        for i, b in enumerate(bounds):
            if b is None:
                bound_params += [ None ]
                continue
            bp = cp.Parameter(2)
            bound_params += [ bp ]
            eop_value = eoplist[i] @ x
            if True:
            #if b[0] is not None and np.isfinite(b[0]):
                constraints_list += [ eop_value >= bp[0] ]
            #if b[1] is not None and np.isfinite(b[1]):
                constraints_list += [ eop_value <= bp[1] ]

    if expect_targets is not None:
        if not hasattr(expect_targets, "__len__"):
            raise ValueError("expect_targets must be list or array")

        if len(expect_targets) != len(expect_op):
            raise ValueError("expected targets should be specified for each expect_op. To disable some targets, use None")
        
        expect_targets_params = []
        
        etexpr = None
            
        # x could be defined by knn_convexity already
        if x is None:
            x = cp.Variable(state_dist.shape)
        
        if expect_target_weights is None:
            expect_target_weights = np.ones(len(expect_targets))
            
        for i, (t, tw) in enumerate(zip(expect_targets, expect_target_weights)):
            if t is not None:
                etp = cp.Parameter(eoplist[i].shape[0])
                if tw is not None and tw != 0:
                    eop_value = eoplist[i] @ x
                    if etexpr is None:
                        etexpr = tw * cp.sum_squares(eop_value - etp)
                    else:
                        etexpr = etexpr + tw * cp.sum_squares(eop_value - etp)
                expect_targets_params += [ etp ]
    
    """
    if constraint == "manifold":
        coords = kwargs['coords']
        tb = kwargs['tangent_bases'] if 'tangent_bases' in kwargs else None
        scale = kwargs['scale'] if 'scale' in kwargs else None
        tangent_dim = kwargs['tangent_dim'] if 'tangent_dim' in kwargs else None
        n0 = kwargs['n0'] if 'n0' in kwargs else None
        n1 = kwargs['n1'] if 'n1' in kwargs else None
            
        from . import SEC
        if tb is None:
            if verbose: print("manifold: building the tangent bases")
            tf = SEC.tangent_frames(scaled_coords, scale, n0, n1)
            tb = SEC.tangent_bases(tf, tangent_dim)
            del tf # saves memory as soon as possible
    """
    
    # No constraints, no problem!
    if constraints_list:
        # preserve the state distribution normalization
        # This is always 1, in whatever scaled or coords unit, since
        # the first eigenvalue is 1
        constraints_list += [ x[0] == 1 ]
        target_x = cp.Parameter(state_dist.shape)
        if expect_targets is not None:
            problem = cp.Problem(cp.Minimize(cp.sum_squares(x - target_x)+etexpr), constraints_list)
        else:
            problem = cp.Problem(cp.Minimize(cp.sum_squares(x - target_x)), constraints_list)
    
        solver_args = kwargs['solver_args'] if 'solver_args' in kwargs else dict()
        
        if solver is None and 'solver' in solver_args:
            solver = solver_args['solver']
    
        if solver is None or solver == 'OSQP_AUTO':
            try:
                import mkl
                solver = 'OSQP_MKL'
            except:
                solver = 'OSQP_QDLDL'

        if solver == 'OSQP' or solver == 'OSQP_QDLDL':
            solver = cp.OSQP

        if solver == 'OSQP_MKL':
            solver = cp.OSQP
            solver_args['linsys_solver'] = "mkl pardiso"
        
        if solver == cp.OSQP:
            if 'verbose' not in solver_args:
                solver_args['verbose'] = False
            if 'eps_abs' not in solver_args:
                solver_args['eps_abs'] = 1e-6
            if 'eps_rel' not in solver_args:
                solver_args['eps_rel'] = 1e-6
            if 'eps_prim_inf' not in solver_args:
                solver_args['eps_prim_inf'] = 1e-6
            if 'eps_dual_inf' not in solver_args:
                solver_args['eps_dual_inf'] = 1e-6
    
        solver_args['solver'] = solver
        
        solver_args['ignore_dpp'] = True

        if verbose>1:
            solver_args['verbose'] = True

    pred = [np.zeros((npred,eop.shape[0])) for eop in eoplist]
    for p in range(npred):
        if verbose: print(f"predict: {p+1}/{npred}")
        if return_dist==2:
            all_state_dists += [state_dist]
            
        # Apply the expectation operator to the current state estimate
        # to make a prediction in data space
        for eidx, eop in enumerate(eoplist):
            pred[eidx][p,:] = (eop @ state_dist).T
            
        # Evolve the current state
            
        # Analog forecast method does not use the shift operator
        if knn_analog is not None:
            kdist, kidx = nbrs.kneighbors((state_dist*eigenvalues[:,None])[knn_dim].T)
            kdist = np.squeeze(kdist)
            kidx = np.squeeze(kidx)
            
            weights = np.exp(-analog_weight_factor * kdist/np.mean(kdist))
            
            # modulation test to generate variability
            #weights = weights * (np.abs(np.sin((np.arange(knn_analog)+p/100)  / knn_analog)))
            # modulation with noise
            #weights *= (1+np.random.uniform(size=len(kidx)))
            
            # all these indices have successors by construction, see above
            samples = np.copy(coords[kidx+1,:])
            state_dist = np.sum(samples * weights[:,None], axis=0)[:,None] / np.sum(weights)
            #state_dist = np.mean(samples, axis=0)[:,None]
            
        # test was done above for one or the other
        else:
            state_dist = shift_op @ state_dist
            # Normalize - not needed anymore by construction of the shift op
            # However, there could still be cumulating numerical inaccuracies
            #state_dist /= state_dist[0,0]
            
        if problem is not None:
            # For knn_convexity, x is not the primary variable
            if knn_convexity is not None:
                kidx = nbrs.kneighbors((state_dist*eigenvalues[:,None])[knn_dim].T)[1][0]
                M.value = coords[kidx,:].T
            elif isinstance(x,cp.Variable):
                x.value = state_dist # warm start
            
            target_x.value = state_dist # also the solution without constraints
            
            if bounds is not None:
                for i in range(len(expect_op)):
                    if bound_params[i] is None:
                        continue
                    bound_params[i].value = bound_values[i][p,:]
                    
            if expect_targets is not None:
                itp = 0
                for i, t in enumerate(expect_targets):
                    if t is not None:
                        if p<t.shape[0]:
                            tvalue = t[p,:]
                        else:
                            tvalue = np.nan
                        # disable optimisation on this target if no target is 
                        # given by optimizing on the currently predicted value
                        if not np.isfinite(tvalue):
                            tvalue = (eoplist[i] @ state_dist)[:,0]
                        expect_targets_params[itp].value = tvalue
                        itp += 1
                    
            try:
                problem.solve(**solver_args)
            except:
                # May lack precision, reached max iter, but still the best
                # approx given the constraints
                pass
            
            if x.value is None:
                raise ValueError("Cannot solve cvxpy")
            
            state_dist = x.value
            # should not be needed mathematically if the 
            # constraint was perfectly respected.
            # There could be numerical inaccuracies in practice
            state_dist /= state_dist[0,0]

        """
        if constraint == "manifold":
            refpt = state_dist.T * eigenvalues
            # knn and knn_dim kwargs are passed, the default is knn=5 in SEC
            proj = SEC.project_to_surface(refpt, scaled_coords, tb, **kwargs)
            
            pstate.value = proj.T # target constant
            x.value = state_dist # warm start
            #problem.solve(verbose = True, solver=cp.CVXOPT, kktsolver='chol', abstol=1e-5, reltol=1e-5)
            eps = 1e-6
            problem.solve(solver=cp.OSQP, linsys_solver="mkl pardiso", verbose=False, eps_abs=eps, eps_rel=eps, eps_prim_inf=eps, eps_dual_inf=eps)
            state_dist = x.value
            #if clip is not None:
            #    proj = clip_along_line(refpt, proj)
            #state_dist = (proj / eigenvalues).T
            
            # should not be needed mathematically if the 
            # constraint was perfectly respected.
            # This is never the case in practice
            state_dist /= state_dist[0,0]
        """
        
        if enforce_positive_weights:
            #print(np.sum(w.value))
            w = cinv @ state_dist
            #w = nnls_impl.nnls(coords.T, state_dist)
            # The min should be 0 or near 0, and the sum near 1
            # This should not be needed in practice
            print(np.min(w), np.sum(w))
            w[w<0] = 0
            state_dist = coords.T @ (w/np.sum(w))

    if not isinstance(expect_op,list) and not isinstance(expect_op,tuple):
        pred = pred[0]
        
    if return_dist == 1:
        return pred, state_dist
    
    elif return_dist == 2:
        return pred, np.concatenate(all_state_dists,axis=1)
    
    return pred
    
