"""
This is the code for the Continuous Causal States method
described in the papers:

  Discovering Causal Structure with Reproducing-Kernel Hilbert Space ε-Machines
  by Nicolas Brodu and James P. Crutchfield

  Inferring Kernel ε-Machines: Discovering Structure in Complex Systems
  by Nicolas Brodu and Alexandra M. Jurgens1

  and possibly more to come
  
This code is free software, according to the MIT license
Author & Contact: nicolas.brodu@inria.fr
"""

import numpy as np
import math

from numba import jit, prange, cuda

# Exponential family, parameters for the distances.
# Gaussian kernel : -0.5 * dist**2
# Laplacian kernal: - dist ** 1
# custom kernel: some_callable, None
# Scales are provided separately
__kernel_params__ = [ -0.5, 2 ]
#__kernel_params__ = [ -1, 1 ]


def set_kernel(kernel_type = 'Gaussian'):
    """
    Sets a kernel type, amongst supported kernels
    kernel_type:
        'Gaussian' string: classical Gausian kernel exp(-0.5 * dist**2 / scale**2)
        'Laplacian' string: classical Laplacian kernel exp(- dist / scale )
        callable: should return the log of the kernel. This is applied to callable(dist**2, scale) in the logk function below
    """
    if type(kernel_type) is str:
        if kernel_type == 'Gaussian':
            __kernel_params__[0] = -0.5
            __kernel_params__[1] = 2
            return
        if kernel_type == 'Laplacian':
            __kernel_params__[0] = -1
            __kernel_params__[1] = 1
            return
        raise ValueError('Invalid kernel type: only Gaussian and Laplacian are supported, or you should provide your own kernel function taking a squared distance as argument')

    raise ValueError('Implementation not complete yet')            
    __kernel_params__[0] = kernel_type
    __kernel_params__[1] = None
    


def compute_index_map_single_source(series, npast, nfuture, skip_start = 0, skip_end = 0, qcflags = None, take2skipX = 0):
    """
    Index mapping helper - for internal use

    Parameters
    ----------
    series : list of arrays
        list of contiguous data blocks
        
    npast : int
        take that consecutive number of samples for building an observed past. This must be an integer, at least 1 value is needed (the present is part of the past)
        
    nfuture : int
        take that consecutive number of samples, just after the past samples, for building an observed future. nfuture can be 0, in which case, the function only looks for valid pasts. combined with npast==1, the function only looks for valid non-NaN values across all data sources
        
    skip_start : int, optional (default 0)
        number of data to mark as invalid at the beginning of each contiguous data block. Useful to align heterogenous data with different npast/nfuture.
        
    skip_end : int, optional (default 0)
        number of data to mark as invalid at the end of each contiguous data block. Useful to align heterogenous data with different npast/nfuture.
        
    qcflags : list of arrays, optional
        quality control flags. <= 0 means invalid data. optional, but if present there must be one qcflag per sample
        
    take2skipX : int, optional (default 0)
        performs a special kind of subsampling: two consecutive samples are retained, X samples are discarded, then this (X+2) sequence repeats. This scheme is designed to preserve consecutive entries for building a shift operator while still allowing subsampling of very large series. The classic subsampling scheme (take one out of X) can also be applied a priori to the original series, but it is equivalent to a (bad) low-pass filtering. Then, the shift operator would be computed on consecutive entries of the subsampled series, hence at a different time scale. The take2skipX allows to still work on the original time scale. Both can be combined (after appropriate low-pass filtering).

    Returns
    -------
    concat_idx_map: (N,) array of int
        Returns the indices of the valid entries, but stacking all series passed as argument in a single series. The indices refer to valid entries in that global stacked array. Hence, discontiguous data blocks generate invalid entries in the global stacked array.
    """
    
    if npast < 1:
        raise ValueError("npast must be a strictly positive integer")
        
    # nfuture can be 0
    if nfuture < 0:
        raise ValueError("nfuture must be a positive integer or null")
    
    if not isinstance(series,list) and not isinstance(series,tuple):
        #raise ValueError("series must be a list of arrays")
        series = [series]

    if qcflags is None:
        qcflags_list = [ None for ser in series ]
    else: 
        qcflags_list = qcflags if isinstance(qcflags,list) or isinstance(qcflags,tuple) else [ qcflags ]
        
    concat_idx_map = None
    
    try:
        # Use cupy if any series is on the GPU
        import cupy as cp
        xp = np
        for ser in series:
            if cp.get_array_module(ser) is cp:
                xp = cp

    except:
        xp = np

    
    nbefore = 0
    for sidx, s in enumerate(series):
        # s is an array
        if s.ndim==1:
            s = s[:,None]
            
        # if any series is on the GPU, so is the resulting map
        # but transfer them one by one
        if xp is not np:
            # ensure all series are on the GPU. No effect if already there
            s = xp.asarray(s)
            
        n = s.shape[0]
        if n < npast + nfuture:
            valid_pf = xp.full(n, False)
        else:
            # valid instantaneous time points
            valid_t = ~xp.isnan(s).any(axis=1)
            if qcflags is not None and qcflags_list[sidx] is not None:
                valid_t[qcflags_list[sidx]<=0] = False
            valid_t[:skip_start] = False
            valid_t[n-skip_end:] = False
            # valid past/future combinations
            valid_pf = valid_t.copy()
            for i in range(1,nfuture+1):
                valid_pf[:n-i] &= valid_t[i:n]
            for i in range(1,npast):
                valid_pf[i:n] &= valid_t[:n-i]
                
            valid_pf[:npast-1] = False
            valid_pf[n-nfuture:] = False

            if take2skipX>0:
                i = npast - 1
                while i<n:
                    i += 2
                    if i>=n:
                        break;
                    for k in range(take2skipX):
                        valid_pf[i] = False
                        i += 1
                        if i>=n: break;
        
        valid_idx = xp.arange(n)[valid_pf]
            
        if concat_idx_map is None:
            concat_idx_map = valid_idx + nbefore
        else:
            concat_idx_map = xp.concatenate([concat_idx_map, valid_idx + nbefore], axis=0)
            
        nbefore += n
        
    return concat_idx_map


def compute_index_map_multiple_sources(series_list, npasts_list, nfutures_list, qcflags_list, take2skipX = 0):
    """
    Index mapping helper - for internal use
    
    See parameters of compute_index_map_single_source
    
    This version takes lists of multiple sources as argument.
    
    """

    try:
        import cupy as cp
        xp = np
        for ser in series_list:
            if cp.get_array_module(ser) is cp:
                xp = cp
                break

    except:
        xp = np

    # Each source may have its own NaN patterns and the computed 
    # index maps do not match. => Compute these index maps for all 
    # heterogenous sources and then retain only these indices that
    # are common to all sources.
    # Then, pass that global index map to the _logk function
    
    max_npast = max(npasts_list)
    max_nfuture = max(nfutures_list)
    
    valid_map = None
    for ser, npa, nfu, qc in zip(series_list, npasts_list, nfutures_list, qcflags_list):
        skip_start = max_npast - npa
        skip_end = max_nfuture - nfu
        
        if xp is not np:
            # ensure all series are on the GPU. No effect if already there
            ser = xp.asarray(ser)
        
        sidxmap = compute_index_map_single_source(ser, npa, nfu, skip_start, skip_end, qc, take2skipX)
        
        # transfer the result maps on GPU if any of the series is on GPU
        if xp is not np:
            if cp.get_array_module(sidxmap) is not cp:
                sidxmap = cp.asarray(sidxmap)
                if valid_map is not None and cp.get_array_module(valid_map) is not cp:
                    valid_map = cp.asarray(valid_map)
        
        # combine global indices
        # retain only indices that are valid for all sources
        # A NaN in one source prevents the kernel combination
        # TODO: another possibility is to ignore the source, but how?
        # replace by mean? divide by nsource-1 ?
        valid_map = sidxmap if valid_map is None else xp.intersect1d(valid_map, sidxmap, assume_unique=True)

    return valid_map


def compute_index_map(series, npast, nfuture, qcflags = None, take2skipX = 0):
    """
    Index mapping helper - return the indices in the series that have valid (past, future) sequences

    Parameters
    ----------
    series : array or list or list of lists
        Each array is a time-contiguous block of data, with as many rows as time samples and each column is a feature. NaN values are allowed and detected. 
        A list may be provided, in which case each list entry is a different data source. These multiple sources must have the same number of values, but they may have their own NaN values and quality flags. Indices that are valid for all sources are computed in the returned map.        
        Each first-level list entry can itself be a second-level list instead of a unique array. In that case, these second-level lists gather data acquired in different time-continuous blocks. Example: in order to enter a unique data source, but with multiple time blocks, the syntax is to use a double list : [[data1, data2, ...]] . There, the first-level list specifies a unique data source, and the second level list specifies the contiguous data blocks. Multiple sources and multiple blocks read: [[src1_block1, src1_block2], [src2_block1, src2_block2]]. Same-numbered blocks from different sources must be of the same length, but different blocks may have different lengths. These blocks must be time-consistent across heterogenous data sources: at each time, each data source must be acquired, even though some sensors may fail (NaN values). The returned indices refer to the concatenated version of all these time blocks. Obviously, (past, future) sequences cannot be made over multiple time blocks and this will reflect in the concatenated indices as gaps in the valid values.                                                    
                                                    
    npast : int or listt
        take that consecutive number of samples for building an observed past. If a single integer is provided it applies to all series. Different values can be provided for different data sources (=each series). It is assumed that data samples are measured at matching times for each data source, hence the first causal state can only be computed after the-largest-npast values have been observed.
        
    nfuture : int or list
        take that consecutive number of samples, just after the past samples, for building an observed future. If a single integer is provided it applies to all series. Similarly as for npast, the largest nfuture value sets the time of the last computable causal state.
        
    qcflags : int or list, optional
        quality control flags. <= 0 means invalid data. optional, but if present there must be one qcflag per sample. The default is None.
        
    take2skipX : int, optional (default 0)
        performs a special kind of subsampling: two consecutive samples are retained, X samples are discarded, then this (X+2) sequence repeats. This scheme is designed to preserve consecutive entries for building a shift operator while still allowing subsampling of very large series. The classic subsampling scheme (take one out of X) can also be applied a priori to the original series, but it is equivalent to a (bad) low-pass filtering. Then, the shift operator would be computed on consecutive entries of the subsampled series, hence at a different time scale. The take2skipX allows to still work on the original time scale. Both can be combined (after appropriate low-pass filtering).

    Returns
    -------
    concat_idx_map: (N,) array of int
        Returns the indices of the valid entries. These indices refer to valid entries in a concatenated sources in the case of multiple time blocks.

    """
    
    series_list = series if isinstance(series,list) or isinstance(series,tuple) else [ series ]
    nseries = len(series_list)
    npasts_list = npast if hasattr(npast, "__len__") else [ npast ] * nseries
    nfutures_list = nfuture if hasattr(nfuture, "__len__") else [ nfuture ] * nseries
    
    if qcflags is None:
        qcflags_list = [ None for ser in series_list ]
    else: 
        qcflags_list = qcflags if isinstance(qcflags,list) or isinstance(qcflags,tuple) else [ qcflags ]
    
    return compute_index_map_multiple_sources(series_list, npasts_list, nfutures_list, qcflags_list, take2skipX)
    

# facility
def split_index_map(series, index_map):
    
    series_list = series if isinstance(series,list) or isinstance(series,tuple) else [ series ]
        
    ser = series_list[0]
    # Second-level lists allowed for discontinuous data blocks
    slist = ser if isinstance(ser,list) or isinstance(ser,tuple) else [ ser ]
    prevn = 0
    valid_list = []
    for s in slist:
        n = prevn + s.shape[0]
        # indices for that s
        sidx = index_map[np.all([index_map>=prevn, index_map<n], axis=0)] - prevn
        valid_list += [sidx]
        prevn = n
    
    # In case a single contiguous block is provided, return a single index map, not a list with a unique element
    if not isinstance(ser,list) and not isinstance(ser,tuple):
        valid_list = valid_list[0]
    
    return valid_list



def expand(data, index_map, N=None):
    """
    Index mapping helper - expand data using NaNs

    Parameters
    ----------
    data : array, 1D
        the data to expand
        
    index_map : array
        the index map as returned by concaust functions
        
    N : int, optional
        The size of the full expansion. The default is None, meaning that the largest index in the map is used

    Returns
    -------
    expanded_data : array
        A data array that has length N, where the indices in the map are filled with the original data values and the others are filled with NaN

    Example
    -------
    Assuming times holds the times of the original data series, which may
    have bad QC flags, missing values, etc
    Then, this plots the first coordinate and leaves holes for missing values
    plt.plot(expand(times[index_map],index_map), expand(coords[:,1],index_map))

    """
    if N is None:
        N = np.max(index_map)+1
    x = np.full(N, np.nan)
    x[index_map] = data
    return x

@jit(nopython=True, cache=True, parallel=True)
def dotprod_to_dist2(G):
    D2 = np.zeros(G.shape)
    for i in range(G.shape[0]):
        for j in range(i):
            d = G[i,i] + G[j,j] - 2 * G[i,j]
            if d<0: 
                d=0
            D2[i,j] = d
            D2[j,i] = d
    return D2


@jit(nopython=True, cache=True, parallel=True)
def dotprod_to_dist(G):
    return np.sqrt(dotprod_to_dist2(G))

@jit(nopython=True, cache=True, parallel=True)
def series_xy_logk_numba(series=np.array([[]]), scale=1, npast=1, nfuture=1, decayRatio = 1, concat_valid_map = None, kernel_params_0 = None, kernel_params_1 = None, localdiff = 0):

    # STILL USES THE OLD DECAY RATIO API
    
    factor_r_past = 1 if npast<=1 else np.exp(np.log(decayRatio)/(npast-1.))
    factor_r_future = 1 if nfuture<=1 else np.exp(np.log(decayRatio)/(nfuture-1.))
    
    sum_r_past = 1
    r = 1
    for t in range(npast-2,-1,-1):
        r *= factor_r_past
        sum_r_past += r
    sum_past_factor = kernel_params_0 / (sum_r_past * scale**kernel_params_1)
    sum_r_future = 1
    r = 1
    for t in range(npast+1,npast+nfuture):
        r *= factor_r_future
        sum_r_future += r
    sum_future_factor = kernel_params_0 / (sum_r_future * scale**kernel_params_1)
    
    
    # The job done for each entry in the matrix
    # computes sum of sq diff for past (sx) and future (sy) sequences
    def sxy_logk(i,j):
            
        # from index of (past,future) pairs to index in data series
        i = concat_valid_map[i]
        j = concat_valid_map[j]

        delta = np.zeros(series.shape[1])
        if localdiff==1:
            # weighted average over each past series
            # diff of these => weighted avg of diffs
            r = 1
            for t in range(npast):
                d = series[i-t,:] - series[j-t,:]
                delta += d * r
                r *= factor_r_past
            delta /= sum_r_past
        elif localdiff==2:
            # value of the "present"
            delta = series[i,:] - series[j,:]
        
        r = 1
        sumx = 0
        for t in range(npast):
            d = series[i-t,:] - series[j-t,:]
            d = d - delta
            ds = np.sum(d*d)
            if kernel_params_1!=2:
                ds = np.power(ds, 0.5*kernel_params_1)
            sumx += ds * r
            r *= factor_r_past
        
        r = 1
        sumy = 0
        for t in range(nfuture):
            d = series[i+1+t,:] - series[j+1+t,:]
            d = d - delta
            ds = np.sum(d*d)
            if kernel_params_1!=2:
                ds = np.power(ds, 0.5*kernel_params_1)
            sumy += ds * r
            r *= factor_r_future
        
        return sumx * sum_past_factor, sumy * sum_future_factor


    # n is the number of valid (past, future) pairs
    n = len(concat_valid_map)
    sx = np.zeros((n,n))
    sy = np.zeros((n,n))
    
    # Triangular indexing, folded
    # x
    # y y     =>    z z z y y
    # z z z         w w w w x
    # w w w w
    # outer loops can now be parallelized - all have about the same duration
    m = n if n%2 == 1 else n-1
    for k in prange((n+1)//2,n):
        for l in range(k):
            i=k
            j=l
            sumx, sumy = sxy_logk(i, j)
            sx[i,j] = sumx
            sx[j,i] = sumx
            sy[i,j] = sumy
            sy[j,i] = sumy
            
        for l in range(k,m):
            i = m - k
            j = m-1-l
            sumx, sumy = sxy_logk(i, j)
            sx[i,j] = sumx
            sx[j,i] = sumx
            sy[i,j] = sumy
            sy[j,i] = sumy
        
    return sx, sy

@cuda.jit
def series_xy_logk_numba_cuda_k(num_folded_rows, num_folded_cols, kbase, sx, sy, concat_valid_map, series, npast, nfuture, factor_r_past, factor_r_future, sum_past_factor, sum_future_factor, kernel_params_0, kernel_params_1):

    # See comments in the other @cuda.jit function for the 'sum' rule below
    # This one was adapted later
    r, c = cuda.grid(2)
    if r>=num_folded_rows or c>=num_folded_cols: 
        return
    
    k = r + kbase     # for k in prange((n+1)//2,n):
    if c < k:         # for l in range(k)
        si = k
        sj = c
    else:             # for l in range(k,m)
        si = num_folded_cols - k
        sj = num_folded_cols - 1 - c
            
    # from index of (past,future) pairs to index in data series
    i = concat_valid_map[si]
    j = concat_valid_map[sj]
    
    r = 1
    sumx = 0
    for t in range(npast):
        # delta was used only for localdiff
        dsq = 0
        for sc in range(series.shape[1]):
            dsc = series[i-t,sc] - series[j-t,sc] #- delta[sc]
            dsq += dsc*dsc
        if kernel_params_1!=2:
            dsq = math.pow(dsq, 0.5*kernel_params_1)
        sumx += dsq * r
        r *= factor_r_past
    
    r = 1
    sumy = 0
    for t in range(nfuture):
        dsq = 0
        for sc in range(series.shape[1]):
            dsc = series[i+1+t,sc] - series[j+1+t,sc] #- delta[sc]
            dsq += dsc*dsc
        if kernel_params_1!=2:
            dsq = math.pow(dsq, 0.5*kernel_params_1)
        sumy += dsq * r
        r *= factor_r_future
    
    sumx *= sum_past_factor
    sumy *= sum_future_factor
    
    sx[si,sj] = sumx
    sx[sj,si] = sumx
    sy[si,sj] = sumy
    sy[sj,si] = sumy
    


def series_xy_logk_numba_cuda(series=np.array([[]]), scale=1, npast=1, nfuture=1, decayRatio = 1, concat_valid_map = None, kernel_params_0 = None, kernel_params_1 = None, localdiff = 0):

    import cupy as cp

    # STILL USES THE OLD DECAY RATIO API
    
    factor_r_past = 1 if npast<=1 else np.exp(np.log(decayRatio)/(npast-1.))
    factor_r_future = 1 if nfuture<=1 else np.exp(np.log(decayRatio)/(nfuture-1.))
    
    sum_r_past = 1
    r = 1
    for t in range(npast-2,-1,-1):
        r *= factor_r_past
        sum_r_past += r
    sum_past_factor = kernel_params_0 / (sum_r_past * scale**kernel_params_1)
    sum_r_future = 1
    r = 1
    for t in range(npast+1,npast+nfuture):
        r *= factor_r_future
        sum_r_future += r
    sum_future_factor = kernel_params_0 / (sum_r_future * scale**kernel_params_1)

    # n is the number of valid (past, future) pairs
    n = len(concat_valid_map)
    sx = cp.zeros((n,n))
    sy = cp.zeros((n,n))

    
    kbase = (n+1)//2
    num_folded_rows = n - kbase
    num_folded_cols = n if n%2 == 1 else n-1
    # https://docs.nvidia.com/cuda/cuda-c-programming-guide/
    # https://numba.readthedocs.io/en/stable/cuda/kernels.html
    threadsperblock = (16,16) # common choice according to cuda doc
    import math
    blockspergrid_x = math.ceil(num_folded_rows / threadsperblock[0])
    blockspergrid_y = math.ceil(num_folded_cols / threadsperblock[1])
    blockspergrid = (blockspergrid_x, blockspergrid_y)

    series_xy_logk_numba_cuda_k[blockspergrid, threadsperblock](num_folded_rows, num_folded_cols, kbase, sx, sy, concat_valid_map, series, npast, nfuture, factor_r_past, factor_r_future, sum_past_factor, sum_future_factor, kernel_params_0, kernel_params_1)
    
    return sx, sy



def series_xy_logk(series, scale, npast, nfuture, decay = False, concat_valid_map = None, localdiff = 0):

    # Switch to the OLD API. TODO: use a decayProfile parameter
    if decay:
        decayRatio = 0.01
    else:
        decayRatio = 1.

    xp = np
    
    # These lines are not supported by Numba
    try:
        # Use cupy if any series is on the GPU
        import cupy as cp
        if isinstance(series,list) or isinstance(series,tuple):
            for ser in series:
                if cp.get_array_module(ser) is cp:
                    xp = cp
                    break
            if xp is cp:
                # TODO: may send error if one is not GPU?
                series = cp.concatenate(series, axis=0)
                
        elif cp.get_array_module(series) is cp:
            xp = cp
        
        if xp is cp:
            # no-op if already on GPU
            if concat_valid_map is not None:
                concat_valid_map = cp.asarray(concat_valid_map)
            if localdiff>0:
                print("Error: Localdiff is not implemented on GPU. Using localdiff=0")
                localdiff = 0

            if series.ndim==1:
                series = series[:,None]

    except:
        pass

    if xp is not np:
        return series_xy_logk_numba_cuda(series, scale, npast, nfuture, decayRatio, concat_valid_map, __kernel_params__[0], __kernel_params__[1], localdiff)
    
    if isinstance(series,list) or isinstance(series,tuple):
        # Concatenate all the series and use the global validity index below
        series = np.concatenate(series, axis=0)
    
    if series.ndim==1:
        series = series[:,None]

    return series_xy_logk_numba(series, scale, npast, nfuture, decayRatio, concat_valid_map, __kernel_params__[0], __kernel_params__[1], localdiff)




@jit(nopython=True, cache=True, parallel=True)
def series_xy_k_sum_numba(series=np.array([[]]), scale=1, npast=1, nfuture=1, decay = False, concat_valid_map = None, kernel_params_0 = None, kernel_params_1 = None, localdiff = 0):
    
    w_past = np.ones(npast)
    w_future = np.ones(nfuture)
    sum_w_past = npast
    sum_w_future = nfuture
    if decay:
        for i in range(npast):
            w_past[i] = (npast-i) / npast
        for i in range(nfuture):
            w_future[i] = (nfuture-i) / nfuture
        sum_w_past = (npast+1)/2
        sum_w_future = (nfuture+1)/2
    
    # The job done for each entry in the matrix
    # computes sum of sq diff for past (sx) and future (sy) sequences
    def sxy_k(i,j):
            
        # from index of (past,future) pairs to index in data series
        i = concat_valid_map[i]
        j = concat_valid_map[j]

        delta = np.zeros(series.shape[1])
        if localdiff==1:
            # weighted average over each past series
            # diff of these => weighted avg of diffs
            for t in range(npast):
                w = w_past[t]
                d = series[i-t,:] - series[j-t,:]
                delta = delta + d * w
            delta /= sum_w_past
        elif localdiff==2:
            # value of the "present"
            delta = series[i,:] - series[j,:]
        
        sumx = 0.
        for t in range(npast):
            d = series[i-t,:] - series[j-t,:]
            d = (d - delta) / scale
            dsq = np.sum(d*d)
            if kernel_params_1!=2:
                dsq = np.power(dsq, 0.5*kernel_params_1)
            sumx += np.exp(dsq * kernel_params_0) * w_past[t]
        
        sumy = 0.
        for t in range(nfuture):
            d = series[i+1+t,:] - series[j+1+t,:]
            d = (d - delta) / scale
            dsq = np.sum(d*d)
            if kernel_params_1!=2:
                dsq = np.power(dsq, 0.5*kernel_params_1)
            sumy += np.exp(dsq * kernel_params_0) * w_future[t]
        
        sumx /= sum_w_past
        sumy /= sum_w_future
        return sumx, sumy


    # n is the number of valid (past, future) pairs
    n = len(concat_valid_map)
    sx = np.ones((n,n))
    sy = np.ones((n,n))
    
    # Triangular indexing, folded
    # x
    # y y     =>    z z z y y
    # z z z         w w w w x
    # w w w w
    # outer loops can now be parallelized - all have about the same duration
    m = n if n%2 == 1 else n-1
    for k in prange((n+1)//2,n):
        for l in range(k):
            i=k
            j=l
            sumx, sumy = sxy_k(i, j)
            sx[i,j] = sumx
            sx[j,i] = sumx
            sy[i,j] = sumy
            sy[j,i] = sumy
            
        for l in range(k,m):
            i = m - k
            j = m-1-l
            sumx, sumy = sxy_k(i, j)
            sx[i,j] = sumx
            sx[j,i] = sumx
            sy[i,j] = sumy
            sy[j,i] = sumy
        
    return sx, sy


import math

# The job done for each entry in the matrix
# computes sum of sq diff for past (sx) and future (sy) sequences
@cuda.jit
def sxy_k_cuda(num_folded_rows, num_folded_cols, kbase, deltazeros, sx, sy, concat_valid_map, series, localdiff, npast, nfuture, w_past, w_future, sum_w_past, sum_w_future, scale, kernel_params_0, kernel_params_1):
    
    r, c = cuda.grid(2)
    if r>=num_folded_rows or c>=num_folded_cols: 
        return


    # Triangular indexing, folded
    # x
    # y y     =>    z z z y y    of size num_folded_rows x num_folded_cols
    # z z z         w w w w x
    # w w w w
    
    """ base code for looping over this function
    for k in prange((n+1)//2,n):
        for l in range(k):
            i=k
            j=l
            sumx, sumy = sxy_k(i, j)
            sx[i,j] = sumx
            sx[j,i] = sumx
            sy[i,j] = sumy
            sy[j,i] = sumy
            
        for l in range(k,m):
            i = m - k
            j = m-1-l
            sumx, sumy = sxy_k(i, j)
            sx[i,j] = sumx
            sx[j,i] = sumx
            sy[i,j] = sumy
            sy[j,i] = sumy
    """
    
    # num_folded_rows is the prange extent in the base code
    # num_folded_cols is m in the base code
    # c is l in the base code
    # kbase is (n+1)//2
    
    k = r + kbase     # for k in prange((n+1)//2,n):
    if c < k:         # for l in range(k)
        si = k
        sj = c
    else:             # for l in range(k,m)
        si = num_folded_cols - k
        sj = num_folded_cols - 1 - c

    # Now we can implement the function body and set sx, sy below
    
    # in each case, the base code for sxy_k(i,j) started
    # by converting indices, so copy/paste this here
    
    # from index of (past,future) pairs to index in data series
    i = concat_valid_map[si]
    j = concat_valid_map[sj]

    """
    #delta = cp.zeros(series.shape[1])
    delta = deltazeros.copy()
    if localdiff==1:
        # weighted average over each past series
        # diff of these => weighted avg of diffs
        for t in range(npast):
            w = w_past[t]
            #d[sc] = series[i-t,:] - series[j-t,:]
            #delta = delta + d * w
            for sc in range(series.shape[1]):
                delta[sc] += (series[i-t,sc] - series[j-t,sc]) * w
        #delta /= sum_w_past
        for sc in range(series.shape[1]):
            delta[sc] /= sum_w_past
    elif localdiff==2:
        # value of the "present"
        #delta = series[i,:] - series[j,:]
        for sc in range(series.shape[1]):
            delta[sc] = series[i,sc] - series[j,sc]
    """
    
    sumx = 0.
    for t in range(npast):
        #d = series[i-t,:] - series[j-t,:]
        #d = (d - delta) / scale
        #dsq = cp.sum(d*d)
        dsq = 0
        for sc in range(series.shape[1]):
            #dsc = (series[i-t,sc] - series[j-t,sc] - delta[sc]) / scale
            dsc = (series[i-t,sc] - series[j-t,sc]) / scale
            dsq += dsc*dsc
            
        if kernel_params_1!=2:
            dsq = math.pow(dsq, 0.5*kernel_params_1)
        sumx += math.exp(dsq * kernel_params_0) * w_past[t]
    
    sumy = 0.
    for t in range(nfuture):
        #d = series[i+1+t,:] - series[j+1+t,:]
        #d = (d - delta) / scale
        #dsq = cp.sum(d*d)
        dsq = 0
        for sc in range(series.shape[1]):
            #dsc = (series[i+1+t,sc] - series[j+1+t,sc] - delta[sc]) / scale
            dsc = (series[i+1+t,sc] - series[j+1+t,sc]) / scale
            dsq += dsc*dsc
            
        if kernel_params_1!=2:
            dsq = math.pow(dsq, 0.5*kernel_params_1)
        sumy += math.exp(dsq * kernel_params_0) * w_future[t]
    
    sumx /= sum_w_past
    sumy /= sum_w_future
    
    # Remains to be done from the above loop conversion
    sx[si,sj] = sumx
    sx[sj,si] = sumx
    sy[si,sj] = sumy
    sy[sj,si] = sumy


def series_xy_k_sum_numba_cuda(series, scale=1, npast=1, nfuture=1, decay = False, concat_valid_map = None, kernel_params_0 = None, kernel_params_1 = None, localdiff = 0):
    
    import cupy as cp
    
    w_past = cp.ones(npast)
    w_future = cp.ones(nfuture)
    sum_w_past = npast
    sum_w_future = nfuture
    if decay:
        for i in range(npast):
            w_past[i] = (npast-i) / npast
        for i in range(nfuture):
            w_future[i] = (nfuture-i) / nfuture
        sum_w_past = (npast+1)/2
        sum_w_future = (nfuture+1)/2

    # n is the number of valid (past, future) pairs
    n = len(concat_valid_map)
    sx = cp.ones((n,n))
    sy = cp.ones((n,n))
    deltazeros = cp.zeros(series.shape[1])
    
    kbase = (n+1)//2
    num_folded_rows = n - kbase
    num_folded_cols = n if n%2 == 1 else n-1
    # https://docs.nvidia.com/cuda/cuda-c-programming-guide/
    # https://numba.readthedocs.io/en/stable/cuda/kernels.html
    threadsperblock = (16,16) # common choice according to cuda doc
    import math
    blockspergrid_x = math.ceil(num_folded_rows / threadsperblock[0])
    blockspergrid_y = math.ceil(num_folded_cols / threadsperblock[1])
    blockspergrid = (blockspergrid_x, blockspergrid_y)
    sxy_k_cuda[blockspergrid, threadsperblock](num_folded_rows, num_folded_cols, kbase, deltazeros, sx, sy, concat_valid_map, series, localdiff, npast, nfuture, w_past, w_future, sum_w_past, sum_w_future, scale, kernel_params_0, kernel_params_1)
    
    return sx, sy



def series_xy_k_sum(series, scale, npast, nfuture, decay = False, concat_valid_map = None, localdiff = 0):
    
    # These lines are not supported by Numba
    
    try:
        # Use cupy if any series is on the GPU
        import cupy as cp
        xp = np
        if isinstance(series,list) or isinstance(series,tuple):
            for ser in series:
                if cp.get_array_module(ser) is cp:
                    xp = cp
                    break
            if xp is cp:
                # TODO: may send error if one is not GPU?
                series = cp.concatenate(series, axis=0)
                
        elif cp.get_array_module(series) is cp:
            xp = cp
        
        if xp is cp:
            # no-op if already on GPU
            if concat_valid_map is not None:
                concat_valid_map = cp.asarray(concat_valid_map)
            if localdiff>0:
                print("Error: Localdiff is not implemented on GPU. Using localdiff=0")

            if series.ndim==1:
                series = series[:,None]

            return series_xy_k_sum_numba_cuda(series, scale, npast, nfuture, decay, concat_valid_map, __kernel_params__[0], __kernel_params__[1], localdiff)

    except:
        pass
    
    if isinstance(series,list) or isinstance(series,tuple):
        # Concatenate all the series and use the global validity index below
        series = np.concatenate(series, axis=0)
        
    if series.ndim==1:
        series = series[:,None]

    return series_xy_k_sum_numba(series, scale, npast, nfuture, decay, concat_valid_map, __kernel_params__[0], __kernel_params__[1], localdiff)




@jit(nopython=True, cache=True, parallel=True)
def parallel_exp(mat, scale):
    N = mat.shape[0]
    invscale = 1./scale
    # outer loops can be parallelized - all have about the same duration
    for k in prange(N):
        mat[k,:] = np.exp(mat[k,:] * invscale)

@jit(nopython=True, cache=True, parallel=True)
def parallel_exp_lowtri(mat, scale):
    N = mat.shape[0]
    M = N if N%2 == 1 else N-1
    invscale = 1./scale
        
    # outer loops can be parallelized - all have about the same duration
    for k in prange((N+1)//2,N):
        for l in range(k):
            i=k
            j=l
            e = np.exp(mat[i,j] * invscale)
            mat[i,j] = e
            mat[j,i] = e
            
        for l in range(k,M):
            i = M - k
            j = M-1-l
            e = np.exp(mat[i,j] * invscale)
            mat[i,j] = e
            mat[j,i] = e

    for d in prange(N):
        mat[d,d] = np.exp(mat[d,d] * invscale)

@jit(nopython=True, cache=True, parallel=True)
def parallel_add_lowtri(total, mat):
    """
    WARNING: ONLY ADDS THE LOWER PART
    """
    N = mat.shape[0]
    M = N if N%2 == 1 else N-1
        
    # outer loops can be parallelized - all have about the same duration
    for k in prange((N+1)//2,N):
        for l in range(k):
            i=k
            j=l
            total[i,j] += mat[i,j]
            
        for l in range(k,M):
            i = M - k
            j = M-1-l
            total[i,j] += mat[i,j]

    for d in prange(N):
        total[d,d] += mat[d,d]

@cuda.jit
def parallel_exp_lowtri_cuda_k(num_folded_rows, num_folded_cols, kbase, mat, invscale):
    # See comments in the other @cuda.jit functions below
    r, c = cuda.grid(2)
    if r>=num_folded_rows or c>=num_folded_cols: 
        return
    
    k = r + kbase     # for k in prange((n+1)//2,n):
    if c < k:         # for l in range(k)
        i = k
        j = c
    else:             # for l in range(k,m)
        i = num_folded_cols - k
        j = num_folded_cols - 1 - c

    e = math.exp(mat[i,j] * invscale)
    mat[i,j] = e
    mat[j,i] = e

@cuda.jit
def parallel_exp_lowtri_cuda_diag_k(n, mat, invscale):
    i = cuda.grid(1)
    if i>=n:
        return
    e = math.exp(mat[i,i] * invscale)
    mat[i,i] = e


def parallel_exp_lowtri_cuda(mat, scale):
    # See comments in the other cuda wrappers below
    n = mat.shape[0]
    kbase = (n+1)//2
    num_folded_rows = n - kbase
    num_folded_cols = n if n%2 == 1 else n-1
    threadsperblock = (16,16) # common choice according to cuda doc
    blockspergrid_x = math.ceil(num_folded_rows / threadsperblock[0])
    blockspergrid_y = math.ceil(num_folded_cols / threadsperblock[1])
    blockspergrid = (blockspergrid_x, blockspergrid_y)
    parallel_exp_lowtri_cuda_k[blockspergrid, threadsperblock](num_folded_rows, num_folded_cols, kbase, mat, 1/scale)
    threadsperblock = 16
    blockspergrid = math.ceil(n / threadsperblock)
    parallel_exp_lowtri_cuda_diag_k[blockspergrid, threadsperblock](n, mat, 1/scale)
    return mat


def gaussianSimilarity(serA, serB=None, bandwidth=1):
    """
    Computes similarity matrix using the Gaussian kernel.
    S_ij = exp(-0.5 * | serA[i,:] - serB[i,:] |**2 / bandwidth**2)

    Parameters
    ----------
    
    serA : array of size N, D
        N = number of samples
        D = dimension of each sample
        
    serB : array of size N, D, optional
        The default None uses serA for serB
        
    bandwidth : real, optional
        gaussian kernel bandwidth. The default is 1.

    Returns
    -------
    
    S : array of size N, N
        The similarity matrix

    """

    import scipy.spatial.distance
    if serB is None:    
        G = scipy.spatial.distance.squareform(np.exp((-0.5/bandwidth**2)*scipy.spatial.distance.pdist(serA, 'sqeuclidean')))
        np.fill_diagonal(G, 1)
        return G
    
    return np.exp((-0.5/bandwidth**2)*scipy.spatial.distance.cdist(serA, serB, 'sqeuclidean'))
    
    
    # This code produces slightly different results
    # Very small numerical innacuracies, but that make
    # the matrix inversion explode. Strange
    """
    sinprodA = np.sum(np.multiply(serA, serA), axis=1)[:, None]
    sinprodB = sinprodA if serB is None else np.sum(np.multiply(serB, serB), axis=1)[:, None]
    soutprod = serA @ serA.T if serB is None else serA @ serB.T
    s = sinprodA + sinprodB.T - 2 * soutprod
    if serB is None:
        parallel_exp_lowtri(s, -2*bandwidth**2)
    else:
        parallel_exp(s, -2*bandwidth**2)
    return s
    """

def series_Gxy(series, scale, npast, nfuture, decay = False, qcflags = None, localdiff = 0, take2skipX = 0, kernelAggregation = 'product', weights = None):
    """
    Compute the past and the future Gram matrices, using the kernel set by set_kernel (defaults to Gaussian)
    
    Arguments:
        series: array or list. Each array is a time-contiguous block of data, with as many rows as time samples and each column is a feature. NaN values are allowed and detected. 
        A list may be provided, in which case each list entry is a different data source. Heterogenous data are supported, so each list entry can be of completely different type. However, each list entry must have the same number of values.
        Each first-level list entry can itself be a second-level list instead of a unique array. In that case, these second-level lists gather data acquired in different time-continuous blocks. These blocks must be consistent across heterogenous data sources: at each time, each data source must be acquired, even though some sensors may fail (NaN values).
        So, in order to enter a unique data source, but with multiple time blocks, the syntax is to use a double list : [[data1, data2, ...]] . There, the first-level list specifies a unique data source, and the second level list specifies the contiguous data blocks.
        Multiple sources and multiple blocks read: [[src1_block1, src1_block2], [src2_block1, src2_block2]. Same-numbered blocks from different sources must be of the same length, but different blocks may have different lengths.

        scale: float, or list of floats. If a single float is provided it applies to all series. A list of floats specifies a different scale for each data source.
         
        npast: take that consecutive number of samples for building an observed past. If a single integer is provided it applies to all series. Different values can be provided for different data sources (=each series). It is assumed that data samples are measured at matching times for each data source, hence the first causal state can only be computed after the-largest-npast values have been observed.
        
        nfuture: take that consecutive number of samples, just after the past samples, for building an observed future. If a single integer is provided it applies to all series. Similarly as for npast, the largest nfuture value sets the time of the last computable causal state.
        
        decay: NEW API: bool (default is False) or list of bool for each data source. Decay the influence of furthermost kernel values in the past or future compared to the present. Due to this reduced causal influence, the past/future duration may need to be increased. See also decayProfile. API Change: OLD Behavior for updating scripts: this used to be the ratio for the furthermost weight compared to the immediate past/future. Used to default to 1 = no decay, while the new default is False = no decay.
        
        qcflags : quality control flags. <= 0 means invalid data. optional, but if present there must be one qcflag per sample
        
        localdiff: how to compute the differences between the sequences, pasts and futures. The default is 0 = just take the difference. This assumes that what matters are the series absolute values. localdiff = 1 removes the average (weigthed using decay) of each sequence before taking their differences. This may be useful for eliminating a slow drift due to the sensor itself, which should be removed from data. localdiff = 2 means that the "present" value is used as reference. This assumes that what matters are relative variations around the "present" value. Warning: this gives equal importance to all such variations. For example, a fluctuation of -3°C in summer (from ~25°C) is not the same as a fluctuation of -3°C in winter (which may very well cross the freezing point).
        
        take2skipX : int, optional (default 0)
            performs a special kind of subsampling: two consecutive samples are retained, X samples are discarded, then this (X+2) sequence repeats. This scheme is designed to preserve consecutive entries for building a shift operator while still allowing subsampling of very large series. The classic subsampling scheme (take one out of X) can also be applied a priori to the original series, but it is equivalent to a (bad) low-pass filtering. Then, the shift operator would be computed on consecutive entries of the subsampled series, hence at a different time scale. The take2skipX allows to still work on the original time scale. Both can be combined (after appropriate low-pass filtering).
        
        kernelAggregation: 'sum', 'product'. How to aggregate kernels acting on multiple sources and for combining past/future values in sequences. Default is to take a product of kernels, but a direct sum of Hilbert spaces is also possible.
        
        weights: for each data source, how much that source contributes to the total. Weights should be positive and will be normalized by sum(weights). The default None corresponds to setting equal weights, hence w = 1/nsources. This parameter must otherwise be an array of the same length as there are data sources
        
        decayProfile (TODO): Sets how the past values are combined. Defaults to 'power', which is the only possibility when kernelAggregation=='product', and the value of the furthermost past influence is set to 1/100th of the present (TODO=allow setting this). Other values for direct sums are 'linear' (default) and 'exponential' (not implemented yet). For now, this parameter is not implemented and only the default decay profiles are implemented.
        
    Returns:
        dpast, dfuture: two square matrices of size N, between each observed sequences of a past/future sample pairs.
        If lists are passed as arguments, each list entry is considered its own
        series object, with a matching scale. The data are then combined into
        a unique Gram matrix

        idxmap : N array of int
            Returns the indices of the valid (not NaN) and contiguous (past,future) pairs. For each returned row/col index of the dpast, dfuture matrices, the idxmap specifies what data in the original series that (past,future) reference pair refers to.
    
    """




    # TODO: Dask auto-detect/use similar to GPU
    # 1. Add a chuncks=(x,y) argument, with defaults None
    #    => if chuncks is not None: this detects that we want to use Dask
    # 2. detect here if the series arrays are actually dask arrays
    #    => if so, auto-set the chunks from the series internal chunks
    # 3. Create Gx, Gy as dask arrays
    # 4. Call ufuncs (coded in external module/py file)
    #    to replace the triangle computations above
    #    see https://discourse.pangeo.io/t/numba-guvectorize-and-distributed/758/3
    # 5. Make it work with Cupy : each dask task uses a local GPU and the
    #    whole array is distributed









    # the series themselves could also be a numpy array or a dask array
    series_list = series if isinstance(series,list) or isinstance(series,tuple) else [ series ]
    nseries = len(series_list)
    # https://stackoverflow.com/questions/16807011/python-how-to-identify-if-a-variable-is-an-array-or-a-scalar
    # Useful trick to check for "__len__": this covers both lists, numpy arrays and dask arrays
    scales_list = scale if hasattr(scale, "__len__") else [ scale ] * nseries
    npasts_list = npast if hasattr(npast, "__len__") else [ npast ] * nseries
    nfutures_list = nfuture if hasattr(nfuture, "__len__") else [ nfuture ] * nseries
    decays_list = decay if hasattr(decay, "__len__") else [ decay ] * nseries
    localdiff_list = localdiff if hasattr(localdiff, "__len__") else [ localdiff ] * nseries

    for dec in decays_list:    
        if type(dec) is not bool:
            print("Error: API change, please update your script. 'decay' is now a bool (True/False) or a list of bool values")
            raise RuntimeError("Error: API change, please update your script. 'decay' is now a bool (True/False) or a list of bool values")

    if qcflags is None:
        qcflags_list = [ None for ser in series_list ]
    else: 
        qcflags_list = qcflags if isinstance(qcflags,list) or isinstance(qcflags,tuple) else [ qcflags ]
    
    if weights is None:
        weights_list = np.ones(nseries)
    elif callable(weights):
        weights_list = [None] * nseries
    else:
        weights_list = weights if hasattr(weights, "__len__") else [ weights ] * nseries
    sumw = np.sum(weights_list)
    
    index_map = compute_index_map_multiple_sources(series_list, npasts_list, nfutures_list, qcflags_list, take2skipX)

    
    # This is an attempt for user-defined aggregation schemes
    # not useful for now, but keep this idea as a comment

    # #  A fully callable aggregation scheme requires the full list of
    # #  kernel similarities, which is nsources times NxN big matrices for
    # #  both kx and ky. This may be much larger than available memory
    # if callable(kernelAggregation):
    #     pass
    
    # # An intermediate solution is to allow some flexibility in the scheme,
    # # but do it incrementally, keeping only the current aggregate and the
    # # new kx, ky. This is exactly what the sum rule does, with the weights 
    # # specifying the aggregation scheme. 'incremental' is a generalization.
    # if kernelAggregation=='incremental' and callable(weights):
    #     total_kx, total_ky = None, None
    #     for idx, (ser, sca, npa, nfu, dec, ldiff) in enumerate(zip(series_list, scales_list, npasts_list, nfutures_list, decays_list, localdiff_list)):
    #         kx, ky = series_xy_k_sum(ser, sca, npa, nfu, dec, index_map, ldiff)
    #         total_kx, total_ky = weights(total_kx, total_ky, kx, ky, idx)
    #     return total_kx, total_ky, index_map
    
    if kernelAggregation=='sum':
        total_kx, total_ky = None, None
        for ser, sca, npa, nfu, dec, ldiff, w in zip(series_list, scales_list, npasts_list, nfutures_list, decays_list, localdiff_list, weights_list):
            kx, ky = series_xy_k_sum(ser, sca, npa, nfu, dec, index_map, ldiff)
            if total_kx is None:
                total_kx, total_ky = kx * w, ky * w
            else:
                total_kx += kx * w
                total_ky += ky * w
        total_kx /= sumw
        total_ky /= sumw
        return total_kx, total_ky, index_map
  

    total_lx, total_ly = None, None
    
    xp = None
    
    for ser, sca, npa, nfu, dec, ldiff, w in zip(series_list, scales_list, npasts_list, nfutures_list, decays_list, localdiff_list, weights_list):
        
        # ser may itself be a list, this is handled by series_xy_logk
        lx, ly = series_xy_logk(ser, sca, npa, nfu, dec, index_map, ldiff)
        
        if xp is None:
            try:
                import cupy as cp
                if cp.get_array_module(lx) is cp:
                    xp = cp
                else:
                    xp = np
            except:
                xp = np
            
        if total_lx is None:
            total_lx, total_ly = lx * w, ly * w
        else:
            # This modifies the arguments - only adds the lower-triangular part
            if xp is np:
                parallel_add_lowtri(total_lx, lx * w)
                parallel_add_lowtri(total_ly, ly * w)
            else:
                # adding on GPU, without kernel, is efficient enough
                total_lx += lx * w
                total_ly += ly * w
    total_lx /= sumw
    total_ly /= sumw

    
    # This modifies the arguments - restores the upper part
    if xp is np:
        parallel_exp_lowtri(total_lx, nseries)
        parallel_exp_lowtri(total_ly, nseries)
    else:
        parallel_exp_lowtri_cuda(total_lx, nseries)
        parallel_exp_lowtri_cuda(total_ly, nseries)
        
    return total_lx, total_ly, index_map
    


def embed_states(Gx, Gy, eps = 1e-8, normalize = False, return_embedder = False, use_GPU = -1):
    """
    Compute a similarity matrix for the embedded causal states, seen as distributions P(Y|X), using the conditional mean embedding.
    
    Arguments:
        Gx: a similarity matrix of pasts X. Gx[i,j] = kernel_X(x_i, x_j) with kernel_X a reproducing kernel for X.
        Gy: a similarity matrix of futures Y.
        eps: amount of regularization. The theory requires a regularizer and shows that the regularized estimator is consistent in the limit of N -> infinity. In practice, using a too small regularizer may cause divergence, too large causes innaccuracies.
        normalize: whether to renormalize the returned similarity matrix, so that each entry along the diagonal is 1. This may sometimes improve numerical stability, but it is not quite correct. See below.
        use_GPU: -1 = auto (default), use GPU if either Gx or Gy are on GPU. 0 = force no GPU, 1 = force GPU
        return_embedder: whether to return an embedder object for new, unknown data. Default is False
        
    Returns:
        Gs: A similarity matrix between each causal states. Entries Gs[i,j] can be seen as inner products between states S_i and S_j. With normalized kernels, such as used in series_Gxy, then the kernel evaluated at each sample k(data, .) are lying on the unit sphere (i.e. k(data,data) = 1, which means their norm is also 1). But then, the distribution embeddings, conditional or not, are just weighted averages of these samples lying on the sphere. So they are barycenters of these points on the sphere, and these barycenters are _NOT_ themselves on the sphere. So, normalizing Gs to have 1s on the diagonal is technically wrong, but it may help for avoiding numerical innacuracies in eigenvalues decompositions so is left as an option. Also, theoretically, that matrix should be positive definite, and theoretically a regularizer is needed in most cases. If positive definiteness issues happens, try using a larger regularizer.

        embedder: (if return_embedder is True) A matrix for embedding new Kx kernel similarity vectors

    """
    
    try:
        # Use cupy if Gx or Gy is on the GPU, use GPU
        import cupy as cp
        if use_GPU==-1 and (cp.get_array_module(Gx) is cp or cp.get_array_module(Gy) is cp):
            use_GPU=1
        else:
            use_GPU=0
            
        if use_GPU==0:
            # force no GPU
            xp = np
            Gx = cp.asnumpy(Gx) # copy on CPU if not there already
            Gy = cp.asnumpy(Gy) # copy on CPU if not there already
        else:
            xp = cp
            Gx = cp.asarray(Gx) # copy on GPU if not there already
            Gy = cp.asarray(Gy) # copy on GPU if not there already

    except:
        xp = np
        if use_GPU==1:
            print("Warning: GPU is not available. Using CPU.")

    
    # Solve is good enough : produces 1/N entries with N = num dups
    Omega = xp.linalg.solve(Gx + xp.diag(xp.full(Gx.shape[0], eps)), Gx)
    #from . import nnls_impl
    #Omega = nnls_impl.nnls(Gx + np.diag(np.ones(Gx.shape[0])*eps), Gx)
    Omega = 0.5 * (Omega + Omega.T) # should not be needed
    

    """
    dists = np.zeros_like(Gx)
    for c in range(Gx.shape[1]):
        dists[:,c] = np.sqrt(np.clip(2 - 2 * Gx[:,c], 0, np.inf))
    Omega = dist_to_weights(dists, 1, 0.1, eps)
    """
    
    embedder = Omega.T @ Gy
    
    Gs = embedder @ Omega
    
    if normalize:
        # This normalization is especially useful for avoiding numerical
        # errors later in the Gs eigendecomposition. But Gs should already
        # have nearly 1s on the diagonal to start with. For embedding new
        # vectors, this refinement is probably overkill
        # TODO: tests if this is really necessary for the embedder,
        # knowing the distributions will be normalized later on.
        # Tests: actually, this impacts a bit the numerical resolution
        # when expressing the embedded Ks as a combination of Gs lines
        # option 1 : all the weights on the left side:
        # embedder /= Gs.diagonal()[:,None]
        # if applied to Gs, this would yield 1s on the diagonal
        # but some asymetry
        # Tests: this options yields some inaccuracies
        # option 2 : use the same normalization as GS:
        # but that normalisation is not applicable on the right
        # to new kvecs
        # Tests: this options yields some inaccuracies, but less than option 1
        d = 1./np.sqrt(Gs.diagonal())
        embedder *= d[:,None]
        # Now Gs
        Gs *= d[:,None] # np.diag(d) @ Gs
        Gs *= d[None,:] # Gs @ np.diag(d)
        Gs = 0.5 * (Gs + Gs.T)
        # just to be extra safe, avoid floating-point residual errors
        xp.fill_diagonal(Gs, 1)
        
        # Tests - this is also possible, using Gs after normalization, 
        # but does not improve the accuracy of the embedder
        #embedder = np.linalg.solve(Omega, Gs).T
        #embedder = 0.5 * (embedder + embedder.T)
    
    if return_embedder:
        return Gs, embedder
    return Gs



# The job done for each entry in the matrix
# computes sum of sq diff for past (sx) and future (sy) sequences
@jit(nopython=True, cache=True)
def sx_k_sum_internal_helper(i,j, new_series, ref_series, npast, scale, w_past, sum_w_past, kernel_params_0, kernel_params_1, localdiff):

    delta = np.zeros(ref_series.shape[1])
    if localdiff==1:
        # weighted average over each past series
        # diff of these => weighted avg of diffs
        for t in range(npast):
            w = w_past[t]
            d = ref_series[i-t,:] - new_series[j-t,:]
            delta = delta + d * w
        delta /= sum_w_past
    elif localdiff==2:
        # value of the "present"
        delta = ref_series[i,:] - new_series[j,:]
    
    sumx = 0.
    for t in range(npast):
        d = ref_series[i-t,:] - new_series[j-t,:]
        d = (d - delta) / scale
        dsq = np.sum(d*d)
        if kernel_params_1!=2:
            dsq = np.power(dsq, 0.5*kernel_params_1)
        sumx += np.exp(dsq * kernel_params_0) * w_past[t]
    
    sumx /= sum_w_past
    return sumx

    

@jit(nopython=True, cache=True)
def sx_logk_internal_helper(i,j, new_series, ref_series, npast, factor_r_past, sum_r_past, sum_past_factor, kernel_params_1, localdiff):
    # from index of (past,future) pairs to index in data series
    #i += npast-1  # ref
    #j += npast-1  # new

    delta = np.zeros(ref_series.shape[1])
    if localdiff==1:
        # weighted average over each past series
        # diff of these => weighted avg of diffs
        r = 1
        for t in range(npast):
            d = ref_series[i-t,:] - new_series[j-t,:]
            delta += d * r
            r *= factor_r_past
        delta /= sum_r_past
    elif localdiff==2:
        # value of the "present"
        delta = ref_series[i,:] - new_series[j,:]
    
    r = 1
    sumx = 0
    for t in range(npast):
        d = ref_series[i-t,:] - new_series[j-t,:]
        d = d - delta
        ds = np.sum(d*d)
        if kernel_params_1!=2:
            ds = np.power(ds, 0.5*kernel_params_1)
        sumx += ds * r
        r *= factor_r_past

    return sumx * sum_past_factor



@jit(nopython=True, cache=True, parallel=True)
def series_newx_k_sum_single_source_numba(new_series=np.array([[]]), new_series_map=np.array([]), ref_series=np.array([[]]), ref_series_map=np.array([]), scale=1, npast=1, decay = False, kernel_params_0 = -0.5, kernel_params_1 = 2.0, localdiff = 0):
    
    w_past = np.ones(npast)
    sum_w_past = npast
    if decay:
        for i in range(npast):
            w_past[i] = (npast-i) / npast
        sum_w_past = (npast+1)/2

    sx = np.zeros((len(ref_series_map),len(new_series_map)))

    nr = len(ref_series_map)
    ns = len(new_series_map)
    #for idx, i in enumerate(ref_series_map):
    #  for jdx, j in enumerate(new_series_map):
    for i_rs in prange(nr*ns):
        idx = i_rs // ns
        jdx = i_rs - idx * ns
        i = ref_series_map[idx]
        j = new_series_map[jdx]
        sx[idx,jdx] = sx_k_sum_internal_helper(i,j, new_series, ref_series, npast, scale, w_past, sum_w_past, kernel_params_0, kernel_params_1, localdiff)

    return sx



@jit(nopython=True, cache=True, parallel=True)
def series_newx_logk_single_source_numba(new_series=np.array([[]]), new_series_map=np.array([]), ref_series=np.array([[]]), ref_series_map=np.array([]), scale=1, npast=1, decayRatio = 1, kernel_params_0 = -0.5, kernel_params_1 = 2.0, localdiff = 0):
    """
    Compute the distances between each past sequence from the new series and the reference sequences
    
    new_series is a MxD array, where each row is a time sample
    
    The other arguments are the same as for series_xy_logk
    
    returns one row for each ref seq, one column for each new past sequence
    """
    
    # STILL USES THE OLD DECAY RATIO API
    
    factor_r_past = 1 if npast<=1 else np.exp(np.log(decayRatio)/(npast-1.))
    
    sum_r_past = 1
    r = 1
    for t in range(npast-2,-1,-1):
        r *= factor_r_past
        sum_r_past += r
    sum_past_factor = kernel_params_0 / (sum_r_past * scale**kernel_params_1)

    sx = np.zeros((len(ref_series_map),len(new_series_map)))

    nr = len(ref_series_map)
    ns = len(new_series_map)
    #for idx, i in enumerate(ref_series_map):
    #  for jdx, j in enumerate(new_series_map):
    for i_rs in prange(nr*ns):
        idx = i_rs // ns
        jdx = i_rs - idx * ns
        i = ref_series_map[idx]
        j = new_series_map[jdx]
        sx[idx,jdx] = sx_logk_internal_helper(i,j, new_series, ref_series, npast, factor_r_past, sum_r_past, sum_past_factor, kernel_params_1, localdiff)

    return sx


def series_newx_single_source(new_series, new_series_map, ref_series, ref_series_map, scale, npast, decay = False, localdiff = 0, kernelAggregation = 'product'):
    
    # These lines are not supported by Numba
    
    if isinstance(ref_series,list) or isinstance(ref_series,tuple):
        # Concatenate all the series and use the global validity index
        ref_series = np.concatenate(ref_series, axis=0)
    
    if isinstance(new_series,list) or isinstance(new_series,tuple):
        # Concatenate all the series and use the global validity index
        new_series = np.concatenate(new_series, axis=0)
    
    if ref_series.ndim==1:
        ref_series = ref_series[:,None]
        
    if new_series.ndim==1:
        new_series = new_series[:,None]

    if kernelAggregation == 'sum':
        return series_newx_k_sum_single_source_numba(new_series, new_series_map, ref_series, ref_series_map, scale, npast, decay, __kernel_params__[0], __kernel_params__[1], localdiff)
    
    else:
        # Switch to the OLD API. TODO: use a decayProfile parameter
        if decay:
            decayRatio = 0.01
        else:
            decayRatio = 1.
        return series_newx_logk_single_source_numba(new_series, new_series_map, ref_series, ref_series_map, scale, npast, decayRatio, __kernel_params__[0], __kernel_params__[1], localdiff)


def series_newKx(new_series, new_series_map, ref_series, ref_series_map, scale, npast, decay = False, localdiff = 0, kernelAggregation = 'product', weights = None):
    
    """
    Computes the kernel vector evaluations between each past sequence from the new series and the reference sequences. 
    
    This is similar to series_Gxy in that it computes a similarity matrix (or the log of it), but here the similarity is computed with respect to new data values, and only for Gx (past sequences)
    
    The new_series_map parameter can be computed using compute_index_map, or set to None to be automatically computed. See series_Gxy and compute_index_map for other arguments/doc. 
    
    """
    
    new_series_list = new_series if isinstance(new_series,list) or isinstance(new_series,tuple) else [ new_series ]
    ref_series_list = ref_series if isinstance(ref_series,list) or isinstance(ref_series,tuple) else [ ref_series ]
    nseries = len(ref_series_list)
    scales_list = scale if hasattr(scale, "__len__") else [ scale ] * nseries
    npasts_list = npast if hasattr(npast, "__len__") else [ npast ] * nseries
    decays_list = decay if hasattr(decay, "__len__") else [ decay ] * nseries
    localdiff_list = localdiff if hasattr(localdiff, "__len__") else [ localdiff ] * nseries
    
    if weights is None:
        weights_list = np.ones(nseries)
    else:
        weights_list = weights if hasattr(weights, "__len__") else [ weights ] * nseries
    sumw = np.sum(weights_list)

    for dec in decays_list:    
        if type(dec) is not bool:
            print("Error: API change, please update your script. 'decay' is now a bool (True/False) or a list of bool values")
            raise RuntimeError("Error: API change, please update your script. 'decay' is now a bool (True/False) or a list of bool values")

    if new_series_map is None:
        new_series_map = compute_index_map_multiple_sources(new_series_list, npasts_list)

    if ref_series_map is None:
        ref_series_map = compute_index_map_multiple_sources(ref_series_list, npasts_list)

    total_korl_x = None
    for nser, rser, sca, npa, dec, ldiff, w in zip(new_series_list, ref_series_list, scales_list, npasts_list, decays_list, localdiff_list, weights_list):
        korl_x = series_newx_single_source(nser, new_series_map, rser, ref_series_map, sca, npa, dec, ldiff, kernelAggregation)
        if total_korl_x is None:
            total_korl_x = korl_x * w
        else:
            total_korl_x += korl_x * w
    total_korl_x /= sumw
    
    if kernelAggregation=='sum':
        return total_korl_x
    
    # product
    return np.exp(total_korl_x)


def embed_Kx(Kx, Gx, embedder, eps = 1e-8):
    """
    Construct a similarity vector in state space, from a similarity vector in X space and using the embbeder returned by embed_states
    
    Arguments:
        Kx: a similarity vector in X space, such as returned by series_newKx.
        Gx: a similarity matrix of pasts X.
        embedder: The embedder returned by embed_states
        eps: amount of regularization. See embed_states

    Returns:
        Ks: the similarity vector, in state space
    """
    
    #omega = np.linalg.solve(Gx + np.diag(np.ones(Gx.shape[0])*eps), Kx)
    from . import nnls_impl
    omega = nnls_impl.nnls(Gx + np.diag(np.ones(Gx.shape[0])*eps), Kx)
    #omega /= np.sum(omega)
    return (embedder @ omega) #/ (omega.T @ Gy @ omega)[0,0]


def dist_to_weights(d, knn=1, factor=None, eps=1e-8):
    """
    Compute weights from distances. The weights sum to 1 and can be used as probabilities. This is softmax, using local bandwidths set by nearest neighbors, hence also a balloon estimator.

    Parameters
    ----------
    d : array, NxM
        The distances are column vectors, specifying the distance of a new point to N reference points. M such new points can be specified
        
    knn : float, optional
        The number of neighbors to seek for each point, based on distances to other points. The distance of the k-th neighbor is used as a bandwidth. The default is 1: this yields 0 bandwidth for new points falling on reference points. In turn, this is converted to a weight of 1 for that reference and 0 for others. This is the expected behavior for consistency when coordinates are computed for new points using these weights, so that new coordinates match reference coordinates for points in the training data set. This might present a risk of overfitting in some situations, in which case increasing knn may help.
        
    factor : float, optional
        Additional factor for the bandwidth, in order to more or less focus the resulting distribution around reference points. The default is None. See the notes. if the factor is set to None, then the standard deviation of the neighbors is used as a factor
        
    eps : float, optional
        Distance below which points are considered equal. The default is 1e-8.

    Returns
    -------
    w : array, NxM
        weights for each of the N reference points, for each of the M new points
        
    Notes
    -----
    Each weight is computed using exp(-d/bw), where d is the distance to the reference point and bw a bandwidth. Then, all weights are normalized to sum to 1. This is the classical softmax formula.
    The bandwidth is bw=dknn*factor, which by default the distance to the nearest neighbor (knn=1) and if factor=1. Hence, the resulting probability distribution is also a baloon estimator, dependant on the local density at each of the M points.
    

    """
    
    # avoid division by 0. All very close points below precision are 
    # considered equal and will be given equal proba
    d[d<eps] = eps
    
    # partition find the k nearest in a fast way (do not sort the entire array)
    # The k-th distance is at its right place, the other distances are shuffled
    # Problem: what if k=1, 1 neighor in partition[1] but partition[0] below ???
    #bandwidth = np.partition(d, knn, axis=0)[knn-1:knn,:]
    #bandwidth = np.mean(np.partition(d, knn, axis=0)[:knn,:])
    ds = np.sort(d, axis=0)
    #bandwidth = ds[knn-1,0]
    bandwidth = np.mean(ds[:knn,:], axis=0)

    # debug
    #ds = np.sort(d, axis=0)
    #print(ds[:5,:5])
    # hack to make the weight of the second neighbor twice as small as the first
    #factor = np.mean(ds[1,:] - ds[0,:]) / np.log(2)
    #if factor<eps: factor=eps
    #print(factor)

    if factor is None:
        factor = np.std(d, axis=0)

    # proba formula = softmax adaptation
    # This is also a balloon estimator using the k-th neighbor as bandwidth
    w = np.exp(-d/(factor*bandwidth))
    wsum = np.sum(w,axis=0)[None,:]
    #if wsum==0:
    #    return np.ones(len(w)) / len(w)
    return w / wsum

def series_newcoords_tests(new_series, new_series_map, ref_series, ref_series_map, scale, npast, coords, decay = False, localdiff = 0, kernelAggregation = 'product', weights = None, regInvGx = None):
    
    Kx = series_newKx(new_series, new_series_map, ref_series, ref_series_map, scale, npast, decay, localdiff, kernelAggregation, weights)
    
    return np.squeeze(coords.T @ regInvGx @ Kx)
    

def series_newcoords(new_series, new_series_map, ref_series, ref_series_map, scale, npast, coords, decay = False, localdiff = 0, kernelAggregation = 'product', weights = None,  knn_for_bandwidth=1, factor_for_bandwidth=None, eps=1e-8, Gx=None, embedder = None, projbasis = None):
    
    Kx = series_newKx(new_series, new_series_map, ref_series, ref_series_map, scale, npast, decay, localdiff, kernelAggregation, weights)

    #Omega = np.linalg.solve(Gx + np.diag(np.full(Gx.shape[0], eps)), Kx)
    #return np.squeeze(coords.T @ Omega)
    
    
    #Ks = embedder @ Omega
    #return np.squeeze(projbasis.T @ Ks)
    
    # New api, uses softmax and KNN search using the X RKHS distance
    
    #np.mean(np.diag(Gx))
    # Kx[i] = ⟨ k(new_x,·), k(ref_xi,·) ⟩ the inner product of kernels evaluated at new and ref points
    # d = Dist[i] = the distance to the new point
    # d = √ ‖ k(new_x,·) - k(ref_xi,·) ‖² 
    # d²  = ⟨ k(new_x,·) - k(ref_xi,·), k(new_x,·) - k(ref_xi,·)⟩ 
    # d²  = 2 - 2 * Kx[i]
    d = np.sqrt(np.clip(2 - 2 * Kx, 0, np.inf))
    
    w = dist_to_weights(d, knn_for_bandwidth, factor_for_bandwidth, eps)

    return w.T @ coords


# OLD HACK -- too specific to the kernel
#def Kx_to_weights(Kx, knn=1, factor=1, eps=1e-8):
#    # convert back to distances / removes the exp in newKx
#    # TODO: better API maybe
#    d = np.abs(np.log(Kx) / __kernel_params__[0])
#    if __kernel_params__[1] != 1:
#        d = np.power(d, 1/__kernel_params__[1])
#
#    return dist_to_weights(d, knn, factor, eps)


#def series_newcoords(new_series, new_series_map, ref_series, ref_series_map, scale, npast, nfuture, coords, decay = False, localdiff = 0, knn=1, factor=1, eps=1e-8):
#    
#    d = np.abs(series_newx_logk(new_series, new_series_map, ref_series, ref_series_map, scale, npast, nfuture, decay, localdiff) / __kernel_params__[0])
#
#    if __kernel_params__[1] != 1:
#        d = np.power(d, 1/__kernel_params__[1])
#    
#    w = dist_to_weights(d, knn, factor, eps)
#    
#    return w.T @ coords

# Kx = concaust.series_newKx(...)
# w = Kx_to_weights(...)
# Ks = embedder @ w ; nc = concaust.new_coords(Ks, Gs, coords, 'unbounded') ; 
# All this is actually equivalent to:
# nc = w.T @ coords
# That shortcut is implemented in
# nc = concaust.series_newcoords(...)
# Using nnls instead of 'unbounded' in new_coords is also possible, then
# more closely related to kernel moment matching instead of Nystrom extension
# (see the concaust ref paper)

#concaust.plot_attractors(coords, index_map, eigenvalues, predictions = nc, annotations=["%d" % year[idx] for idx in np.nditer(minindices)], annoted_coords=coords[minindices_ev,:], num_vars = 5, num_displayed_vars = 8, title="Sunspots")















