"""
Source code from the Continuous Causal States method,
described in the paper :
  Discovering Causal Structure with Reproducing-Kernel Hilbert Space ε-Machines
  by Nicolas Brodu and James P. Crutchfield

This code is free software, according to the MIT license
Author & Contact: nicolas.brodu@inria.fr
"""
from . import npda_helper
import numpy as np

# Non-negative least square implementation
# freely adapted from https://github.com/mikeiovine/fast-nnls
# To make it work with Dask and/or numpy and for the use of this project
def nnls_internal(AtA, Aty, iter_max=None, epsilon=None, force_dask = None):
  # avoid FutureWarning: `rcond` parameter will change to the default of machine precision, which we don't care, in the lstsq routine
  import warnings
  with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    
    npda = npda_helper.import_dask_or_numpy(force_dask)
    
    if epsilon is None:
        epsilon = np.finfo(np.float64).eps

    n = AtA.shape[0]

    if iter_max is None:
        iter_max = min(1000, 30 * n)

    if Aty.ndim != 1 or Aty.shape[0] != n:
        raise ValueError('Invalid dimension; got Aty vector of size {}, ' \
                         'expected {}'.format(Aty.shape, n))

    # Represents passive and active sets.
    # If sets[j] is 0, then index j is in the active set (R in literature).
    # Else, it is in the passive set (P).
    sets = np.zeros(n, dtype=bool)
    # The set of all possible indices. Construct P, R by using `sets` as a mask
    ind = np.arange(n, dtype=int)
    P = ind[sets]
    R = ind[~sets]

    x = np.zeros(n, dtype=np.float64)
    w = Aty
    s = np.zeros(n, dtype=np.float64)

    i = 0
    # While R not empty and max_(n \in R) w_n > epsilon
    #while da.sum(sets)<n and da.max(w[R]) > epsilon and i < iter_max:
    while not np.all(sets) and np.max(w[R]) > epsilon and i < iter_max:
        # Find index of maximum element of w which is in active set.
        j = npda_helper.npda_compute(npda.argmax(w[R]))
        # We have the index in MASKED w.
        # The real index is stored in the j-th position of R.
        m = R[j]

        # Move index from active set to passive set.
        sets[m] = True
        P = ind[sets]
        R = ind[~sets]

        # Get the rows, cols in AtA corresponding to P
        AtA_in_p = AtA[P][:, P]
        # Do the same for Aty
        Aty_in_p = Aty[P]
        
        # Update s. Solve (AtA)^p * s^p = (Aty)^p
        #s[P] = np.linalg.lstsq(AtA_in_p, Aty_in_p, rcond=None)[0]
        s[P] = npda_helper.npda_compute(npda.linalg.lstsq(AtA_in_p, Aty_in_p)[0])
        s[R] = 0.

        while np.any(s[P] <= epsilon):
            i += 1

            mask = (s[P] <= epsilon)
            alpha = np.min(x[P][mask] / (x[P][mask] - s[P][mask]))
            x += alpha * (s - x)

            # Move all indices j in P such that x[j] = 0 to R
            # First get all indices where x == 0 in the MASKED x
            zero_mask = (x[P] < epsilon)
            # These correspond to indices in P
            zeros = P[zero_mask]
            # Finally, update the passive/active sets.
            sets[zeros] = False
            P = ind[sets]
            R = ind[~sets]

            # Get the rows, cols in AtA corresponding to P
            AtA_in_p = AtA[P][:, P]
            # Do the same for Aty
            Aty_in_p = Aty[P]

            # Update s. Solve (AtA)^p * s^p = (Aty)^p
            s[P] = npda_helper.npda_compute(npda.linalg.lstsq(AtA_in_p, Aty_in_p)[0])
            s[R] = 0.

        x = s.copy()
        w = Aty - AtA.dot(x)

    return x


# Same on GPU. TODO: remove dask and use/adapt/simplify this code instead
def nnls_internal_GPU(xp, AtA, Aty, iter_max=None, epsilon=None):
  # avoid FutureWarning: `rcond` parameter will change to the default of machine precision, which we don't care, in the lstsq routine
  import warnings
  with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    
    if epsilon is None:
        epsilon = xp.finfo(xp.float64).eps

    n = AtA.shape[0]

    if iter_max is None:
        iter_max = min(1000, 30 * n)

    if Aty.ndim != 1 or Aty.shape[0] != n:
        raise ValueError('Invalid dimension; got Aty vector of size {}, ' \
                         'expected {}'.format(Aty.shape, n))

    # Represents passive and active sets.
    # If sets[j] is 0, then index j is in the active set (R in literature).
    # Else, it is in the passive set (P).
    sets = xp.zeros(n, dtype=bool)
    # The set of all possible indices. Construct P, R by using `sets` as a mask
    ind = xp.arange(n, dtype=int)
    P = ind[sets]
    R = ind[~sets]

    x = xp.zeros(n, dtype=xp.float64)
    w = Aty
    s = xp.zeros(n, dtype=xp.float64)

    i = 0
    # While R not empty and max_(n \in R) w_n > epsilon
    #while da.sum(sets)<n and da.max(w[R]) > epsilon and i < iter_max:
    while not xp.all(sets) and xp.max(w[R]) > epsilon and i < iter_max:
        # Find index of maximum element of w which is in active set.
        j = xp.argmax(w[R])
        # We have the index in MASKED w.
        # The real index is stored in the j-th position of R.
        m = R[j]

        # Move index from active set to passive set.
        sets[m] = True
        P = ind[sets]
        R = ind[~sets]

        # Get the rows, cols in AtA corresponding to P
        AtA_in_p = AtA[P][:, P]
        # Do the same for Aty
        Aty_in_p = Aty[P]
        
        # Update s. Solve (AtA)^p * s^p = (Aty)^p
        #s[P] = np.linalg.lstsq(AtA_in_p, Aty_in_p, rcond=None)[0]
        s[P] = xp.linalg.lstsq(AtA_in_p, Aty_in_p, rcond=None)[0]
        s[R] = 0.

        while np.any(s[P] <= epsilon):
            i += 1

            mask = (s[P] <= epsilon)
            alpha = xp.min(x[P][mask] / (x[P][mask] - s[P][mask]))
            x += alpha * (s - x)

            # Move all indices j in P such that x[j] = 0 to R
            # First get all indices where x == 0 in the MASKED x
            zero_mask = (x[P] < epsilon)
            # These correspond to indices in P
            zeros = P[zero_mask]
            # Finally, update the passive/active sets.
            sets[zeros] = False
            P = ind[sets]
            R = ind[~sets]

            # Get the rows, cols in AtA corresponding to P
            AtA_in_p = AtA[P][:, P]
            # Do the same for Aty
            Aty_in_p = Aty[P]

            # Update s. Solve (AtA)^p * s^p = (Aty)^p
            s[P] = xp.linalg.lstsq(AtA_in_p, Aty_in_p, rcond=None)[0]
            s[R] = 0.

        x = s.copy()
        w = Aty - AtA.dot(x)

    return x





# NNLS on each column of b
def nnls(A, b, max_iter=1000, eps=None, use_GPU = -1, use_dask = None, verbose = False):
    """
    Non-negative least square, minimizes |Ax - b|^2, subject to x>=0
    arguments: 
        A, b: numpy or cupy or dask arrays
        use_GPU: -1 = auto (default), use GPU if either A or b are on GPU. 0 = force no GPU, 1 = force GPU. Has precedence over use_dask (TODO: deprecate dask)
        use_dask: Whether to distribute the computations with Dask
        - None (default): Use the package flag (itself defaulting to False)
        - False or 0: Do not use Dask
        - True or 1: over every column of b. Best performance, may use more memory
        - 2: within the solving of each each column of b. May be useful when only one column is solved to still distribute some computations (warning: but this is not necessarily faster)
        max_iter, eps: internal parameter for stopping the iterative algorithm
    returns:
        The solution matrix, each column matches one column of b.
        always a Numpy object, even if Dask is used
    """
    try:
        # Use cupy if Gx or Gy is on the GPU, use GPU
        import cupy as cp
        if use_GPU==-1 and (cp.get_array_module(A) is cp or cp.get_array_module(b) is cp):
            use_GPU=1
        else:
            use_GPU=0
            
        if use_GPU==0:
            # force no GPU
            xp = np
            A = cp.asnumpy(A) # copy on CPU if not there already
            b = cp.asnumpy(b) # copy on CPU if not there already
        else:
            xp = cp
            A = cp.asarray(A) # copy on GPU if not there already
            b = cp.asarray(b) # copy on GPU if not there already
            use_dask = 0

    except:
        xp = np
        if use_GPU==1:
            print("Warning: GPU is not available. Using CPU.")
            
    ret = xp.empty((A.shape[1], b.shape[1]))

    if xp is not np:
        AtA = A.T @ A
        for c in range(b.shape[1]):
            Aty = A.T @ b[:,c]
            ret[:, c] = nnls_internal_GPU(xp, AtA, Aty, max_iter, eps)
        return ret

    
    if use_dask is None:
        use_dask = npda_helper.use_dask()
        
    if use_dask == 1:
        AtA = A.T @ A
        for c in range(b.shape[1]):
            Aty = A.T @ b[:,c]
            ret[:, c] = nnls_internal(AtA, Aty, max_iter, eps, force_dask = True)
        return ret
    
    elif use_dask == 2:
        AtA = A.T @ A
        import dask
        task_list = []
        for c in range(b.shape[1]):
            Aty = A.T @ b[:,c]
            task_list += [ dask.delayed(nnls_internal)(AtA, Aty, max_iter, eps, force_dask = False) ]
        task_list = dask.compute(task_list)
        for c in range(b.shape[1]):
            ret[:, c] = task_list[c]
        return ret

    # plain numpy - no parallelism unless linalg.lsq does it with other means
    #for c in range(b.shape[1]):
    #    Aty = A.T @ b[:,c]
    #    ret[:, c] = nnls_internal(AtA, Aty, max_iter, eps, force_dask = False)
    import scipy.optimize
    for c in range(b.shape[1]):
        if verbose: print(f'nnls: {c+1}/{b.shape[1]}')
        ret[:, c] = scipy.optimize.nnls(A, b[:,c])[0]
    return ret


