"""
Source code from the Continuous Causal States method,
described in the paper :
  Discovering Causal Structure with Reproducing-Kernel Hilbert Space ε-Machines
  by Nicolas Brodu and James P. Crutchfield

This code is free software, according to the MIT license
Author & Contact: nicolas.brodu@inria.fr
"""

# Using a double-underscore to avoid exporting module configuration variables
# Make these variables lists, because the variables would become read-only
# when called from outside the module. However, a list entry is actually
# a pointer to memory that can be modified. Useful and usual Python trick!

__use_dask__ = [ False ]

__chunk_size__ = [ 1024 ]

def use_dask(flag = None):
    """
    Specifies or returns whether the module should use Dask.
    The default is False.
    If the argument is not specified (default is None), then
    the current usage is returned.
    If the argument is specified, set the value to this argument
    Returns whether this module will use Dask or not
    """
    if flag is not None:
        __use_dask__[0] = flag
    return __use_dask__[0]

def chunk_size(size = None):
    """
    Specifies or returns the chunk size to use along each dimension,
    in order to chop numpy arrays into smaller Dask block.
    Dask can only parallelize over these blocks, so they should not be too
    large. Yet, too small means more overhead.
    If the argument is not specified (default is None), then
    the current block size is returned.
    If the argument is specified, set the value to this argument
    Returns the current or updated chunk size value
    """
    if size is not None:
        __chunk_size__[0] = size
    return __chunk_size__[0]

def import_dask_or_numpy(force_dask = None):
    """
    Returns either the numpy or the dask module
    Can be forced, or use the default flag if no/None argument is passed
    For internal use, not exported in __init__
    """
    if force_dask is not None:
        if force_dask:
            import dask.array as npda
        else:
            import numpy as npda
    else:
        if __use_dask__[0]:
            import dask.array as npda
        else:
            import numpy as npda
    return npda

# internal helper functions -- need not be exported

def npda_compute(x):
    if __use_dask__[0]:
        return x.compute()
    return x
    

def pad_to_square_chunks(a):
    """
    Enlarges a to be a multiple of the chunk size in each dimension >1

    Parameters
    ----------
    a : dask array.

    Returns
    -------
    an array with filled-up values

    """
    import dask.array as da
    
    next0 = ((a.shape[0]+chunk_size()-1) // chunk_size()) * chunk_size()
    next1 = ((a.shape[1]+chunk_size()-1) // chunk_size()) * chunk_size()
    rem0 = next0 - a.shape[0]
    rem1 = next1 - a.shape[1]
    if rem0 != rem1:
        raise ValueError("pad_to_square_chunks received a non-square matrix")
    
    ret = da.concatenate(
        [da.concatenate([a, da.zeros((rem0, a.shape[1]))], axis=0),
         da.concatenate([da.zeros((a.shape[0], rem1)), da.eye(rem1)], axis=0)],
        axis=1)

    return ret.rechunk((chunk_size(), chunk_size()))
