#!/usr/bin/env python3
"""
This is the code for the Continuous Causal States method
described in the papers:

  Discovering Causal Structure with Reproducing-Kernel Hilbert Space ε-Machines
  by Nicolas Brodu and James P. Crutchfield

  Inferring Kernel ε-Machines: Discovering Structure in Complex Systems
  by Nicolas Brodu and Alexandra M. Jurgens1

  and possibly more to come
  
This code is free software, according to the MIT license
Author & Contact: nicolas.brodu@inria.fr
"""

import numpy as np
import scipy.linalg
import scipy.signal

# This is for running the script from the source code of the uninstalled package
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
# Otherwise you can just import concaust and skip the above 3 lines
import concaust

import scipy.io
import scipy.linalg
import concaust.ccutils
import sys


# ==========================================
# Plotting routines - The main code is below
# ==========================================


import pyvista as pv

def polyline_from_points(points):
    poly = pv.PolyData()
    poly.points = points
    the_cell = np.arange(0, len(points), dtype=np.int_)
    the_cell = np.insert(the_cell, 0, len(points))
    poly.lines = the_cell
    return poly
    

def plot_3D_pyvista(coords, phi, snapshot=None, return_cpos=False):
    
    dphi = np.diff(phi)
    # add duplicate points for wraps
    for i,idx in enumerate(np.where(dphi>180)):
        coords = np.insert(coords, idx+i+1, coords[idx+i+1,:], axis=0)
        phi = np.insert(phi, idx+i+1, phi[idx+i+1]-360)
    dphi = np.diff(phi)
    # add duplicate points for wraps
    for i,idx in enumerate(np.where(dphi<-180)):
        coords = np.insert(coords, idx+i+1, coords[idx+i+1,:], axis=0)
        phi = np.insert(phi, idx+i+1, phi[idx+i+1]+360)

    from matplotlib.colors import LinearSegmentedColormap
    hsv3 = LinearSegmentedColormap.from_list("hsv3", ["#ff0000", "#ffa700", "#afff00", "#08ff00", "#00ff9f", "#00b7ff", "#0010ff", "#9700ff", "#ff00bf", "#ff0000", "#ffa700", "#afff00", "#08ff00", "#00ff9f", "#00b7ff", "#0010ff", "#9700ff", "#ff00bf", "#ff0000", "#ffa700", "#afff00", "#08ff00", "#00ff9f", "#00b7ff", "#0010ff", "#9700ff", "#ff00bf", "#ff0000"])
    
    color_scalar = (phi + 360) / 1080
    
    polyline = polyline_from_points(coords[:,1:4])
    polyline["scalars"] = color_scalar
    tube = polyline.tube(radius=0.0075)
    p = pv.Plotter(window_size=[2000,2000], notebook=False, off_screen = bool(snapshot))
    p.add_mesh(tube, smooth_shading=True, cmap=hsv3, ambient=0.5, opacity=0.15, clim=[0,1])
    
    p.add_mesh(polyline, smooth_shading=True, cmap=hsv3, ambient=0.5, render_points_as_spheres=True, style='points', point_size=30)
    
    # PCA on the bottom clusters to get a plane below them
    # HACK: hardcoded threshold using the positions
    c = np.copy(coords[coords[:,3]<-0.8,1:4])
    mc = np.mean(c, axis=0)
    c -= mc
    cov = np.cov(c, rowvar = False)
    eigval, eigvec = scipy.linalg.eigh(cov)
    vec = eigvec[:,0]
    if vec[2] < 0:
        vec = -vec
    plane_center = mc - vec/2
    p.add_mesh(
        pv.Plane(center=tuple(plane_center), direction=tuple(vec), i_size=3.75, j_size=4.75),
        color='#eeeeee',ambient=0,smooth_shading=True
    )
    
    
    # frame -- new, less obtrusive and at the bottom
    frame_origin = plane_center + np.asarray([1.45,0.6,0.2]) #np.asarray([0.21,0,0])
    shaft_radius = 0.02
    tip_radius = 0.05
    tip_length = 0.2
    scale = 0.5
    p.add_mesh(
        pv.Arrow(start=frame_origin, direction=(1,0,0), tip_length=tip_length, tip_radius=tip_radius, shaft_radius=shaft_radius, scale=scale),
        color='black',ambient=0,smooth_shading=False
    )
    p.add_mesh(
        pv.Arrow(start=frame_origin, direction=(0,1,0), tip_length=tip_length, tip_radius=tip_radius, shaft_radius=shaft_radius, scale=scale),
        color='black',ambient=0,smooth_shading=False
    )
    p.add_mesh(
        pv.Arrow(start=frame_origin, direction=(0,0,1), tip_length=tip_length, tip_radius=tip_radius, shaft_radius=shaft_radius, scale=scale),
        color='black',ambient=0,smooth_shading=False
    )
    
    camera_pos = [(9.218831621418648, -0.8022138598005301, 4.493273664699347), (0.01082022549120465, -0.07704698360971318, -0.3773915013778544), (-0.4715716074918947, -0.12161024172157756, 0.8734020655551085)]
    
    p.enable_ssao()
    p.enable_anti_aliasing('ssaa')
    p.hide_axes()
    p.remove_scalar_bar()
    camera_pos = p.show(return_cpos=return_cpos, cpos=camera_pos)
    if snapshot:
        p.screenshot(snapshot)
        
    return camera_pos
    

def visualize_molecule_motion(positions, snapshot=None, return_cpos=False):
    
    # Use the middle of the center carbon link
    mpos = np.empty((positions.shape[0], 3))
    for i in range(positions.shape[0]):
        mpos[i,:] = (positions[i,12:15] + positions[i,21:24]) / 2
        
    diffuse = 1
    ambient = 0
    specular = 0
    metallic = 0.2
    pbr = True
    
    p = pv.Plotter(window_size=[2000,2000], notebook=False, lighting='none', off_screen = bool(snapshot))
    p.add_light(pv.Light(light_type='camera light', position=(1,1,1)))
    p.set_environment_texture(pv.cubemap("./sky_cubemap", "", ext=".png"))

    polyline = polyline_from_points(mpos)
    polyline["scalars"] = np.linspace(0,1,mpos.shape[0])
    tube = polyline.tube(radius=0.1)
    
    from matplotlib import cm
    p.add_mesh(tube, cmap=cm.get_cmap('viridis'), diffuse=diffuse, ambient=ambient,specular=specular,smooth_shading=True, pbr=pbr, metallic=metallic)
    
    p.enable_ssao(radius=1)
    p.enable_anti_aliasing('ssaa')
    p.remove_scalar_bar()
    camera_pos = [(258.08100301384627, 95.37933846474786, 128.5089577805058),
 (28.39752080907665, -38.35854974231903, -33.583398697457326),
 (-0.45215873549616337, -0.2582080469402082, 0.8537453264351342)]
    camera_pos = p.show(return_cpos=return_cpos, cpos=camera_pos)
    if snapshot:
        p.screenshot(snapshot)

    return camera_pos


def visualize_molecule_pyvista(positions, snapshot=None, color_HC_for_clusters = False, color_methyl_groups = False, color_H_all_gray = False, display_frame=False):
    import matplotlib.pyplot as plt
    
    cidx = [0,4,7,10]
    hidx = [1,2,3,5,6,8,9,11,12,13]
    
    diffuse = 1
    ambient = 0
    specular = 0
    metallic = 0.2
    pbr = True
    
    p = pv.Plotter(window_size=[2000,2000], notebook=False, lighting='none', off_screen = bool(snapshot))
    p.add_light(pv.Light(light_type='camera light', position=(1,1,1)))
    
    for i in cidx:
        
        if color_methyl_groups and (i==0 or i==10):
            Ccolor = '#0000ff'
        elif color_HC_for_clusters and (i==10):
            Ccolor = '#ffff00'
        else:
            Ccolor = '#ffffff'
            
        p.add_mesh(
            pv.Icosphere(center=tuple(positions[i,:]), radius=0.4 if color_methyl_groups else 0.6, nsub=5),
            color=Ccolor, diffuse=diffuse, ambient=ambient,specular=specular,smooth_shading=True, pbr=pbr, metallic=metallic
        )
        
    hcmap = plt.get_cmap('jet')
    color_h_idx = [2, 4, 8, 7, 0, 9, 5, 6, 3, 1]
    for loopidx, (ci, i) in enumerate(zip(color_h_idx, hidx)):
        color=hcmap(ci/(len(hidx)-1))
        if (color_HC_for_clusters and loopidx>2) or color_H_all_gray:
            color='#444444'

        if color_methyl_groups:
            if loopidx<3 or loopidx>=7:
                color='#0000ff'
            else:
                color='#222222'
        
        p.add_mesh(
            pv.Icosphere(center=tuple(positions[i,:]), radius=0.25 if color_methyl_groups else 0.4, nsub=5),
            color=color,
            diffuse=diffuse, ambient=ambient,specular=specular,smooth_shading=True, pbr=pbr, metallic=metallic
        )
        
    def join(i,j):
        p.add_mesh(
            pv.Cylinder(center=tuple((positions[i,:]+positions[j,:])/2), direction=tuple(positions[j,:]-positions[i,:]), radius=0.05, height=np.linalg.norm(positions[j,:]-positions[i,:])),
            color='#222222', diffuse=diffuse, ambient=ambient,specular=specular,smooth_shading=True, pbr=pbr, metallic=metallic
        )
    
    # carbon links
    join(0,4)
    join(4,7)
    join(7,10)
    # hydrogen links
    join(0,1) ; join(0,2) ; join(0,3)
    join(4,5) ; join(4,6)
    join(7,8) ; join(7,9)
    join(10,11) ; join(10,12) ; join(10,13)


    if display_frame:
    
        arrow_scale = 3
        shaft_radius = 0.04
        tip_radius = 0.1
        
        # frame
        arrx = pv.Arrow(start=(0,0,0), direction=(1,0,0), tip_length=0.2, tip_radius=tip_radius, shaft_radius=shaft_radius, scale = arrow_scale)
        p.add_mesh(
            arrx,
            color='red',ambient=1,smooth_shading=False
        )
        p.add_mesh(
            pv.Arrow(start=(0,0,0), direction=(0,1,0), tip_length=0.2, tip_radius=tip_radius, shaft_radius=shaft_radius, scale = arrow_scale),
            color='red',ambient=1,smooth_shading=False
        )
        p.add_mesh(
            pv.Arrow(start=(0,0,0), direction=(0,0,1), tip_length=0.2, tip_radius=tip_radius, shaft_radius=shaft_radius, scale = arrow_scale),
            color='red',ambient=1,smooth_shading=False
        )



    p.set_environment_texture(pv.cubemap("./sky_cubemap", "", ext=".png"))
    
    p.enable_ssao()
    p.enable_anti_aliasing('ssaa')
    if display_frame:
        p.show_axes()
    else:
        p.hide_axes()
    p.set_position([-3,-5,-2])
    p.set_viewup([0,0,1])
    p.show()
    if snapshot:
        p.screenshot(snapshot)


def post_process(data, co, im, title=''):
    from sklearn.cluster import DBSCAN
    

    # Cluster the states using DBSCAN. Some fiddling may be needed
    # with the arbitrary threshold of DBSCAN, especially when including
    # larger dimensions (which get progressively meaningless)
    # Use clustering in high-dimension (at least these which were computed
    # by the main function)
    # However using all dimensions allows to label the clusters using markers,
    # thus help visualize how the low-dim clusters split
    #db = DBSCAN(eps=0.03, min_samples=100).fit(co)
    
    # The alternative is to restrict the search of clusters using only the 
    # first, more meaningful, state variables
    db = DBSCAN(eps=0.25, min_samples=100).fit(co[:,2:5])
    
    # In this example, there should be 9 clusters, 
    # from 0 to 8, with -1 for outliers
    labels = db.labels_
    nlabels = np.max(labels)+1
    #assert(nlabels==9)
    
    print("Cluster positions:")
    for i in range(nlabels):
        # compute the average positions for the atoms in that cluster
        meanpos = np.mean(data[im[labels==i],:], axis=0)
        #visualize_molecule(np.reshape(meanpos, (14,3)), f"Cluster {i}")
        visualize_molecule_pyvista(np.reshape(meanpos, (14,3)), f"butane_cluster_{i}.png", color_HC_for_clusters = True, display_frame = False)
        print(i, np.mean(co[labels==i,1:4], axis=0))

    # This can be done in the shell to make the background transparent
    # for f in butane_cluster_*.png ; do convert $f -transparent white transparent_$f ; done


    # Load the dihedral angle data
    phi_mat = scipy.io.loadmat('data/butane/phi.mat')
    # Express the dihedral angle in degrees
    # select the angles matching the state indices
    phi = np.squeeze(phi_mat['phi'])[im] * 180 / np.pi


    # This interactive visualization tool is very experimental and its API
    # is completely unstable.
    
    #concaust.plot_trajectories(co, im, ev, num_vars=6, color_by_values=[[phi,0,360], labels], color_by_names=["dihedral angle","clusters"], color_by_scales=['HSV','Rainbow'], title=title, annotations=[f"────── Cluster {i}" for i in range(nlabels)], annoted_coords=np.asarray([np.mean(co[labels==i,:], axis=0) for i in range(9)]), annotations_alignment="middle right", annotations_legend='Clusters', display=True, axes_mode=2)
    
    # Use this to save the plot to a file
    #concaust.plot_trajectories(co, im, ev, num_vars=6, color_by_values=[[phi,0,360], labels], color_by_names=["dihedral angle","clusters"], color_by_scales=['HSV','Rainbow'], title=title, filename="butane_efftraj.html")
    
    #concaust.plot_trajectories(co, im, ev, num_vars=6, color_by_values=[[phi,0,360], labels], color_by_names=["dihedral angle","clusters"], color_by_scales=['HSV','Rainbow'], title=title, axes_mode=2, filename="butane_efftraj_noframe.html")

    plot_3D_pyvista(co,phi,"butane_plot.png")

#sys.exit(0)


# =========
# Main code
# =========

if __name__ == '__main__':
    
    # Details of the data in: Stefan Klus, Péter Koltai, Christof Schütte, "On the
    # numerical approximation of the Perron–Frobenius and Koopman operator", 
    # Journal of Computational Dynamics, 2016, 3(1): 51-79. doi: 10.3934/jcd.2016003
    # It consists of a simulation of the positions of the n-butane molecule atoms
    # every 2fs for a total duration of 10ns, subsambled by a factor 100.
    # Hence 42 atom positions and 50000 data points
    # This data set is kindly provided by Stefan Klus
    
    print("Loading and preprocessing")
    butane_mat = scipy.io.loadmat('data/butane/amberSim.mat')
    positions = butane_mat['x'].T

    # The whole molecule moves in space. If we look at raw atom positions,
    # then the algorithm finds back the molecule average position and
    # the 3 space coordinates as the main causal components.
    # But we are not interested here in the global random walk.
    # Rather, we'd like to find the internal molecular states. For this, we
    # express the atom positions with respect to an internal frame that moves
    # with the molecule. Thus, we can study the molecular conformations
    # independently of where the molecule is located globally.
    for i in range(positions.shape[0]):
        # carbon atoms
        ci = np.reshape(positions[i,np.r_[0:3, 12:15, 21:24, 30:33]], (4,3))
        # middle branch: first basis vector
        e0 = ci[2,:]-ci[1,:]
        e0 = e0 / np.linalg.norm(e0)
        # cross product of middle branch with the first
        # These cannot be colinear, so e1 is orthogonal to e0
        e1 = np.cross(e0, ci[0,:]-ci[1,:])
        e1 = e1 / np.linalg.norm(e1)
        # complete the basis
        e2 = np.cross(e0, e1)
        # new origin = the middle of the middle branch
        o = (ci[1,:]+ci[2,:]) * 0.5
        # center positions and project in the new frame of reference
        pi = np.reshape(positions[i,:], (14,3))
        pi = (pi - o[None,:]) @ np.asarray([e0, e1, e2]).T
        positions[i:i+1,:] = np.reshape(pi, (1,42))

    # Number of time steps to retain. 
    # We use as much data as our GPU memory allows
    # Better solutions are work in progress
    N = 25000
    selected_data = positions[:N, :]
    
    # We need a scale for the algorithm
    # Let's use the full molecule span
    span = np.empty(positions.shape[0])
    for i in range(positions.shape[0]):
        pi = np.reshape(positions[i,:], (14,3))
        span[i] = np.max(scipy.spatial.distance.pdist(pi))
    scale = np.mean(span)

    # Stop here to just load the data... 
    # ...or comment to pursue on a powerful enough machine
    sys.exit(0)
    
    # Transfer to GPU. Comment these 2 lines if you don't have cupy or a GPU
    # or if you wish to wait for a long time and use (hopefully larger) CPU mem
    import cupy as cp
    selected_data = cp.asarray(selected_data)

    # Consider all atoms as separate data sources
    selected_data = [ selected_data[:N, i*3:i*3+3] for i in range(14) ]

    # In this example, we'd like to study instantaneous conformation changes.
    # Let's take as small histories as possible
    # These match the temporal 200fs scale of the original data
    npast = 1
    nfuture = 1
    
    print("Computing the similarity matrices in RKHS")
    Gx, Gy, index_map = concaust.series_Gxy(selected_data, scale, npast, nfuture)
    
    print("Embedding the states")
    Gs = concaust.embed_states(Gx, Gy)
    
    # mem cleanup for very large arrays, before spectral computations
    del Gx, Gy
    
    print("Computing the reduced dimension basis")
    eigenvalues, basis, coords = concaust.spectral_basis(Gs, num_basis=10)

    print("Main computations done")

    # This script is run on a remote machine with a powerful NVidia a100 GPU
    # Save on that machine the useful info for further post-processing
    np.savez("butane_effvars.npz", eigenvalues=eigenvalues, coords=coords, index_map=index_map)

    # Comment this or copy-paste the code below to post-process the data
    sys.exit(0)

    bres = np.load('butane_effvars.npz')
    eigenvalues, coords, index_map = bres['eigenvalues'], bres['coords'], bres['index_map']
    coords[:,1] = -coords[:,1]
    selected_data = positions
    # Adjust the title according to the parameters used above
    post_process(selected_data, coords, index_map, "Histories of 1 sample (200fs). Analysis scale = average span")
    
    # Visualizing one of the molecule positions, not the average
    # in clusters, and with the local frame for reference
    visualize_molecule_pyvista(np.reshape(positions[1000:1001,:], (14,3)), "butane_ref_position.png", color_H_all_gray = True, display_frame = True)
    
    # Plotting the eigenvalues
    from plot_eigenvalues import plot_eigenvalues
    plot_eigenvalues(eigenvalues,'Butane molecule', 'butane_eigenvalues')


