#!/usr/bin/env python3
"""
This is the code for the Continuous Causal States method
described in the papers:

  Discovering Causal Structure with Reproducing-Kernel Hilbert Space ε-Machines
  by Nicolas Brodu and James P. Crutchfield

  Inferring Kernel ε-Machines: Discovering Structure in Complex Systems
  by Nicolas Brodu and Alexandra M. Jurgens1

  and possibly more to come
  
This code is free software, according to the MIT license
Author & Contact: nicolas.brodu@inria.fr
"""

import numpy as np
import pandas as pd

# This is for running the script from the source code of the uninstalled package
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
# Otherwise you can just import concaust and skip the above 3 lines
import concaust

import pyvista as pv


# ==========================================
# Plotting routines - The main code is below
# ==========================================


import matplotlib.pyplot as plt
import matplotlib.dates as mdates

def plot_data_series(gap_filled_series, qc_data, timestamps, data_labels):
    fig = plt.figure()
    gs = fig.add_gridspec(6, hspace=0)
    axs = gs.subplots(sharex=True)
    for i in range(6):
        qc = qc_data[i]
        ser = np.copy(gap_filled_series[i])
        gap_ser = np.copy(gap_filled_series[i])
        ser[qc==0] = np.nan
        gap_ser[qc!=0] = np.nan
        vidx = np.where(qc==0)[0]
        vprev = -1
        for v in vidx:
            if v-vprev > 1:
                gap_ser[v-1] = ser[v-1]
                if vprev != -1:
                    gap_ser[vprev+1] = ser[vprev+1]
            vprev = v
            
        axs[i].plot(timestamps, ser, linewidth=0.25, color='k')
        axs[i].plot(timestamps, gap_ser, linewidth=0.75, color='orange')
        #axs[i].set_ylabel(data_labels[i], rotation=0, labelpad=40)
        axs[i].label_outer()
        axs[i].tick_params(axis='y',labelsize=14)
        axs[i].margins(x=0)
        
        for first, last, c in culture_dates:
            axs[i].axvspan(first, last, facecolor=culture_colors_ref[c], alpha=0.25)
        
        axs[i].axvspan(np.datetime64('2004-01-01'), np.datetime64('2004-01-01') + np.timedelta64('90', 'D'), facecolor='#aaaaaa', alpha=0.25)
        axs[i].axvspan(np.datetime64('2014-12-31')-np.timedelta64('90', 'D'), np.datetime64('2014-12-31'), facecolor='#aaaaaa', alpha=0.25)
        
    # hard-coded formatting...
    axs[0].yaxis.set_ticks([0,10,20])
    axs[0].yaxis.set_ticklabels(["0","","20"])
    axs[1].yaxis.set_ticks([0,100,200,300])
    axs[1].yaxis.set_ticklabels(["0","","200",""])
    axs[2].yaxis.set_ticks([0,10,20])
    axs[2].yaxis.set_ticklabels(["0","","20"])
    axs[3].yaxis.set_ticks([0,10,20,30,40,50])
    axs[3].yaxis.set_ticklabels(["0","","20","","40",""])
    axs[4].yaxis.set_ticks([0,50,100,150])
    axs[4].yaxis.set_ticklabels(["0","","100",""])
    axs[5].yaxis.set_ticks([-15,-10,-5,0,5])
    axs[5].yaxis.set_ticklabels(["","-10","","0",""])
    plt.xlim([np.datetime64('2004-01-01'), np.datetime64('2015-01-01')])
    plt.gca().xaxis.set_major_locator(mdates.YearLocator())
    plt.gca().xaxis.set_minor_locator(mdates.MonthLocator())
    plt.gca().tick_params(axis='x',labelsize=14)
    
    
    return axs


# color-blind friendly palette from : https://jfly.uni-koeln.de/color/
# Unfortunately, the 3D shading completely destroys the luminance differences
# between these colors. 
# The results are worse for everyone, including color-blind people
# culture_colors_ref = {
#     'Mustard' : '#e69f00',
#     'Maize' : '#0072b2',
#     'Wheat' : '#009e73',
#     'Triticale' : '#009e73',
#     'Barley' : '#009e73',
#     'Phacelia' : '#cc79a7',
#     'Rapeseed' : '#d55e00',
# }

# Even the matplotlib tab10 is not contrasted enough
#from matplotlib import cm
#tab10 = cm.get_cmap('tab10')

culture_colors_ref = {
    'Mustard' : 'orange', #tab10(1),
    'Maize' : 'blue', #tab10(0),
    'Wheat' : 'green', #tab10(2),
    'Triticale' : 'green', #tab10(2),
    'Barley' : 'green', #tab10(2),
    'Phacelia' : 'magenta', #tab10(4),
    'Rapeseed' : 'red', #tab10(3)
}


def polyline_from_points(points):
    poly = pv.PolyData()
    poly.points = points
    the_cell = np.arange(0, len(points), dtype=np.int_)
    the_cell = np.insert(the_cell, 0, len(points))
    poly.lines = the_cell
    return poly


def plot_3D_pyvista(plot_coords, relevance_factors, days_of_year, culture_dates, index_map, title=None, snapshot=None, return_cpos=False, cpos=None):
    
    p = pv.Plotter(window_size=[2000,2000], notebook=False, off_screen = bool(snapshot), lighting='none')
    p.add_light(pv.Light(light_type='scene light', position=(3,3,3), positional=True, intensity=1, cone_angle=150 ))
    p.add_light(pv.Light(light_type='scene light', position=(3,0,0), positional=True, intensity=0.5, cone_angle=150 ))
    p.add_light(pv.Light(light_type='scene light', position=(3,-3,0), positional=True, intensity=1, cone_angle=150 ))

    pbr = True
    ambient = 0.5
    specular = 0.5
    metallic = 0

    tube_radius = 0.010
    projtube_radius = 0.016
    tube_opacity = 0.25
    points_size = 18


    default_color = '#888888'
    iprev = 0
    for first, last, c in culture_dates:
        iarr = np.where(timestamps[index_map]==np.datetime64(first))[0]
        if iarr.size==0:
            continue
        ifirst = iarr[0]
        iarr = np.where(timestamps[index_map]==np.datetime64(last))[0]
        if iarr.size==0: 
            continue
        ilast = iarr[0]
    
        polyline = polyline_from_points(plot_coords[iprev:ifirst+1,:])
        tube = polyline.tube(radius=tube_radius)
        p.add_mesh(tube, smooth_shading=True, color=default_color, ambient=ambient, specular=specular, pbr=pbr, metallic=metallic, opacity=tube_opacity)
        
        p.add_mesh(polyline, smooth_shading=True, color=default_color, ambient=ambient, render_points_as_spheres=True, style='points', point_size=points_size)

        polyline = polyline_from_points(plot_coords[ifirst:ilast+1,:])
        polyline["scalars"] = np.linspace(0,1,ilast-ifirst+1)
        tube = polyline.tube(radius=tube_radius)
        p.add_mesh(tube, smooth_shading=True, color=culture_colors_ref[c], ambient=ambient, specular=specular, pbr=pbr, metallic=metallic, opacity=tube_opacity)
        
        p.add_mesh(polyline, smooth_shading=True, color=culture_colors_ref[c], ambient=ambient, render_points_as_spheres=True, style='points', point_size=points_size)
        
        iprev = ilast
    # last segment
    polyline = polyline_from_points(plot_coords[iprev:,:])
    tube = polyline.tube(radius=tube_radius)
    p.add_mesh(tube, smooth_shading=True, color=default_color, ambient=ambient, specular=specular, pbr=pbr, metallic=metallic, opacity=tube_opacity)

    p.add_mesh(polyline, smooth_shading=True, color=default_color, ambient=ambient, render_points_as_spheres=True, style='points', point_size=points_size)
    
    projection_plane_origin = np.asarray([0,0,-3.3])
    
    doy = np.copy(days_of_year)
    ddoy = np.diff(days_of_year)
    idprev = -1
    for i,idx in enumerate(np.where(ddoy<0)[0]):
        ndays = idx - idprev
        
        polyline = polyline_from_points(plot_coords[idprev+1:idx+2,:])
        color_scalar = np.copy(doy[idprev+1:idx+2])
        color_scalar[-1] = color_scalar[-2] + 1
        color_scalar = color_scalar / ndays / 2
        polyline["scalars"] = color_scalar
        tube = polyline.tube(radius=tube_radius)
        projected = polyline.project_points_to_plane(origin=projection_plane_origin).tube(radius=projtube_radius)
        p.add_mesh(projected, smooth_shading=True, cmap='hsv', ambient=0.5, specular=1, pbr=False, metallic=1)
        
        idprev = idx
   
    polyline = polyline_from_points(plot_coords[idprev+1:,:])
    color_scalar = np.copy(doy[idprev+1:]) / 365 / 2
    polyline["scalars"] = color_scalar
    tube = polyline.tube(radius=tube_radius)
    projected = polyline.project_points_to_plane(origin=projection_plane_origin).tube(radius=tube_radius)
    p.add_mesh(projected, smooth_shading=True, cmap='hsv', ambient=0.5, specular=1, pbr=False, metallic=1)

    if pbr:
        p.set_environment_texture(pv.cubemap("../../sky_cubemap", "", ext=".png"))
    
    
    # big walls increase shading + allow to easy visualize the projection
    
    p.add_mesh(
        pv.Plane(center=projection_plane_origin*1.002, direction=(0,0,1), i_size=10, j_size=10),
        color='#ffffff',ambient=0,smooth_shading=True, pbr=pbr
    )

    p.add_mesh(
        pv.Plane(center=(0,-2,0), direction=(0,1,0), i_size=10, j_size=10),
        color='#ffffff',ambient=0,smooth_shading=True, pbr=pbr
    )

    p.add_mesh(
        pv.Plane(center=(-2,0,0), direction=(1,0,0), i_size=10, j_size=10),
        color='#ffffff',ambient=0,smooth_shading=True, pbr=pbr
    )

    p.enable_ssao()
    p.enable_anti_aliasing('ssaa')
    p.hide_axes()
    p.remove_scalar_bar()
    camera_pos = p.show(return_cpos=return_cpos, cpos=cpos, title=title)
    if snapshot:
        p.screenshot(snapshot)
        
    return camera_pos




# =========
# Main code
# =========



variable_names = {
    'TA_F' : 'Temperature',
    'SW_IN_F' : 'Shortwave radiation',
    'VPD_F' : 'Vapor pressure deficit',
    'P_F' : 'Precipitation',
    'LE_F_MDS' : 'Latent heat',
    'NEE_VUT_REF' : 'CO₂ flux',
}
data_var = ['TA_F', 'SW_IN_F', 'VPD_F', 'P_F', 'LE_F_MDS', 'NEE_VUT_REF']


# The data was measured by a flux tower at the Grignon site, 40km from Paris
# Data is provided by ICOS (Integrated Carbon Observatory System) and licensed
# under Creative Commons Attribution 4.0 International licence (CC BY 4.0)
# See this page: https://meta.icos-cp.eu/resources/stations/ES_FR-Gri
# The original data can be downloaded here
# https://meta.icos-cp.eu/objects/zEKOPCS-ErtqYDGuqYykEsOK
# Only a relevant subset is included in our git repo
data = pd.read_csv('data/crop_field/FR-Gri_daily.csv', skiprows=1)
data[data<=-9999] = np.nan

# copy so filtering out bad qc data and scaling does not modify the original
series = [np.copy(data[k].to_numpy()[:,None]) for k in data_var]

labels = [variable_names[k] for k in data_var]
timestamps = pd.to_datetime(data['TIMESTAMP'], format='%Y%m%d').to_numpy(dtype='datetime64[D]')

# Quality Control data. The ICOS group does their best to provide a quality
# index for the flux data. Low quality data should be discarded, but we can 
# still inspect the original measurements for qc flags close to 1.
qc_data = [np.ones(data.shape[0], dtype=int) for k in data_var]
if 'LE_F_MDS' in data_var and 'LE_F_MDS_QC' in data.columns:
    qc_data[data_var.index('LE_F_MDS')] = data['LE_F_MDS_QC'].to_numpy()
if 'NEE_VUT_REF' in data_var and 'NEE_VUT_REF_QC' in data.columns:
    qc_data[data_var.index('NEE_VUT_REF')] = data['NEE_VUT_REF_QC'].to_numpy()

for k in range(len(series)):
    qc_data[k][np.isnan(series[k]).any(axis=1)] = 0
    qc_data[k][np.isnan(qc_data[k])] = 0
    if np.max(qc_data[k]) == 0:
        qc_data[k].fill(1) # unreliable qc data - do not use qc, i.e use data

    series[k][qc_data[k]<=0] = np.nan


# ==== cultures database ====
# Data provided by the ICOS group managing the Grignon field
# Ranges are from sowing/seedling to cut/harvest/grinding
culture_dates = [
    [ '2004-09-02', '2005-05-07', 'Mustard' ],
    [ '2005-05-09', '2005-09-28', 'Maize' ],
    [ '2005-10-28', '2006-07-15', 'Wheat' ],
    [ '2006-10-04', '2007-06-29', 'Barley' ],
    [ '2007-09-25', '2008-04-14', 'Mustard' ],
    [ '2008-04-28', '2008-09-25', 'Maize' ],
    [ '2008-10-17', '2009-07-31', 'Wheat' ],
    [ '2009-10-14', '2010-07-19', 'Triticale' ],
    [ '2010-09-14', '2011-03-23', 'Phacelia' ],
    [ '2011-04-21', '2011-09-06', 'Maize' ],
    [ '2011-10-20', '2012-08-03', 'Wheat' ],
    [ '2012-08-31', '2013-08-15', 'Rapeseed' ],
    [ '2013-10-08', '2014-08-05', 'Wheat' ],
    [ '2014-09-10', '2014-09-11', 'Rapeseed' ],
]


# ==== Concaust method ====

# 3 months
npast = 91
nfuture = npast

data_scales = [ np.nanstd(s) for s in series ]

print("Computing the similarity matrices in RKHS")
Gx, Gy, index_map = concaust.series_Gxy(series, data_scales, npast, nfuture, qcflags=qc_data)

print("Embedding the states")
Gs = concaust.embed_states(Gx, Gy)

print("Computing the reduced dimension basis")
eigenvalues, basis, coords = concaust.spectral_basis(Gs)

print("Computing linear operators for gap filling")
scaled_coords = coords * eigenvalues
shift_op = concaust.shift_operator(scaled_coords, index_map)
expect_op = concaust.expectation_operator(scaled_coords, index_map, series)

print("Gap filling, states extrapolation and data fitting - first pass")
# This method can be roughly summarized as:
# - Start from the last known/valid point of an invalid data source
# - Use the shift/expectation operators to produce data values for all series,
#   the invalid data source and all other valid sources
# - Compare these predictions with the real data on all valid data sources. Use
#   a minimizer for adjusting the predicted states so that valid data is best 
#   recovered.
# - This provides a prediction for invalid series, that is consistent with
#   the dynamics (using shift op) over all the other valid series
# - The above works for forward predictions, such as at the end of the series. 
#   For invalid data gaps in the middle of the time range, predictions are run 
#   from both sides (forward and reverse) and linearly combined so that the
#   known end points match their actual values
exseries, exstates, qc_fill_pass1 = concaust.extrapolate_contiguous(scaled_coords, index_map, shift_op, expect_op, series, qc_data)


print("Gap filling, states extrapolation and data fitting - second pass")
# Re-build the causal states projection and the shift operators
# using the gap-filled states on the partial history ranges.
# The bad QC data will be re gap-filled.
# The minimizer ensured that the states match the observed data, even for these 
# states missing due to partial histories. Hence, what usually happens is 
# that states filled in these partial histories regions (before/after a missing 
# data gap, or at the beginning/end) predict data very close to the actual
# observations. This is one way to exploit all available data at our disposal.
# This second pass does very little change to the causal diffusion components,
# since the projection for the "new" states matching partial histories are
# consistent with their filled version in the "old" projection.
# However this really improves the quality of the gap-filled section in data
# space. One reason is that the interpolation range is much smaller, starting
# and ending just at the bad QC flags. In the first pass, these ranges were
# extended to missing states that could not be computed due to lacking full
# histories, hence gaps larger by (history length - 1) before/after the gap.
# It is possible to iterate in an third pass, etc, but the minimal gains 
# are not worth the cost
coords_ori, index_map_ori = coords, index_map
index_map = np.arange(len(qc_fill_pass1))[qc_fill_pass1>=1]
scaled_coords = exstates[qc_fill_pass1>=1, :]
shift_op = concaust.shift_operator(scaled_coords, index_map)
expect_op = concaust.expectation_operator(scaled_coords, index_map, series)
exseries, exstates, _ = concaust.extrapolate_contiguous(scaled_coords, index_map, shift_op, expect_op, series, qc_data)
index_map_ex = np.arange(len(timestamps), dtype=int)

gap_filled_series = []
for i, s in enumerate(series):
    gs = np.copy(s)
    bad_entries = np.isnan(s) # nan was set also for bad QC flags
    gs[bad_entries] = exseries[i][bad_entries]
    gap_filled_series += [ gs ]

# convert back to unscaled coordinates
exstates /= eigenvalues 

# === 3D and data series visualization for the article ====

# The sign of a basis eigenvector has no meaning - prefer this one for display
st = np.copy(exstates)
st[:,3] = - st[:,3]
st[:,2] = - st[:,2]

days_of_year = np.asarray([ ts.timetuple().tm_yday for ts in timestamps.astype(object) ])

cpos = plot_3D_pyvista(st[:, 1:4], eigenvalues[2:4]/eigenvalues[1], days_of_year, culture_dates, index_map_ex, title="Effective variables 1,2,3", return_cpos = True, cpos = np.asarray([(8.852597651251275, 11.047484271366736, 3.313895419195384), (0.02798216429757841, -0.039926141968956705, -0.5639586604423752), (-0.17183857903877114, -0.20056534608627188, 0.9644921174914962)]), snapshot='crop_123.png')

cpos = plot_3D_pyvista(st[:, [1,2,4]], eigenvalues[2:4]/eigenvalues[1], days_of_year, culture_dates, index_map_ex, title="Effective variables 1,2,4", return_cpos = True, cpos = np.asarray([(8.852597651251275, 11.047484271366736, 3.313895419195384), (0.02798216429757841, -0.039926141968956705, -0.5639586604423752), (-0.17183857903877114, -0.20056534608627188, 0.9644921174914962)]), snapshot='crop_124.png')

axs = plot_data_series(gap_filled_series, qc_data, timestamps, [variable_names[d] for d in data_var])

# Display the legend in separate window
import matplotlib.patches as mpatches
plt.figure()
plt.legend(handles=[mpatches.Patch(color=c, label=l) for l,c in culture_colors_ref.items()])


# === Below are calls to home-made data exploration tools ===
# === API is susceptible to change - may not work with the latest version ===


# But we plot the original data, the sensor measurements in the bad QC regions,
# to compare with the gap-filled versions
concaust.plot_series([data[k].to_numpy()[:,None] for k in data_var], extrapolations=gap_filled_series, timestamps=timestamps, qcflags=qc_data, labels=labels, title="Grignon field data, gap filled using 3 months states", filename='GRI_3months_gap_filled_data.html', display=True)


color_by_names = ['Culture type', 'Day of the year'] + labels

culture_colors = np.asarray(['#b3b3b3'] * len(timestamps))
iprev = 0
for first, last, c in culture_dates:
    culture_colors[np.logical_and(timestamps >= np.datetime64(first), timestamps <= np.datetime64(last))] = culture_colors_ref[c]


color_by_values = [culture_colors] + [np.asarray([ ts.timetuple().tm_yday for ts in timestamps.astype(object) ])] + [ np.squeeze(series[i]) for i in range(len(series)) ]

color_by_scales = [None, 'HSV'] + ['Rainbow'] * len(series)

# year annotations are also displayed for the extended states
index_map_ex = np.arange(len(timestamps), dtype=int)
seldates = [ci for ci,i in enumerate(index_map_ex) if (timestamps[i].astype(object).month-1)%3==0 and timestamps[i].astype(object).day==1]
annotations = ['%d-%02d' % (timestamps[i].astype(object).year, timestamps[i].astype(object).month) for i in index_map_ex[seldates]]
annoted_coords = exstates[seldates,:]
annotations_legend = 'Date'


concaust.plot_trajectories(scaled_coords/eigenvalues, index_map, eigenvalues, N = timestamps.shape[0], extrapolations = exstates, annotations=annotations, annoted_coords=annoted_coords, annotations_legend=annotations_legend, color_by_values=color_by_values, color_by_scales=color_by_scales, color_by_names=color_by_names, num_vars = 8, title="Grignon Field - RKHS states and trajectories - 3 months history", filename='GRI_3months_RKHS_states.html', display=True)


from plot_eigenvalues import plot_eigenvalues
plot_eigenvalues(eigenvalues,'Crop field', 'crop_eigenvalues')

axs = plot_data_series(gap_filled_series, qc_data, timestamps, [variable_names[d] for d in data_var])





