#!/usr/bin/env python3
"""
This is the code for the Continuous Causal States method
described in the papers:

  Discovering Causal Structure with Reproducing-Kernel Hilbert Space ε-Machines
  by Nicolas Brodu and James P. Crutchfield

  Inferring Kernel ε-Machines: Discovering Structure in Complex Systems
  by Nicolas Brodu and Alexandra M. Jurgens1

  and possibly more to come
  
This code is free software, according to the MIT license
Author & Contact: nicolas.brodu@inria.fr
"""

import numpy as np
from scipy import integrate
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection

import os
import sys
sys.path.insert(0, os.path.abspath('..'))
import concaust


# ==== simulation code and helper routine ====

def pendulum_sim(theta_velocity, t):
    theta_dot = theta_velocity[1]
    velocity_dot = -b/m*theta_velocity[1] - g/L*np.sin(theta_velocity[0])
    return theta_dot, velocity_dot

def plot_colored_by_scalar(x,y,c, **lc_kwargs):
    from matplotlib.collections import LineCollection
    # Thanks https://stackoverflow.com/questions/10252412/matplotlib-varying-color-of-line-to-capture-natural-time-parameterization-in-da
    # https://matplotlib.org/stable/gallery/lines_bars_and_markers/multicolored_line.html
    points = np.array([x,y]).transpose().reshape(-1,1,2)
    segs = np.concatenate([points[:-1],points[1:]],axis=1)
    default_kwargs = {"capstyle": "butt", "cmap": "viridis"}
    default_kwargs.update(lc_kwargs)
    lc = LineCollection(segs, **default_kwargs)
    lc.set_array(c)
    plt.gca().add_collection(lc)


def main_experiment(title, color = None):
    
    print("==== Processing experiment: %s ==== " % title)
    
    # Generate the data with the selected parameters
    
    t = np.linspace(0, t_max, int(t_max/delta_t))
    θ_ω = integrate.odeint(pendulum_sim, [θ_ini,ω_ini], t)
    # positions - no velocity, not a 1 to 1 mapping of state space
    # => require history for reconstruction
    x = L * np.sin(θ_ω[::sampling,0])
    y = L - L * np.cos(θ_ω[::sampling,0])
    t = t[::sampling]

    # Parameters for the causal state analysis

    # average pseudo-period = time between crossings of x==0 in the same direction
    npast = int(np.mean(np.diff(np.where(np.logical_and(x[:-1]>0, x[1:]<0))))) 
    nfuture = npast
    
    print(f"Using sequences past/future of {npast} samples")
    
    if use_GPU:
        import cupy as cp
        xcpu = x
        ycpu = y
        x = cp.asarray(x)
        y = cp.asarray(y)
    
    print("Building similarity matrices")
    Gx, Gy, index_map = concaust.series_Gxy([x, y], scale, npast, nfuture) #, kernelAggregation='sum')
    
    print("Embedding to get the similarity matrix between conditional distributions")
    Gs = concaust.embed_states(Gx, Gy)
    
    print("Find a reduced dimension embedding and extract the first coordinates")
    eigenvalues, basis, coords = concaust.spectral_basis(Gs,11)

    print("Spectral decay (in % of the first relevant eigenvalue)")    
    print(100 * eigenvalues[1:] / eigenvalues[1])
    
    if use_GPU:
        coords = cp.asnumpy(coords)
        eigenvalues = cp.asnumpy(eigenvalues)
        index_map = cp.asnumpy(index_map)
        x = xcpu
        y = ycpu

    return x,y,t,θ_ω,npast,nfuture, coords, eigenvalues, index_map, title, color


def plot_colored_by_time(x,y,t, nthick=None):
    # Thanks https://stackoverflow.com/questions/10252412/matplotlib-varying-color-of-line-to-capture-natural-time-parameterization-in-da
    points = np.array([x,y]).transpose().reshape(-1,1,2)
    segs = np.concatenate([points[:-1],points[1:]],axis=1)
    widths = np.ones(len(segs)) * 1.5
    if nthick is not None:
        widths[:nthick] = 2.5
    lc = LineCollection(segs, cmap=plt.get_cmap('jet'), linewidths=widths)
    lc.set_array(t)
    plt.gca().add_collection(lc)
    plt.xlim(x.min()*1.05, x.max()*1.05)
    plt.ylim(y.min()*1.05, y.max()*1.05)

def plot_main_experiment(x,y,t,θ_ω,npast,nfuture, coords, eigenvalues, index_map, title, color):
    #plt.figure()
    #plot_colored_by_time(x,y,t)
    #plt.axis('equal')
    #plt.title("%s - Motion of the pendulum" % title)
    #plt.show()

    plt.figure(layout='constrained')
    if color is None:
        plot_colored_by_time(θ_ω[::sampling,0][npast:-nfuture],θ_ω[::sampling,1][npast:-nfuture],t[npast:-nfuture],nfuture)
        #plt.plot(θ_ω[::sampling,0][npast:-nfuture],θ_ω[::sampling,1][npast:-nfuture],color='black')
        plt.plot(θ_ω[::sampling,0][:npast], θ_ω[::sampling,1][:npast], color='#b3b3b3')
        plt.plot(θ_ω[::sampling,0][-nfuture:], θ_ω[::sampling,1][-nfuture:], color='#b3b3b3')
        #plt.plot(θ_ω[::sampling,0][npast+nfuture], θ_ω[::sampling,1][npast+nfuture], '.k')
    else:
        plt.plot(θ_ω[::sampling,0],θ_ω[::sampling,1], color=color)
        
    plt.xlim(θ_ω[::sampling,0].min()*1.05, θ_ω[::sampling,0].max()*1.05)
    #plt.xticks([-np.pi/2, -np.pi/4, 0, np.pi/4, np.pi/2], ['$-\\pi/2$', '$-\\pi/4$', '0', '$\\pi/4$', '$\\pi/2$'])
    plt.xticks([-np.pi/2, -np.pi/4, 0, np.pi/4, np.pi/2], ['-π/2', '-π/4', '0', 'π/4', 'π/2'])
    plt.ylim(θ_ω[::sampling,1].min()*1.05, θ_ω[::sampling,1].max()*1.05)
    plt.yticks([-4,-2,0,2,4], ['-4','-2','0','2','4'])
    #plt.title("%s - Phase space" % title)
    plt.gca().tick_params(axis='x',labelsize=14)
    plt.gca().tick_params(axis='y',labelsize=14)
    plt.show()

    plt.figure(layout='constrained')
    if color is None:
        plot_colored_by_time((coords)[:,1],(coords)[:,2],t[index_map])
    else:
        plt.plot(coords[:,1],coords[:,2], color=color)
    plt.title("%s - Effective space" % title)
    plt.gca().tick_params(axis='x',labelsize=14)
    plt.gca().tick_params(axis='y',labelsize=14)
    plt.show()


#sys.exit(0)

# ==== Script entry point - calls the above with various parameters ====


if __name__ == '__main__':
    
    
    # https://en.wikipedia.org/wiki/Separatrix_(mathematics)
    # Separatrix equation in phase space : 1/2 ω² - g/L cos θ = g/L
    #   with ω = dθ/dt
    # If θ = -π/2 (our initial condition):
    #   1/2 ω² = g/L
    #   => ω² = 2g/L
    # Then, the initial velocity is mathematically exactly what is needed to 
    # stop the pendulum on the top point. This is extremely unstable and
    # numerically pointless to try to integrate exactly.
    # But the closer to that situation, and the longer the pendulum "stays"
    # on the top part before going back down, on either side.
    # This means, in our context, that longer and longer histories are needed.
    # Also, especially for velocities just below the critical one, the pendulum
    # reaches positions which are close to each other on both sides, but with
    # very distinct futures. As the pendulum stays there for a while, this 
    # creates portions of histories which are very close to each other.
    # So, the "scale" parameter, which sets relevant differences in data 
    # values, needs to be adjusted down to finer and finer values as the top
    # positions on either side get closer and closer.
    # On the other hand, in the dissipative case, the pendulum never actually
    # returns to the same conditions. Then, the scale needs to be adjusted 
    # so that histories with different energy levels in the spiral can be
    # compared to each other.
    # Below, the choice is made to use a scale=1 for that situation. For 
    # consistency we keep the same scale in the undamped scenarios, but then,
    # we use initial conditions further away from the separatrix.
    
    
    # ==== Physical constants ====
    m = 1 # mass (kg)
    L = 1 # length (m)
    g = 9.81 # gravity (m/s^2)
    tini = -np.pi/2  # Initial angle (rad)
    
    # ==== Numerical parameters ====
    t_max = 40      # max sim time (seconds)
    delta_t = 0.005 # time step size (seconds)
    sampling = 1    # retain 1 in every so many samples
    scale = 1       # for comparing data values -- see above
    
    use_GPU = False # enable this and install cupy if you have CUDA-compatible hardware
    b = 0
    π = np.pi
    
    levels_color = '#222222'
    separatrix_color = levels_color
    levels_sep = 1
    

    plt.figure()
    ω_sep = np.sqrt(4 * g/L)
    
    # Plot the separatrix    
    # 1/2 ω² - g/L cos θ = g/L
    # ω² = 2 * g/L * (1 + cos θ)
    θ = np.linspace(-2*π, 2*π, 500)
    plt.plot(θ, np.sqrt(2 * g/L * (1 + np.cos(θ))), '--', color=separatrix_color)
    plt.plot(θ, -np.sqrt(2 * g/L * (1 + np.cos(θ))), '--', color=separatrix_color)

    # H = 1/2 ω² - g/L cos θ is conserved
    # case for H = 0, used as initial condition also for the damped case
    #θ_ini = -π/2
    #ω_ini = 0
    # Equivalently, 1/2 ω² - g/L cos θ = 0 with θ=0 implies:
    θ_ini = 0
    ω_ref = np.sqrt(2 * g/L)
    ω_ini = ω_ref
    
    def plot_for_θ_ω_ini(linecolor, numturns = 1):
        t = np.linspace(0, 10, int(t_max/delta_t))
        θ_ω = integrate.odeint(pendulum_sim, [θ_ini,ω_ini], t)
        x = L * np.sin(θ_ω[::sampling,0])
        npts = numturns * int(np.max(np.diff(np.where(np.logical_and(x[:-1]>0, x[1:]<0)))))
        plt.plot(θ_ω[::sampling,0][:npts],θ_ω[::sampling,1][:npts],color=linecolor)

    plot_for_θ_ω_ini('blue')
    
    # Increment in ω for lines to plot on the diagram
    Δω = (ω_sep - ω_ini) / 2

    # Build one line between the separatrix and the above initial condition
    plot_intermediate_lines = False
    ω_ini += Δω
    if plot_intermediate_lines: plot_for_θ_ω_ini(levels_color)
    
    # Build lines below
    for i in range(4):
        ω_ini = ω_ref - Δω * (i+1)
        if plot_intermediate_lines or (i&1==1): plot_for_θ_ω_ini(levels_color)
    
    # Build one line on the other side of the separatrix
    θ_ini = -2*π # or 0, but for plotting the wavy curve it's better
    ω_ini = ω_sep + Δω
    if plot_intermediate_lines: plot_for_θ_ω_ini(levels_color, 2)

    # Go symmetric in ω on the other side of the separatrix
    ω_ini += Δω
    plot_for_θ_ω_ini('red', 2)
    # complete with partial lines visible above and on the sides
    ω_ini += Δω
    if plot_intermediate_lines: plot_for_θ_ω_ini(levels_color, 2)
    ω_ini += Δω
    plot_for_θ_ω_ini(levels_color, 2)
    
    # Complete with lines below the separatrix
    θ_ini = 2*π
    for i in range(4):
        ω_ini = -ω_sep - Δω * (i+1)
        if plot_intermediate_lines or (i&1==1): plot_for_θ_ω_ini(levels_color, 2)
        


    plt.xlim([-π*1.25, π*1.25])
    plt.xticks(np.linspace(-π*1.25,π*1.25,11), ['-5π/4','-π','-3π/4','-π/2','-π/4','0','π/4','π/2','3π/4','π','5π/4'])
    plt.ylim([-8.5,8.5])
    plt.yticks(np.linspace(-8,8,9))
    plt.show()


    #sys.exit(0)
    from plot_eigenvalues import plot_eigenvalues

    # ==== Dissipative case ====
    θ_ini = -π/2 # Initial angle
    ω_ini = 0    # Initial velocity
    b = 0.1      # damping value (kg/m^2-s)

    res_dissip = main_experiment("Dissipative case")
    plot_main_experiment(*res_dissip)
    plot_eigenvalues(res_dissip[7], "Dissipative case", 'pendulum_dissipative_eigenvalues')
 
    
    # ==== Undamped case ====
    b = 0
    res_below = main_experiment("Undamped case, θ_ini = -π/2, ω_ini = 0", 'blue')
    plot_main_experiment(*res_below)
    plot_eigenvalues(res_below[7], "Conservative case below $E_{sep}$", 'pendulum_conservative_below_eigenvalues')
    
    θ_ini = -2*π # or 0, but for plotting
    ω_ini = ω_sep * 2 - ω_ref
    res_above = main_experiment("Undamped case, θ_ini = 0, ω_ini above critical value", 'red')
    plot_main_experiment(*res_above)
    plot_eigenvalues(res_above[7], "Conservative case above $E_{sep}$", 'pendulum_conservative_above_eigenvalues')
    
    # Spyder tip for assessing the effect of GPU: 
    # %timeit -r1 -n1 run -i pendulum.py
    
    # Tip for better printing numbers
    # np.set_printoptions(suppress=True)
    
    # Spyder tip for producing better quality images
    # %matplotlib qt
    # run -i pendulum.py
    # then save as SVG
    
    
    
    
