#!/usr/bin/env python3
"""
This is the code for the Continuous Causal States method
described in the papers:

  Discovering Causal Structure with Reproducing-Kernel Hilbert Space ε-Machines
  by Nicolas Brodu and James P. Crutchfield

  Inferring Kernel ε-Machines: Discovering Structure in Complex Systems
  by Nicolas Brodu and Alexandra M. Jurgens1

  and possibly more to come
  
This code is free software, according to the MIT license
Author & Contact: nicolas.brodu@inria.fr
"""

import numpy as np
import scipy.linalg
import scipy.signal

# This is for running the script from the source code of the uninstalled package
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
# Otherwise you can just import concaust and skip the above 3 lines
import concaust


# ==========================================
# Plotting routines - The main code is below
# ==========================================

import pyvista as pv
import matplotlib.pyplot as plt

def compute_data_phase(minindices, maxindices, n):
    phase = np.full(n, np.nan)
    nmin = len(minindices)
    for i in range(nmin-1):
        m0 = minindices[i]
        m1 = minindices[i+1]
        themax = -1
        for m in maxindices:
            if m > m0 and m < m1:
                themax = m
                break
        if themax==-1:
            continue
        for j in range(m0,themax):
            phase[j] = np.pi * (j - m0) / (themax - m0)
        for j in range(themax,m1):
            phase[j] = np.pi + np.pi * (j - themax) / (m1 - themax)
    return phase

def polyline_from_points(points):
    poly = pv.PolyData()
    poly.points = points
    the_cell = np.arange(0, len(points), dtype=np.int_)
    the_cell = np.insert(the_cell, 0, len(points))
    poly.lines = the_cell
    return poly

def plot_3D_pyvista(coords, data_phase, title=None, snapshot=None, cpos=None):
    from matplotlib.colors import LinearSegmentedColormap
    
    p = pv.Plotter(window_size=[2000,2000], notebook=False, off_screen = bool(snapshot))
    pbr = False
    
    # The cyclic colormap is OK for plotting points, but segments between
    # points create artifacts. Example plotting the phase in a cycle of 
    # length 5: 0 1 2 3 4 0 1 2 ...
    # The segments between consecutive values (e.g. 0, 1) are interpolated 
    # (e.g. 0.1, 0.2...) for all pixels along the segment.
    # BUT at the end of the cycle (from 4 to 0) the pixels are interpolated
    # backwards. In terms of colors, this creates a crazy erroneous segment
    # running through all the cyclic palette backwards
    # The solution below avoids this issue by creating separate tubes,
    # fixing the end values so they overlap, and to repeat the colormap
    hsv3 = LinearSegmentedColormap.from_list("hsv3", ["#ff0000", "#ffa700", "#afff00", "#08ff00", "#00ff9f", "#00b7ff", "#0010ff", "#9700ff", "#ff00bf", "#ff0000", "#ffa700", "#afff00", "#08ff00", "#00ff9f", "#00b7ff", "#0010ff", "#9700ff", "#ff00bf", "#ff0000", "#ffa700", "#afff00", "#08ff00", "#00ff9f", "#00b7ff", "#0010ff", "#9700ff", "#ff00bf", "#ff0000"])
    data_phase = data_phase / (2 * np.pi)
    dphi = np.diff(data_phase)
    previ = 0
    for i,idx in enumerate(np.where(dphi<-0.5)[0] + 1):
        polyline = polyline_from_points(coords[previ:idx+1,1:4])
        dseg = (data_phase[previ:idx+1] + 1) / 3
        dseg[-1] = dseg[-2]
        dseg[0] = dseg[1]
        polyline["scalars"] = dseg
        tube = polyline.tube(radius=0.008)
        p.add_mesh(tube, smooth_shading=True, cmap=hsv3, ambient=0.5, pbr=pbr, clim=[0,1], opacity=0.2)
        p.add_mesh(polyline, smooth_shading=True, cmap=hsv3, ambient=0.5, render_points_as_spheres=True, style='points', point_size=30)
        previ = idx-1
    
    scale = 0.5
    shaft_radius = 0.02
    tip_radius = 0.05
    tip_length = 0.2
    plane_center = np.asarray([0,0,-2.25])
    frame_origin = plane_center + np.asarray([-0.3,-2,0.1])
    p.add_mesh(
        pv.Arrow(start=frame_origin, direction=(1,0,0), tip_length=tip_length, tip_radius=tip_radius, shaft_radius=shaft_radius, scale=scale),
        color='black',ambient=0,smooth_shading=False
    )
    p.add_mesh(
        pv.Arrow(start=frame_origin, direction=(0,1,0), tip_length=tip_length, tip_radius=tip_radius, shaft_radius=shaft_radius, scale=scale),
        color='black',ambient=0,smooth_shading=False
    )
    p.add_mesh(
        pv.Arrow(start=frame_origin, direction=(0,0,1), tip_length=tip_length, tip_radius=tip_radius, shaft_radius=shaft_radius, scale=scale),
        color='black',ambient=0,smooth_shading=False
    )
    
    p.add_mesh(
        pv.Plane(center=plane_center, direction=(0,0,1), i_size=6, j_size=6),
        color='#eeeeee',ambient=0,smooth_shading=True, pbr=pbr
    )
    
    p.enable_ssao()
    p.enable_anti_aliasing('ssaa')
    p.hide_axes()
    p.remove_scalar_bar()
    p.set_position([0.9,0.9,1.1])
    p.set_viewup([0,0,1])
    
    #p.export_gltf('sunspots.gltf') # to import in Blender
    #p.export_html("test_pv.html")
    if cpos is None:
        cpos = [(9.045539299059834, 0, 2.821684124548503),(-0.06366092297680231, 0.0, -0.05889586782942299),(-0.30154477917071953, 0, 0.9534483462944139)]
        
    if title is not None:
        p.add_text(title)
        
    camera_pos = p.show(return_cpos=True, cpos=cpos, title=title)

    if snapshot:
        p.screenshot(snapshot)

    return camera_pos


def plot_sunspot_data(sunspots, minindices, maxindices, npast, nfuture, coords, index_map):
    fig, ax1 = plt.subplots(layout='constrained')
    ax2 = ax1.twinx()
    
    n = sunspots.size
    ax1.plot(index_map, sunspots[index_map], '#b8b8ff')
    
    ax1.plot(np.arange(npast), sunspots[:npast], color='lightgrey')
    ax1.plot(np.arange(n-nfuture,n), sunspots[-nfuture:], color='lightgrey')


    ax2.plot(index_map+132, coords[:,3], '#ffaaff', linewidth=1.5, linestyle='solid')

    ax2.plot(index_map, coords[:,3], 'black', linewidth=1.5)
    
    for mi in minindices:
        ax1.plot([mi, mi], [-15, -5], c='#ff0000')
    for ma in maxindices:
        #ax1.plot([ma, ma], [-15, -5], c='#00ffff')
        ax1.scatter([ma], [-10], c='#00aaaa', s=1.5)
    
    xdelta = 10
    xlocs = np.squeeze(np.where((month==1) & (year%xdelta==0)))
    plt.xticks(xlocs, ["%d" % y if y % 50==0 else "" for y in year[xlocs]])
    ax1.set_ylim([-20,420])
    ax2.set_ylim([-3,2.5])
    plt.xlim([0,n])
    ax1.tick_params(axis='y',labelsize=14)
    ax2.tick_params(axis='y',labelsize=14)
    ax1.tick_params(axis='x',labelsize=14)
    plt.show()

#plot_sunspot_data(sunspots, minindices, maxindices, npast, nfuture, coords, index_map)
#sys.exit(0)



# =========
# Main code
# =========
if __name__ == '__main__':
    
    # Sunspots monthly series from SILSO. Data is under the CC BY-NC4.0 license
    # File format documentation and updates at http://sidc.oma.be/silso/datafiles
    sraw = np.genfromtxt('data/sunspots/SN_m_tot_V2.0.csv', delimiter=';')
    year = sraw[:,0]
    month = sraw[:,1]
    sunspots = sraw[:,3]
    
    # ==== Data loading and series preparation ====

    # A low-pass filter is applied to the sunspots series to get rid of the
    # fast variations of the sunspot numbers. This yields a much better
    # estimate of the min/max years for each "cycle".
    # Traditionally (early 20th century) a simpler smoothing average was used,
    # with remaining artifacts. See also https://www.sidc.be/SILSO/faq6
    smoothed_sunspots = scipy.signal.sosfiltfilt(scipy.signal.butter(4, 1/44, 'lowpass', fs=1, output='sos'), sunspots)
    
    # Min/max years(/indices) in each cycle
    maxindices = scipy.signal.find_peaks(smoothed_sunspots)[0]
    minindices = scipy.signal.find_peaks(-smoothed_sunspots)[0]
    
    # Take the average cycle height as the data scale
    cycle_heights = []
    for i in range(len(minindices)-1):
        m0 = minindices[i]
        m1 = minindices[i+1]
        srange = sunspots[np.max([0,m0-24]):m1+24]
        cycle_heights += [ np.max(srange) - np.min(srange) ]
    scale = np.mean(cycle_heights)

    data_phase = compute_data_phase(minindices, maxindices, len(sunspots))
    
    # Compute only the first eigenvalues / vectors -- faster
    num_basis = 10
    # Analyze the monthly data at the temporal scale of 22 years
    # (a full magnetic cycle)
    npast = 22*12
    nfuture = 22*12
    
    # Demo of GPU usage - set to true if needed
    # The computations should run fast enough on any recent CPU
    use_GPU = True
    if use_GPU:
        import cupy as cp
        sunspots_CPU = sunspots
        sunspots = cp.asarray(sunspots)
    
    
    # Compute the kernelized similarity matrices
    # For each row in the matrices (including coords below), the index map
    # gives the "present" value in the sunspots series. In this example, 
    # this is just a consecutive range of data starting at index npast-1
    Gx, Gy, index_map = concaust.series_Gxy(sunspots, scale, npast, nfuture)

    # Compute the state similarity matrix
    Gs = concaust.embed_states(Gx, Gy)
        
    # Compute a spectral basis for representing the causal states. See the paper.
    # Note that the smallest eigenvalues are irrelevant (below the precision
    # of the probability distribution RKHS mapping). The default is to retain 
    # half the components, but smaller numbers are faster to compute
    eigenvalues, basis, coords = concaust.spectral_basis(Gs, num_basis=num_basis)
    
    # Bring the results back to CPU if needed
    if use_GPU:
        coords = cp.asnumpy(coords)
        eigenvalues = cp.asnumpy(eigenvalues)
        index_map = cp.asnumpy(index_map)
        sunspots = sunspots_CPU
       
    # Plot results
    plot_sunspot_data(sunspots, minindices, maxindices, npast, nfuture, coords, index_map)
    plot_3D_pyvista(coords, data_phase[index_map], snapshot="sunspots_states_3D_visualization.png")
    
    from plot_eigenvalues import plot_eigenvalues
    plot_eigenvalues(eigenvalues, "Sunspots", 'sunspots_eigenvalues')
 






