#!/usr/bin/env python3

from mpl_toolkits.mplot3d import Axes3D
from scipy.integrate import odeint
import matplotlib.pyplot as plt
import numpy as np

# This is for running the script from the source code of the uninstalled package
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
# Otherwise you can just import concaust and skip the above 3 lines
import concaust

# Parameters
# ==========

dimrange = [5] #np.arange(5,101,5)
npast = 5
nfuture = 5
# total number of sample pairs
N = 5000
# subdivide this number N in that many independent series taken 
# on the attractor after a transient, starting from a random point
nser = 10
# dimension for the final space using random projections
# use -1 to disable projection and use the original space
embed_dims = 1000
# noise added in that random space - effectively creating data spanning all
# dimensions, not just the low-dimensional initial structure
# The scaling is so the noise level is the same when data are projected back
# in the original space, so tests with different embed_dims are comparable
# use 0 to keep the data clean
added_noise_std = 0.25/np.sqrt(embed_dims)

# plots
plotdims = [5] #dimrange

def L96(x, t, num_dims, forcing_constant):
    """
    Lorenz 96 model with constant forcing
    Adapted from Wikipedia reference implementation
    https://en.wikipedia.org/wiki/Lorenz_96_model
    """
    d = np.zeros(num_dims)
    for i in range(num_dims):
        d[i] = (x[(i + 1) % num_dims] - x[i - 2]) * x[i - 1] - x[i] + forcing_constant
    return d
    

eigenvalues_per_ndim = []

for ndim in dimrange:

    # random seed for reproducible results - reset per ndim loop, so
    # these could be reproduced independently of each other
    seed = 42
    rng = np.random.default_rng(seed)
    
    # Forcing constant in the L96 definition
    forcing = 8
    
    # data scale - itself scaled with the unit cube diagonal
    # in the embedding dimension
    scale = 10 * np.sqrt(embed_dims if embed_dims > 0 else ndim)

    # generate data
    npts_per_ser = (N // nser) + npast + nfuture - 1
    dt = 0.01
    transient = 10
    tmax = npts_per_ser * dt + transient
    
    # random projection in high-dim behaves like a pseudo-basis of 
    # nearly orthogonal vectors
    if embed_dims>0:
        random_proj = rng.standard_normal((ndim, embed_dims)) / embed_dims
    else:
        random_proj = np.eye(ndim)
    
    clean_x = []
    clean_s = []
    series = []
    
    for i in range(nser):
        #x0 = forcing * np.ones(ndim)  # Initial state (equilibrium)
        # Add a perturbation to the first variable
        #x0[0] += 0.1 * rng.standard_normal()
        x0 = rng.uniform(-forcing/2, +forcing/2, ndim)
        t = np.arange(0.0, tmax, dt)
        
        x = odeint(L96, x0, t, args=(ndim,forcing))
        x = x[round(transient/dt):,:]
        clean_x += [ x ]
        
        s = x @ random_proj
        
        clean_s += [ s.copy() ]
        
        s += rng.standard_normal(s.shape) * added_noise_std
        
        series += [ s ]

    
    # Use [series] as these are different blocks from a single data source
    # See the series_Gxy documentation
    Gx, Gy, index_map = concaust.series_Gxy([series], scale, npast, nfuture)
    
    Gs = concaust.embed_states(Gx, Gy)
    
    eigenvalues, basis, coords = concaust.spectral_basis(Gs)

    eigenvalues_per_ndim += [ eigenvalues ]


    if ndim in plotdims:
        
        # plot the original clean data
        fig = plt.figure()
        ax = fig.gca(projection="3d")
        for i in range(nser):
            ax.plot(clean_x[i][:,0], clean_x[i][:,1], clean_x[i][:,2])
        ax.set_xlabel("$x_1$", fontsize=16)
        ax.set_ylabel("$x_2$", fontsize=16)
        ax.set_zlabel("$x_3$", fontsize=16)
        #plt.title(f"Original {ndim}-dimensional data, first 3 dimensions", fontsize=18)
        plt.show()

        # plot the noisy data, projected back into the original space
        rpinv = np.linalg.pinv(random_proj)
        fig = plt.figure()
        ax = fig.gca(projection="3d")
        for i in range(nser):
            backproj = series[i] @ rpinv
            ax.plot(backproj[:, 0], backproj[:, 1], backproj[:, 2])
        ax.set_xlabel("noisy $x_1$", fontsize=16)
        ax.set_ylabel("noisy $x_2$", fontsize=16)
        ax.set_zlabel("noisy $x_3$", fontsize=16)
        #plt.title(f"Noisy {embed_dims if embed_dims>0 else ndim}-dimensional data,\n projected back to the original space", fontsize=18)
        plt.show()
        
        #split_pts = np.concatenate((np.where(np.diff(index_map)>1)[0]+1, [N]))
        split_pts = np.arange(1,nser+1) * (N//nser)
        fig = plt.figure()
        ax = fig.gca(projection="3d")
        beg = 0
        for sp in split_pts:
            # Note: these coordinates may not correspond to the original,
            # the algorithm only finds equivalent variables holding
            # information. Intuitively, this is similar for example to a 
            # cartesian-polar coordinate transform, except that here, we
            # do not a priori know the transform
            ax.plot(coords[beg:sp, 1], coords[beg:sp, 2], coords[beg:sp, 3])
            beg = sp
        ax.set_xlabel("$c_1$", fontsize=16)
        ax.set_ylabel("$c_2$", fontsize=16)
        ax.set_zlabel("$c_3$", fontsize=16)
        #plt.title("First 3 variables of the inferred attractor", fontsize=18)
        plt.show()
        
        #fig = plt.figure()
        #neig = len(eigenvalues[1:dimrange[-1]+6])
        #plt.plot(np.arange(1,neig+1), eigenvalues[1:neig+1]/eigenvalues[1])
        #plt.title("Eigenvalues, % of the first")
        #plt.show()


#np.savez_compressed('lor96_eigval_highdim_noisy.npz',eigenvalues_per_ndim=eigenvalues_per_ndim,dimrange=dimrange)


# post-processing code, put here for convenience
# can be activated if needed for reproducing plots in the paper
if False:
    f=np.load('lor96_eigval_highdim_noisy.npz')
    eigenvalues_per_ndim=f['eigenvalues_per_ndim']
    dimrange=f['dimrange']
    step=dimrange[1]-dimrange[0]
    r = np.arange(dimrange[0], dimrange[-1]+2*step, step)
    plt.figure()
    for i,d in enumerate(dimrange):
        plt.plot(np.arange(1,r[-1]+1), eigenvalues_per_ndim[i][1:r[-1]+1]/eigenvalues_per_ndim[i][1]*100, label=f"dim={d}")
    plt.xticks(np.arange(0,r[-1],step), fontsize=14)
    plt.xlim((0,r[-1]))
    plt.yticks(np.arange(0,101,10), fontsize=14)
    plt.ylim((0,100))
    plt.xlabel("Eigenvalue index", fontsize=16)
    plt.ylabel("% of the first eigenvalue", fontsize=16)
    plt.legend(fontsize=12)
    plt.show()


# spectral gaps - no graphics required
# formatting for easier copy-pasting into latex
if True:
    
    # leftover value after the gap, % of the 1st eigenvalue
    print(" & ".join(["%.2f" % (eigenvalues[ndim+1]/eigenvalues[1]*100) for eigenvalues, ndim in zip(eigenvalues_per_ndim, dimrange)]))
    
    # relative spectral gap ev(N)/ev(N+1) 
    print(" & ".join(["%.1f" % (eigenvalues[ndim]/eigenvalues[ndim+1]) for eigenvalues, ndim in zip(eigenvalues_per_ndim, dimrange)]))








