#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Example code for the paper :
  Discovering Causal Structure with Reproducing-Kernel Hilbert Space ε-Machines
  by Nicolas Brodu and James P. Crutchfield

This code is free software, according to the MIT license
Author & Contact: nicolas.brodu@inria.fr
"""

import numpy as np
import scipy
import scipy.linalg
import scipy.integrate

# This is for running the script from the source code of the uninstalled package
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
# Otherwise you can just import concaust and skip the above 3 lines
import concaust

import argparse

parser = argparse.ArgumentParser(description='Process an experiment with the Lorenz attrator.')
parser.add_argument('--seed', default=42, type=int, help='run exp with that seed')
parser.add_argument('--added_noise', default=0, type=float, help="Noise level (isotropic gaussian, std-dev) to add the time series generated from the Lorenz equations")
parser.add_argument('--sde_noise', default=0, type=float, help="Noise level (isotropic gaussian, std-dev) to add to each dX/dt component, turning the Lorenz equations into a SDE")
parser.add_argument('--npast', default=5, type=int, help='past series size')
parser.add_argument('--nfuture', default=5, type=int, help='future series size')
parser.add_argument('--subsampling', default=1, type=int, help='retain 1 every X samples')
parser.add_argument('--dt', default=0.01, type=float, help='time step for integrating the ODE')
parser.add_argument('--ntransient', default=2000, type=int, help='discard that number of points after the random initial condition')
parser.add_argument('--N', default=10000, type=int, help='Number of training samples')
parser.add_argument('--P', default=100, type=int, help='Number of validations samples')
parser.add_argument('--Psep', default=50, type=int, help="Separate each validation sample by this number of time steps, so they do not appear correlated")
parser.add_argument('--num_preds', default=100, type=int, help='Number of predictions that will be computed')
parser.add_argument('--pred_sep', default=5, type=int, help="Separate predictions that many steps. Operators are exponentiated in between.")
parser.add_argument('--verbose', default=0, type=int, help='display progress, increasing levels of')
parser.add_argument('--rawbw', default=1.1, type=float, help='scale for the raw data kernel exp(- 0.5 * diff**2 / rawbw**2)')
parser.add_argument('--kbw', default=40*np.sqrt(2.), type=float, help='scale for the reproducing kernel')
parser.add_argument('--decay', default=0.01, type=float, help='decay parameter for scaling the influence of past/future values')
parser.add_argument('--nbasis', default=0, type=int, help="number of components to retain for the operator basis. default = all")
parser.add_argument('--skip_raw_data', default=0, type=float, help="Whether to skip raw data computations")

args = parser.parse_args()
globals().update(args.__dict__)

seed=args.seed
added_noise=args.added_noise
sde_noise=args.sde_noise
npast=args.npast
nfuture=args.nfuture
subsampling=args.subsampling
dt=args.dt
ntransient=args.ntransient
N=args.N
P=args.P
Psep=args.Psep
num_preds=args.num_preds
pred_sep=args.pred_sep
verbose=args.verbose
rawbw=args.rawbw
kbw=args.kbw
decay=args.decay
nbasis=args.nbasis
skip_raw_data=args.skip_raw_data


np.set_printoptions(precision=5)
np.set_printoptions(suppress = True)
np.set_printoptions(linewidth=100000)

rng = np.random.default_rng(seed)

# Lorenz parameters
sigma, beta, rho = 10, 8./3., 28

def lorenz_equations(X, t):
    u, v, w = X
    up = -sigma*(u - v)
    vp = rho*u - v - u*w
    wp = -beta*w + u*v
    return np.asarray([up, vp, wp])

def stochastic_noise(X, t):
    # sdeint uses m independent Wiener processes
    # and asks that we specify the dim x m matrix
    # here dim = m = 3. We keep all coordinates independent
    # with an isotropic noise of fixed standard deviation
    return np.diag([sde_noise,sde_noise,sde_noise])

print("=== Arguments ===")
print(args)
print("=================")

u0 = rng.uniform(-20,20)
v0 = rng.uniform(-25,25)
w0 = rng.uniform(0,45)

window_size = npast + nfuture

# total number of time points
series_length = (N + window_size-1) * subsampling

if verbose: print('Generating the Lorenz attractor training data...')
t = np.linspace(0, (ntransient+series_length-1) * dt, ntransient+series_length)

if sde_noise>0:
    import sdeint
    lor = sdeint.itoint(lorenz_equations, stochastic_noise, (u0, v0, w0), t)
else:
    lor = scipy.integrate.odeint(lorenz_equations, (u0, v0, w0), t)

series = lor[ntransient::subsampling,:]

if added_noise>0:
    series += rng.standard_normal(series.shape) * added_noise

if verbose: print('computing gram matrices')

Gx, Gy, index_map = concaust.series_Gxy(series, kbw, npast, nfuture, decay)

if verbose: print('building the states')

Gs = concaust.embed_states(Gx, Gy)


if skip_raw_data==0:
    
    # VBDM is Variable-Bandwidth Diffusion Maps. The C++ code was validated
    # to reproduce the reference matlab implementation from 
    # Nonparametric forecasting of low-dimensional dynamical systems
    # by Tyrus Berry, Dimitrios Giannakis and John Harlim
    # Physical Review E, 91(3), 032915
    
    if verbose: print('building the shift operator in data space')
    dimension_d, density_d, eigenvalues_d, eigenvectors_d, invariant_d = concaust.VBDM(series, op_type = 2, debug=np.clip(verbose-1,0,10))
    sd2im_d = invariant_d / density_d
    normalized_basis_d = eigenvectors_d * sd2im_d[:, None]
    normalizer_d = np.mean(normalized_basis_d, 0)
    shift_op_d = eigenvectors_d[1:,:].T  @  normalized_basis_d[:-1,:]
    shift_op_d /= series.shape[0] - 1
    
    # This is the function we wish to take the expectation of. That expectation is 
    # taken according to the distribution which is evolved by the shift operator.
    # It respects the steps introduced above:
    # - 1. representation by values on each data sample
    #      => the function f(x,y,z) = x,y,z is here simply the values which we wish to recover
    #      so, series[:N,:] give these samples values... at each sampling point !
    # - 2. projection on the functional basis
    # Below, we will simply perform an dot product <f,c> = meanFun @ coefs
    # in order to recover the expectation sum_k density_k * functional_k
    # The division by the number of terms (converts sum->expectation) is included here
    meanFun_d = series.T @ normalized_basis_d / normalized_basis_d.shape[0]


# Now the real stuff
if verbose: print('Computing the spectral basis')

eigenvalues, basis, coords = concaust.spectral_basis(Gs)

if verbose: print('Computing the shift operator')

shift_op = concaust.shift_operator(coords, index_map)

if verbose: print('Computing the expectation operator')

expect_op = concaust.expectation_operator(coords, index_map, series)


# =============
#  Predictions
# =============

# We generate another series and use a completely different starting point
# for the validation set.

u1 = rng.uniform(-20,20)
v1 = rng.uniform(-25,25)
w1 = rng.uniform(0,45)

series_length = (P * Psep + window_size-1 + num_preds*pred_sep) * subsampling

if verbose: print('Generating the Lorenz attractor validation data...')
t = np.linspace(0, (ntransient+series_length-1) * dt, ntransient+series_length)

if sde_noise>0:
    lorvalid = sdeint.itoint(lorenz_equations, stochastic_noise, (u1, v1, w1), t)
else:
    lorvalid = scipy.integrate.odeint(lorenz_equations, (u1, v1, w1), t)

valid_series = lorvalid[ntransient::subsampling,:]

if added_noise>0:
    clean_valid = np.copy(valid_series)
    valid_series += rng.standard_normal(valid_series.shape) * added_noise
    
# Find the inner products for each PAST sequence in the validation data set
# This builds one column per such past sequence. 
# Each row is associated to a kernel coefficient for each training data entry
kxvalid = np.concatenate(
        [ concaust.new_kernel_vectors_from_time_series(
                valid_series[itest:itest+npast, :], 
                series, npast, nfuture, scale_x, decay)
          for itest in range(0,P*Psep,Psep)
        ], axis=1)
    
if verbose: print('Projecting the new data ...')
omega_new = concaust.NNLS(Gx, kxvalid, max_iter=1000, verbose=verbose).T

pred_discrepancy = np.zeros(num_preds)
pred_discrepancy_var = np.zeros(num_preds)
pred_discrepancy_min = np.zeros(num_preds) + np.inf
pred_discrepancy_max = np.zeros(num_preds) - np.inf
    
if skip_raw_data==0:
    pred_discrepancy_d = np.zeros(num_preds)
    pred_discrepancy_var_d = np.zeros(num_preds)
    pred_discrepancy_min_d = np.zeros(num_preds) + np.inf
    pred_discrepancy_max_d = np.zeros(num_preds) - np.inf

if verbose: print('Computing the predictions accuracy...')
for itest in range(0,P*Psep,Psep):
    # start with the "present" = the last entry in the past sequence
    # in the raw data case, the prediction starts at index npast-1+1=npast
    # which is, also, the first entry in the first future sequence
    vidx = npast - 1 + itest

    # Initial distribution to evolve.
    # - 1. defined as values expressed for each training data sample.
    #      Here, use a fixed probability mass, spread with a Gaussian kernel
    #      with pre-set bandwidth. Preliminary experiment cross-validated
    #      the default bandwith of this script with a reduced data set
    #      The use of the normalizer below allows to use an unnormalized
    #      gaussian kernel here, so long as it is of fixed bandwidth
    
    if skip_raw_data==0:
        broadcasted_diff = series - valid_series[vidx, :]
        
        tsqdiff = np.sum(broadcasted_diff ** 2, 1)

        p0_d = np.exp(- tsqdiff / rawbw**2)
        
        
        # - 2. projected on the functional basis, for the shift operator to act on
        #      This is the formula for the VBDM operator, see the above ref.
        coefs_d = eigenvectors_d.T @ (p0_d[:,None] / density_d[:,None]) / eigenvectors_d.shape[0]
        coefs_d /= normalizer_d @ coefs_d
    
    p0 = omega_new[itest//Psep,:]
    
    p0 /= np.sum(p0)
    
    coefs = right_eigenvectors[:,:nbasis].T @ p0[:,None]

    # "clean" reference data to compare against
    if sde_noise>0:
        # Start from the "present" point, then integrate without noise
        t = np.linspace(0, num_preds*pred_sep* dt, num_preds*pred_sep+1)
        clean_next = scipy.integrate.odeint(lorenz_equations, valid_series[vidx,:], t)
        # Subsample every pred_sep
        clean_ref = clean_next[1:1+num_preds*pred_sep:pred_sep, :]
    elif added_noise>0:
        clean_ref = clean_valid[vidx+1:vidx+1+num_preds*pred_sep:pred_sep, :]
    else:
        clean_ref = valid_series[vidx+1:vidx+1+num_preds*pred_sep:pred_sep, :]


    # make predictions for num_preds in the future
    for ipred in range(num_preds):
        if skip_raw_data==0:
            # apply the shift operator
            coefs_d = shift_op_d @ coefs_d
            coefs_d /= normalizer_d @ coefs_d
            # recover a triplet from the evolved coefficients
            triplet_d = meanFun_d @ coefs_d
            # prepare for next time
            for i in range(pred_sep-1):
                coefs_d = shift_op_d @ coefs_d
                coefs_d /= normalizer_d @ coefs_d
            # compare with the validation data
            disc_d = np.linalg.norm(triplet_d - clean_ref[ipred,:][:,None])
            pred_discrepancy_d[ipred] += disc_d
            pred_discrepancy_var_d[ipred] += disc_d * disc_d
            pred_discrepancy_min_d[ipred] = np.min((pred_discrepancy_min_d[ipred], disc_d))
            pred_discrepancy_max_d[ipred] = np.max((pred_discrepancy_max_d[ipred], disc_d))
            
            if verbose>1 and ipred==0:
                print("Validation %d/%d (DATA 0)" % (itest//Psep, P), valid_series[vidx+1+ipred*pred_sep,:], triplet_d.T, disc_d)
        
        # same thing using the states rkhs
        #triplet = meanFun @ (coefs / coefs[0,0])
        triplet = meanFun[:,:nbasis] @ coefs
        err = np.linalg.norm(triplet - clean_ref[ipred,:][:,None])
        pred_discrepancy[ipred] += err
        pred_discrepancy_var[ipred] += err * err
        pred_discrepancy_min[ipred] = np.min((pred_discrepancy_min[ipred], err))
        pred_discrepancy_max[ipred] = np.max((pred_discrepancy_max[ipred], err))
            
        if verbose and not err>=0:
            print("nan at itest=%d, ipred=%d" % (itest//Psep, ipred))
            input()
        # shift after - meanfun was defined on the first value in the future sequence
        for i in range(pred_sep):
            coefs = shift_op @ coefs
            coefs /= coefs[0,0]

        if verbose>1 and ipred==0:
            print("Validation %d/%d (0)" % (itest//Psep, P), valid_series[vidx+1+ipred*pred_sep,:], triplet.T, err)
        if verbose>2 and ipred==1:
            print("Validation %d/%d (1)" % (itest//Psep, P), valid_series[vidx+1+ipred*pred_sep,:], triplet.T, err)
        if verbose>3 and ipred==num_preds//2:
            print("Validation %d/%d (num_preds//2)" % (itest//Psep, P), valid_series[vidx+1+ipred*pred_sep,:], triplet.T, err)
        if verbose>4 and ipred==num_preds-1:
            print("Validation %d/%d (num_preds-1)" % (itest//Psep, P), valid_series[vidx+1+ipred*pred_sep,:], triplet.T, err)

print("eigenvalues=", np.array2string(eigenvalues[:11], separator=', '))

if skip_raw_data==0:
    pred_discrepancy_d /= P
    pred_discrepancy_var_d = np.sqrt((pred_discrepancy_var_d - pred_discrepancy_d**2 * P) / (P - 1))
    if verbose>1:
        print("dev_d=", np.array2string(pred_discrepancy_var_d, separator=', '))
        print("min_d=", np.array2string(pred_discrepancy_min_d, separator=', '))
        print("max_d=", np.array2string(pred_discrepancy_max_d, separator=', '))
    print("avg_d=", np.array2string(pred_discrepancy_d, separator=', '))

pred_discrepancy /= P
pred_discrepancy_var = np.sqrt((pred_discrepancy_var - pred_discrepancy**2 * P) / (P - 1))
if verbose>1:
    print("dev=", np.array2string(pred_discrepancy_var, separator=', '))
    print("min=", np.array2string(pred_discrepancy_min, separator=', '))
    print("max=", np.array2string(pred_discrepancy_max, separator=', '))
print("avg=", np.array2string(pred_discrepancy, separator=', '))

np.savez_compressed('lores_%g_%g.npz' % (sde_noise, added_noise),
    series = series,
    valid_series = valid_series,
    eigval = eigenvalues[:11],
    rev = right_eigenvectors[:,:11],
    dev_d = pred_discrepancy_var_d,
    min_d = pred_discrepancy_min_d,
    max_d = pred_discrepancy_max_d,
    avg_d = pred_discrepancy_d,
    dev_r = pred_discrepancy_var,
    min_r = pred_discrepancy_min,
    max_r = pred_discrepancy_max,
    avg_r = pred_discrepancy,
    )
