#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Example code for the paper :
  Discovering Causal Structure with Reproducing-Kernel Hilbert Space ε-Machines
  by Nicolas Brodu and James P. Crutchfield

This code is free software, according to the MIT license
Author & Contact: nicolas.brodu@inria.fr
"""

import numpy as np
import scipy.linalg as lin
import scipy.integrate
from numpy.random import default_rng
from numba import jit, njit, prange

import concaust

import argparse

parser = argparse.ArgumentParser(description='Perform an experiment with the Mess3 process.')
parser.add_argument('--seed', default=42, type=int, help='run exp with that seed')
parser.add_argument('--npast', default=15, type=int, help='past series size')
parser.add_argument('--nfuture', default=1, type=int, help='future series size')
parser.add_argument('--bandwidth', default=0.1, type=float, help='bandwidth (0=auto)')
parser.add_argument('--N', default=5000, type=int, help='Number of training samples')
parser.add_argument('--ntransient', default=1000, type=int, help='first discard that number of samples when generating the series')

args = parser.parse_args()
globals().update(args.__dict__)

print("=== Mess3 experiment with seed = %d ====" % seed)
print(args)

np.set_printoptions(precision=5)
np.set_printoptions(suppress = True)
np.set_printoptions(linewidth=100000)

"""
Thanks Alexandra Jurgens for the code to generate the non-unifilar mess3 
machine, its mixed states and the symbolic series.
"""
@njit
def mess3_transitions(a, x):
    """Machine introduced by Sarah in the original dimensions paper."""
    b = (1-a)/2.0
    y = 1 - 2.0*x
    return [np.array([[a*y, b*x, b*x],
                      [a*x, b*y, b*x],
                      [a*x, b*x, b*y]]),
            np.array([[b*y, a*x, b*x],
                      [b*x, a*y, b*x],
                      [b*x, a*x, b*y]]),
            np.array([[b*y, b*x, a*x],
                      [b*x, b*y, a*x],
                      [b*x, b*x, a*y]])]
@jit(forceobj=True)
def prob(Z, T):
    """Return the probability of map T given a point Z on the n-simplex."""
    Z = Z/np.sum(Z)
    return np.sum(np.matmul(Z, T))

@jit(forceobj=True)
def evolve(Z, T, n=1):
    """Given a matrix T, evolve a point Z on the n-simplex and return it."""
    Z = Z/np.sum(Z)
    for i in range(n):
        if prob(Z, T) != 0.0:
            Z = np.matmul(Z, T)/prob(Z, T)
        else:
            Z = Z
    return Z

@jit(forceobj=True)
def generate_mixed_state_process_symbolic_series(rng, transitions, N=10000, N_trans=5000):
    """Generate and store N symbols of the process, throwing away the first N_trans pointransitions."""
    data = []
    M = len(transitions[0])
    N_matrices = range(len(transitions))

    Z = [rng.random() for i in range(M)]
    Z = Z/np.sum(Z)
    for n in range(N_trans):
        p = [prob(Z, T) for T in transitions]
        sym = rng.choice(N_matrices, p=p)
        T = transitions[sym]
        Z = evolve(Z, T)
    for n in range(N):
        p = [prob(Z, T) for T in transitions]
        sym = rng.choice(N_matrices, p=p)
        T = transitions[sym]
        Z = evolve(Z, T)
        data.append(sym)
    return data



window_size = npast + nfuture

rng = np.random.default_rng(seed)

series_length = N + window_size-1

print("generating the series")
transitions = mess3_transitions(0.6, 0.15)
series = np.array(generate_mixed_state_process_symbolic_series(rng, transitions, series_length, ntransient), np.float64)

print("generating gram matrices")
Gx, Gy, scale_x, scale_x = concaust.gram_matrices_from_time_series(series[:N + window_size-1],npast,nfuture,bandwidth)

N = Gx.shape[0]

print("computing Gs")
Omega = scipy.linalg.solve(Gx + np.diag(np.ones(Gx.shape[0])*1e-6), Gx, assume_a = "pos")
Omega = 0.5 * (Omega + Omega.T)
Gs = Omega.T @ Gy @ Omega

print("Diffusion map on %d samples (with state similarity kernel)" % N)
eigenvalues, left_eigenvectors, right_eigenvectors = concaust.DMap_K(Gs, num_basis = 5)
print("eigenvalues: ", eigenvalues)

#Visualization
import matplotlib.pyplot as plt
plt.scatter(right_eigenvectors[:,1],right_eigenvectors[:,2])
plt.show()

# Batch processing
#np.savetxt("mess3-p%d-f%d-N%d-bw%.03f-seed%s.txt" % (npast, nfuture, N, bandwidth, seed), right_eigenvectors)
#np.savetxt("mess3-eval-p%d-f%d-N%d-bw%.03f-seed%s.txt" % (npast, nfuture, N, bandwidth, seed), eigenvalues)




